// Generated by CoffeeScript 2.3.1
  //#*
  // The central module for DocPad
  // @module DocPad
  //#

  // =====================================
  // This block *must* come first

  // Node 0.10 compat
  /**
   * Contains methods for managing the DocPad application.
   * This includes managing a DocPad projects files and
   * documents, watching directories, emitting events and
   * managing the node.js/express.js web server.
   * Extends https://github.com/bevry/event-emitter-grouped
   *
   * The class is instantiated in the docpad-server.js file
   * which is the entry point for a DocPad application.
   *
   * 	new DocPad(docpadConfig, function(err, docpad) {
   * 		if (err) {
   * 			return docpadUtil.writeError(err);
   * 		}
   * 		return docpad.action(action, function(err) {
   * 			if (err) {
   * 				return docpadUtil.writeError(err);
   * 			}
   * 			return console.log('OK');
   * 		});
   * 	});
   *
   * @class Docpad
   * @constructor
   * @extends EventEmitterGrouped
   */
var BasePlugin, CSON, Collection, DocPad, DocumentModel, ElementsCollection, EventEmitterGrouped, Events, FileModel, FilesCollection, MetaCollection, Model, PluginLoader, Progress, QueryCollection, ScriptsCollection, StylesCollection, TaskGroup, ambi, balUtil, corePath, docpadUtil, eachr, extendr, extractOptsAndCallback, ignorefs, isUser, lazyRequire, pathUtil, pick, queryEngine, rimraf, safefs, safeps, scandir, setImmediate, superAgent, typeChecker, unbounded, union, uniq, util,
  hasProp = {}.hasOwnProperty,
  indexOf = [].indexOf,
  boundMethodCheck = function(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new Error('Bound instance method accessed before binding'); } };

require('babel-polyfill');

// Better debugging information
require('unbounded').patch();

// Important
pathUtil = require('path');

lazyRequire = require('lazy-require');

corePath = pathUtil.resolve(__dirname, '..', '..');

// =====================================
// Requires

// Standard Library
util = require('util');

// External
queryEngine = require('query-engine');

({uniq, union, pick} = require('underscore'));

CSON = require('cson');

balUtil = require('bal-util');

scandir = require('scandirectory');

extendr = require('extendr');

eachr = require('eachr');

typeChecker = require('typechecker');

ambi = require('ambi');

unbounded = require('unbounded');

({TaskGroup} = require('taskgroup'));

safefs = require('safefs');

safeps = require('safeps');

ignorefs = require('ignorefs');

rimraf = require('rimraf');

Progress = require('progress-title');

superAgent = require('superagent');

extractOptsAndCallback = require('extract-opts');

({EventEmitterGrouped} = require('event-emitter-grouped'));

// Base
({Events, Model, Collection, QueryCollection} = require('./base'));

// Utils
docpadUtil = require('./util');

// Models
FileModel = require('./models/file');

DocumentModel = require('./models/document');

// Collections
FilesCollection = require('./collections/files');

ElementsCollection = require('./collections/elements');

MetaCollection = require('./collections/meta');

ScriptsCollection = require('./collections/scripts');

StylesCollection = require('./collections/styles');

// Plugins
PluginLoader = require('./plugin-loader');

BasePlugin = require('./plugin');

// ---------------------------------
// Helpers
setImmediate = (typeof global !== "undefined" && global !== null ? global.setImmediate : void 0) || process.nextTick; // node 0.8 b/c


// ---------------------------------
// Variables
isUser = docpadUtil.isUser();

DocPad = (function() {
  class DocPad extends EventEmitterGrouped {
    // Libraries
    // Here for legacy API reasons
    //@DocPad: DocPad
    //@Backbone: require('backbone')
    //@queryEngine: queryEngine

    // Allow for `DocPad.create()` as an alias for `new DocPad()`
    // Allow for `DocPad.createInstance()` as an alias for `new DocPad()` (legacy alias)
    static create(...args) {
      return new this(...args);
    }

    static createInstance(...args) {
      return new this(...args);
    }

    // Require a local DocPad file
    // Before v6.73.0 this allowed requiring of files inside src/lib, as well as files inside src
    // Now it only allows requiring of files inside src/lib as that makes more sense
    static require(relativePath) {
      var absolutePath;
      // Absolute the path
      absolutePath = pathUtil.normalize(pathUtil.join(__dirname, relativePath));
      // Check if we are actually a local docpad file
      if (absolutePath.replace(__dirname, '') === absolutePath) {
        throw new Error(`docpad.require is limited to local docpad files only: ${relativePath}`);
      }
      // Require the path
      return require(absolutePath);
    }

    /**
     * Get the DocPad version number
     * @method getVersion
     * @return {Number}
     */
    getVersion() {
      if (this.version == null) {
        this.version = require(this.packagePath).version;
      }
      return this.version;
    }

    /**
     * Get the DocPad version string
     * @method getVersionString
     * @return {String}
     */
    getVersionString() {
      if (docpadUtil.isLocalDocPadExecutable()) {
        return util.format(this.getLocale().versionLocal, this.getVersion(), this.corePath);
      } else {
        return util.format(this.getLocale().versionGlobal, this.getVersion(), this.corePath);
      }
    }

    // Process getters
    /**
     * Get the process platform
     * @method getProcessPlatform
     * @return {Object}
     */
    getProcessPlatform() {
      return process.platform;
    }

    /**
     * Get the process version string
     * @method getProcessVersion
     * @return {String}
     */
    getProcessVersion() {
      return process.version.replace(/^v/, '');
    }

    /**
     * Get the DocPad express.js server instance and, optionally,
     * the node.js https server instance
     * @method getServer
     * @param {Boolean} [both=false]
     * @return {Object}
     */
    getServer(both = false) {
      var serverExpress, serverHttp;
      ({serverExpress, serverHttp} = this);
      if (both) {
        return {serverExpress, serverHttp};
      } else {
        return serverExpress;
      }
    }

    /**
     * Set the express.js server and node.js http server
     * to bind to DocPad
     * @method setServer
     * @param {Object} servers
     */
    setServer(servers) {
      // Apply
      if (servers.serverExpress && servers.serverHttp) {
        this.serverExpress = servers.serverExpress;
        this.serverHttp = servers.serverHttp;
      }
      // Cleanup
      delete this.config.serverHttp;
      delete this.config.serverExpress;
      return delete this.config.server;
    }

    /**
     * Destructor. Close and destroy the node.js http server
     * @private
     * @method destroyServer
     */
    destroyServer() {
      var ref;
      if ((ref = this.serverHttp) != null) {
        ref.close();
      }
      return this.serverHttp = null;
    }

    /**
     * Get the caterpillar logger instance bound to DocPad
     * @method getLogger
     * @return {Object} caterpillar logger
     */
    getLogger() {
      var ref;
      return (ref = this.loggerInstances) != null ? ref.logger : void 0;
    }

    /**
     * Get all the caterpillar logger instances bound to DocPad
     * @method getLoggers
     * @return {Object} collection of caterpillar loggers
     */
    getLoggers() {
      return this.loggerInstances;
    }

    /**
     * Sets the caterpillar logger instances bound to DocPad
     * @method setLoggers
     * @param {Object} loggers
     * @return {Object} logger instances bound to DocPad
     */
    setLoggers(loggers) {
      if (this.loggerInstances) {
        this.warn(this.getLocale().loggersAlreadyDefined);
      } else {
        this.loggerInstances = loggers;
        this.loggerInstances.logger.setConfig({
          dry: true
        });
        this.loggerInstances.console.setConfig({
          dry: false
        }).pipe(process.stdout);
      }
      return loggers;
    }

    /**
     * Destructor. Destroy the caterpillar logger instances bound to DocPad
     * @private
     * @method {Object} destroyLoggers
     */
    destroyLoggers() {
      var key, ref, value;
      if (this.loggerInstances) {
        ref = this.loggerInstances;
        for (key in ref) {
          if (!hasProp.call(ref, key)) continue;
          value = ref[key];
          value.end();
          this.loggerInstances[key] = null;
        }
        this.loggerInstances = null;
      }
      return this;
    }

    /**
     * Update the configuration of the progress instance, to either enable it or disable it
     * Progress will be enabled if the DocPad config 'progress' and 'prompts' are both true
     * @private
     * @method updateProgress
     * @param {boolean} [enabled] manually enable or disable the progress bar
     */
    updateProgress(enabled) {
      var config, debug, docpad, options;
      // Prepare
      docpad = this;
      config = docpad.getConfig();
      debug = this.getDebugging();
      // Enabled
      if (enabled == null) {
        enabled = config.progress && config.prompts;
      }
      // If we are in debug mode, then output more detailed title messages
      options = {};
      if (debug) {
        options.verbose = true;
        options.interval = 0;
      }
      // options.log = true

      // If we wish to have it enabled
      if (enabled) {
        if (this.progressInstance) {
          this.progressInstance.pause().configure(options).resume();
        } else {
          this.progressInstance = Progress.create(options).start();
        }
      } else if (this.progressInstance) {
        this.progressInstance.stop().configure(options);
      }
      // Return
      return this;
    }

    /**
     * Get the action runner instance bound to docpad
     * @method getActionRunner
     * @return {Object} the action runner instance
     */
    getActionRunner() {
      return this.actionRunnerInstance;
    }

    /**
     * Apply the passed DocPad action arguments
     * @method {Object} action
     * @param {Object} args
     * @return {Object}
     */
    action(...args) {
      return docpadUtil.action.apply(this, args);
    }

    /**
     * Get the list of available events
     * @method getEvents
     * @return {Object} string array of event names
     */
    getEvents() {
      return this.events;
    }

    /**
     * Description for getDatabase
     * @method {Object} getDatabase
     */
    getDatabase() {
      return this.database;
    }

    /**
     * Safe method for retrieving the database by
     * either returning the database itself or the temporary
     * database cache
     * @method getDatabaseSafe
     * @return {Object}
     */
    getDatabaseSafe() {
      return this.databaseTempCache || this.database;
    }

    /**
     * Destructor. Destroy the DocPad database
     * @private
     * @method destroyDatabase
     */
    destroyDatabase() {
      if (this.database != null) {
        this.database.destroy();
        this.database = null;
      }
      if (this.databaseTempCache != null) {
        this.databaseTempCache.destroy();
        this.databaseTempCache = null;
      }
      return this;
    }

    /* {
     * A collection of meta elements
    	meta: null  # Elements Collection

     * A collection of script elements
    	scripts: null  # Scripts Collection

     * Collection of style elements
    	styles: null  # Styles Collection
    }  */
    /**
     * Get a block by block name. Optionally clone block.
     * @method getBlock
     * @param {String} name
     * @param {Object} [clone]
     * @return {Object} block
     */
    getBlock(name, clone) {
      var block, classname;
      block = this.blocks[name];
      if (clone) {
        classname = name[0].toUpperCase() + name.slice(1) + 'Collection';
        block = new this[classname](block.models);
      }
      return block;
    }

    /**
     * Set a block by name and value
     * @method setBlock
     * @param {String} name
     * @param {Object} value
     */
    setBlock(name, value) {
      if (this.blocks[name] != null) {
        this.blocks[name].destroy();
        if (value) {
          this.blocks[name] = value;
        } else {
          delete this.blocks[name];
        }
      } else {
        this.blocks[name] = value;
      }
      return this;
    }

    /**
     * Get all blocks
     * @method getBlocks
     * @return {Object} collection of blocks
     */
    getBlocks() {
      return this.blocks;
    }

    /**
     * Set all blocks
     * @method setBlocks
     * @param {Object} blocks
     */
    setBlocks(blocks) {
      var name, value;
      for (name in blocks) {
        if (!hasProp.call(blocks, name)) continue;
        value = blocks[name];
        this.setBlock(name, value);
      }
      return this;
    }

    /**
     * Apply the passed function to each block
     * @method eachBlock
     * @param {Function} fn
     */
    eachBlock(fn) {
      eachr(this.blocks, fn);
      return this;
    }

    /**
     * Destructor. Destroy all blocks
     * @private
     * @method destroyBlocks
     */
    destroyBlocks() {
      var block, name, ref;
      if (this.blocks) {
        ref = this.blocks;
        for (name in ref) {
          if (!hasProp.call(ref, name)) continue;
          block = ref[name];
          block.destroy();
          this.blocks[name] = null;
        }
      }
      return this;
    }

    /**
     * Get a collection by collection name or key.
     * This is often accessed within the docpad.coffee
     * file or a layout/page via @getCollection.
     * Because getCollection returns a docpad collection,
     * a call to this method is often chained with a
     * QueryEngine style query.
     *
     * 	@getCollection('documents').findAllLive({relativeOutDirPath: 'posts'},[{date:-1}])
     *
     * @method getCollection
     * @param {String} value
     * @return {Object} collection
     */
    getCollection(value) {
      var collection, i, j, len, len1, ref, ref1;
      if (value) {
        if (typeof value === 'string') {
          if (value === 'database') {
            return this.getDatabase();
          } else {
            ref = this.collections;
            for (i = 0, len = ref.length; i < len; i++) {
              collection = ref[i];
              if (value === collection.options.name || value === collection.options.key) {
                return collection;
              }
            }
          }
        } else {
          ref1 = this.collections;
          for (j = 0, len1 = ref1.length; j < len1; j++) {
            collection = ref1[j];
            if (value === collection) {
              return collection;
            }
          }
        }
      }
      return null;
    }

    /**
     * Destroy a collection by collection name or key
     * @method destroyCollection
     * @param {String} value
     * @return {Object} description
     */
    destroyCollection(value) {
      if (value) {
        if (typeof value === 'string' && value !== 'database') {
          this.collections = this.collections.filter(function(collection) {
            if (value === collection.options.name || value === collection.options.key) {
              if (collection != null) {
                collection.destroy();
              }
              return false;
            } else {
              return true;
            }
          });
        } else if (value !== this.getDatabase()) {
          this.collections = this.collections.filter(function(collection) {
            if (value === collection) {
              if (collection != null) {
                collection.destroy();
              }
              return false;
            } else {
              return true;
            }
          });
        }
      }
      return null;
    }

    /**
     * Add a collection
     * @method addCollection
     * @param {Object} collection
     */
    addCollection(collection) {
      if (collection && (collection !== this.getDatabase() && collection !== this.getCollection(collection))) {
        this.collections.push(collection);
      }
      return this;
    }

    /**
     * Set a name for a collection.
     * A collection can have multiple names
     *
     * The partials plugin (https://github.com/docpad/docpad-plugin-partials)
     * creates a live collection and passes this to setCollection with
     * the name 'partials'.
     *
     * 	# Add our partials collection
     *	docpad.setCollection('partials', database.createLiveChildCollection()
     *		.setQuery('isPartial', {
     *				$or:
     *					isPartial: true
     *					fullPath: $startsWith: config.partialsPath
     *		})
     *		.on('add', (model) ->
     *			docpad.log('debug', util.format(locale.addingPartial, model.getFilePath()))
     *			model.setDefaults(
     *				isPartial: true
     *				render: false
     *				write: false
     *			)
     *		)
     *	)
     *
     *
     * @method setCollection
     * @param {String} name the name to give to the collection
     * @param {Object} collection a DocPad collection
     */
    setCollection(name, collection) {
      if (collection) {
        if (name) {
          collection.options.name = name;
          if (this.getCollection(name) !== collection) {
            this.destroyCollection(name);
          }
        }
        return this.addCollection(collection);
      } else {
        return this.destroyCollection(name);
      }
    }

    /**
     * Get the DocPad project's collections
     * @method getCollections
     * @return {Object} the collections
     */
    getCollections() {
      return this.collections;
    }

    /**
     * Set the DocPad project's collections
     * @method setCollections
     */
    setCollections(collections) {
      var i, len, name, value;
      if (Array.isArray(collections)) {
        for (i = 0, len = collections.length; i < len; i++) {
          value = collections[i];
          this.addCollection(value);
        }
      } else {
        for (name in collections) {
          if (!hasProp.call(collections, name)) continue;
          value = collections[name];
          this.setCollection(name, value);
        }
      }
      return this;
    }

    /**
     * Apply the passed function to each collection
     * @method eachCollection
     * @param {Function} fn
     */
    eachCollection(fn) {
      var collection, i, index, len, ref;
      fn(this.getDatabase(), 'database');
      ref = this.collections;
      for (index = i = 0, len = ref.length; i < len; index = ++i) {
        collection = ref[index];
        fn(collection, collection.options.name || collection.options.key || index);
      }
      return this;
    }

    /**
     * Destructor. Destroy the DocPad project's collections.
     * @private
     * @method destroyCollections
     */
    destroyCollections() {
      var collection, i, len, ref;
      if (this.collections) {
        ref = this.collections;
        for (i = 0, len = ref.length; i < len; i++) {
          collection = ref[i];
          collection.destroy();
        }
        this.collections = [];
      }
      return this;
    }

    // ---------------------------------
    // Collection Helpers
    /**
     * Get all the files in the DocPad database (will use live collections)
     * @method getFiles
     * @param {Object} query
     * @param {Object} sorting
     * @param {Object} paging
     * @return {Object} collection
     */
    getFiles(query, sorting, paging) {
      var collection, key;
      key = JSON.stringify({query, sorting, paging});
      collection = this.getCollection(key);
      if (!collection) {
        collection = this.getDatabase().findAllLive(query, sorting, paging);
        collection.options.key = key;
        this.addCollection(collection);
      }
      return collection;
    }

    /**
     * Get a single file based on a query
     * @method getFile
     * @param {Object} query
     * @param {Object} sorting
     * @param {Object} paging
     * @return {Object} a file
     */
    getFile(query, sorting, paging) {
      var file;
      file = this.getDatabase().findOne(query, sorting, paging);
      return file;
    }

    /**
     * Get files at a path
     * @method getFilesAtPath
     * @param {String} path
     * @param {Object} sorting
     * @param {Object} paging
     * @return {Object} files
     */
    getFilesAtPath(path, sorting, paging) {
      var files, query;
      query = {
        $or: [
          {
            relativePath: {
              $startsWith: path
            }
          },
          {
            fullPath: {
              $startsWith: path
            }
          }
        ]
      };
      files = this.getFiles(query, sorting, paging);
      return files;
    }

    /**
     * Get a file at a relative or absolute path or url
     * @method getFileAtPath
     * @param {String} path
     * @param {Object} sorting
     * @param {Object} paging
     * @return {Object} a file
     */
    getFileAtPath(path, sorting, paging) {
      var file;
      file = this.getDatabase().fuzzyFindOne(path, sorting, paging);
      return file;
    }

    // TODO: Does this still work???
    /**
     * Get a file by its url
     * @method getFileByUrl
     * @param {String} url
     * @param {Object} [opts={}]
     * @return {Object} a file
     */
    getFileByUrl(url, opts = {}) {
      var file;
      if (opts.collection == null) {
        opts.collection = this.getDatabase();
      }
      file = opts.collection.get(this.filesByUrl[url]);
      return file;
    }

    /**
     * Get a file by its id
     * @method getFileById
     * @param {String} id
     * @param {Object} [opts={}]
     * @return {Object} a file
     */
    getFileById(id, opts = {}) {
      var file;
      if (opts.collection == null) {
        opts.collection = this.getDatabase();
      }
      file = opts.collection.get(id);
      return file;
    }

    /**
     * Remove the query string from a url
     * Pathname convention taken from document.location.pathname
     * @method getUrlPathname
     * @param {String} url
     * @return {String}
     */
    getUrlPathname(url) {
      return url.replace(/\?.*/, '');
    }

    /**
     * Get a file by its route and return
     * it to the supplied callback.
     * @method getFileByRoute
     * @param {String} url
     * @param {Object} next
     * @param {Error} next.err
     * @param {String} next.file
     */
    getFileByRoute(url, next) {
      var cleanUrl, database, docpad, file;
      // Prepare
      docpad = this;
      // If we have not performed a generation yet then wait until the initial generation has completed
      if (docpad.generated === false) {
        // Wait until generation has completed and recall ourselves
        docpad.once('generated', function() {
          return docpad.getFileByRoute(url, next);
        });
        // hain
        return this;
      }
      // @TODO the above causes a signifcant delay when importing external documents (like tumblr data) into the database
      // we need to figure out a better way of doing this
      // perhaps it is via `writeSource: once` for imported documents
      // or providing an option to disable this so it forward onto the static handler instead

      // Prepare
      database = docpad.getDatabaseSafe();
      // Fetch
      cleanUrl = docpad.getUrlPathname(url);
      file = docpad.getFileByUrl(url, {
        collection: database
      }) || docpad.getFileByUrl(cleanUrl, {
        collection: database
      });
      // Forward
      next(null, file);
      return this;
    }

    // TODO: What on earth is a selector?
    /**
     * Get a file by its selector
     * @method getFileBySelector
     * @param {Object} selector
     * @param {Object} [opts={}]
     * @return {Object} a file
     */
    getFileBySelector(selector, opts = {}) {
      var file;
      if (opts.collection == null) {
        opts.collection = this.getDatabase();
      }
      file = opts.collection.get(this.filesBySelector[selector]);
      if (!file) {
        file = opts.collection.fuzzyFindOne(selector);
        if (file) {
          this.filesBySelector[selector] = file.id;
        }
      }
      return file;
    }

    /**
     * Get Skeletons
     * Get all the available skeletons with their details and
     * return this collection to the supplied callback.
     * @method getSkeletons
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.skeletonsCollection DocPad collection of skeletons
     * @return {Object} DocPad skeleton collection
     */
    getSkeletons(next) {
      var docpad, locale;
      // Prepare
      docpad = this;
      locale = this.getLocale();
      // Check if we have cached locally
      if (this.skeletonsCollection != null) {
        return next(null, this.skeletonsCollection);
      }
      // Fetch the skeletons from the exchange
      this.skeletonsCollection = new Collection();
      this.skeletonsCollection.comparator = queryEngine.generateComparator({
        position: 1,
        name: 1
      });
      this.getExchange(function(err, exchange) {
        var index;
        if (err) {
          // Check
          return next(err);
        }
        // Prepare
        index = 0;
        // If we have the exchange data, then add the skeletons from it
        if (exchange) {
          eachr(exchange.skeletons, function(skeleton, skeletonKey) {
            if (skeleton.id == null) {
              skeleton.id = skeletonKey;
            }
            if (skeleton.name == null) {
              skeleton.name = skeletonKey;
            }
            if (skeleton.position == null) {
              skeleton.position = index;
            }
            docpad.skeletonsCollection.add(new Model(skeleton));
            return ++index;
          });
        }
        // Add No Skeleton Option
        docpad.skeletonsCollection.add(new Model({
          id: 'none',
          name: locale.skeletonNoneName,
          description: locale.skeletonNoneDescription,
          position: index
        }));
        // Return Collection
        return next(null, docpad.skeletonsCollection);
      });
      return this;
    }

    /**
     * Get Complete Template Data
     * @method getTemplateData
     * @param {Object} userTemplateData
     * @return {Object} templateData
     */
    getTemplateData(userTemplateData) {
      var base, base1, base2, docpad, locale, templateData;
      // Prepare
      userTemplateData || (userTemplateData = {});
      docpad = this;
      locale = this.getLocale();
      // Set the initial docpad template data
      if (this.initialTemplateData == null) {
        this.initialTemplateData = {
          // Site Properties
          site: {},
          // Environment
          getEnvironment: function() {
            return docpad.getEnvironment();
          },
          // Environments
          getEnvironments: function() {
            return docpad.getEnvironments();
          },
          // Set that we reference other files
          referencesOthers: function(flag) {
            var document;
            document = this.getDocument();
            document.referencesOthers();
            return null;
          },
          // Get the Document
          getDocument: function() {
            return this.documentModel;
          },
          // Get a Path in respect to the current document
          getPath: function(path, parentPath) {
            var document;
            document = this.getDocument();
            path = document.getPath(path, parentPath);
            return path;
          },
          // Get Files
          getFiles: function(query, sorting, paging) {
            var result;
            this.referencesOthers();
            result = docpad.getFiles(query, sorting, paging);
            return result;
          },
          // Get another file's URL based on a relative path
          getFile: function(query, sorting, paging) {
            var result;
            this.referencesOthers();
            result = docpad.getFile(query, sorting, paging);
            return result;
          },
          // Get Files At Path
          getFilesAtPath: function(path, sorting, paging) {
            var result;
            this.referencesOthers();
            path = this.getPath(path);
            result = docpad.getFilesAtPath(path, sorting, paging);
            return result;
          },
          // Get another file's model based on a relative path
          getFileAtPath: function(relativePath) {
            var path, result;
            this.referencesOthers();
            path = this.getPath(relativePath);
            result = docpad.getFileAtPath(path);
            return result;
          },
          // Get a specific file by its id
          getFileById: function(id) {
            var result;
            this.referencesOthers();
            result = docpad.getFileById(id);
            return result;
          },
          // Get the entire database
          getDatabase: function() {
            this.referencesOthers();
            return docpad.getDatabase();
          },
          // Get a pre-defined collection
          getCollection: function(name) {
            this.referencesOthers();
            return docpad.getCollection(name);
          },
          // Get a block
          getBlock: function(name) {
            return docpad.getBlock(name, true);
          },
          // Include another file taking in a relative path
          include: function(subRelativePath, strict = true) {
            var err, file;
            file = this.getFileAtPath(subRelativePath);
            if (file) {
              if (strict && file.get('rendered') === false) {
                if (docpad.getConfig().renderPasses === 1) {
                  docpad.warn(util.format(locale.renderedEarlyViaInclude, subRelativePath));
                }
                return null;
              }
              return file.getOutContent();
            } else {
              err = new Error(util.format(locale.includeFailed, subRelativePath));
              throw err;
            }
          }
        };
      }
      // Fetch our result template data
      templateData = extendr.extend({}, this.initialTemplateData, this.pluginsTemplateData, this.getConfig().templateData, userTemplateData);
      // Add site data
      (base = templateData.site).url || (base.url = this.getSimpleServerUrl());
      (base1 = templateData.site).date || (base1.date = new Date());
      (base2 = templateData.site).keywords || (base2.keywords = []);
      if (typeChecker.isString(templateData.site.keywords)) {
        templateData.site.keywords = templateData.site.keywords.split(/,\s*/g);
      }
      // Return
      return templateData;
    }

    /**
     * Get the locale (language code and locale code)
     * @method getLocale
     * @return {Object} locale
     */
    getLocale() {
      var code, codes, config, locales;
      if ((this.locale != null) === false) {
        config = this.getConfig();
        codes = uniq(['en', safeps.getLanguageCode(config.localeCode), safeps.getLanguageCode(safeps.getLocaleCode()), safeps.getLocaleCode(config.localeCode), safeps.getLocaleCode(safeps.getLocaleCode())]);
        locales = ((function() {
          var i, len, results;
          results = [];
          for (i = 0, len = codes.length; i < len; i++) {
            code = codes[i];
            results.push(this.loadLocale(code));
          }
          return results;
        }).call(this)).filter(function(l) {
          return l;
        });
        this.locale = extendr.extend(...locales);
      }
      return this.locale;
    }

    /**
     * Load the locale
     * @method loadLocale
     * @param {String} code
     * @return {Object} locale
     */
    loadLocale(code) {
      var docpad, locale, localeFilename, localePath;
      // Prepare
      docpad = this;
      // Check if it exists
      localeFilename = `${code}.cson`;
      localePath = pathUtil.join(this.localePath, localeFilename);
      if (!safefs.existsSync(localePath)) {
        return null;
      }
      // Load it
      locale = CSON.parseCSONFile(localePath);
      // Log the error in the background and continue
      if (locale instanceof Error) {
        locale.context = `Failed to parse the CSON locale file: ${localePath}`;
        docpad.error(locale); // @TODO: should this be a fatal error instead?
        return null;
      }
      // Success
      return locale;
    }

    // -----------------------------
    // Environments
    /**
     * Get the DocPad environment, eg: development,
     * production or static
     * @method getEnvironment
     * @return {String} the environment
     */
    getEnvironment() {
      var env;
      env = this.getConfig().env || 'development';
      return env;
    }

    /**
     * Get the environments
     * @method getEnvironments
     * @return {Array} array of environment strings
     */
    getEnvironments() {
      var env, envs;
      env = this.getEnvironment();
      envs = env.split(/[, ]+/);
      return envs;
    }

    /**
     * Get the DocPad configuration. Commonly
     * called within the docpad.coffee file or within
     * plugins to access application specific configurations.
     * 	serverExtend: (opts) ->
     * Extract the server from the options
    		{server} = opts
    		docpad = @docpad

     * As we are now running in an event,
     * ensure we are using the latest copy of the docpad configuraiton
     * and fetch our urls from it
    		latestConfig = docpad.getConfig()
    		oldUrls = latestConfig.templateData.site.oldUrls or []
    		newUrl = latestConfig.templateData.site.url

     * Redirect any requests accessing one of our sites oldUrls to the new site url
    		server.use (req,res,next) ->
    			...
     * @method getConfig
     * @return {Object} the DocPad configuration object
     */
    getConfig() {
      return this.config || {};
    }

    /**
     * Get the port that DocPad is listening on (eg 9778)
     * @method getPort
     * @return {Number} the port number
     */
    getPort() {
      var ref, ref1;
      return (ref = (ref1 = this.getConfig().port) != null ? ref1 : require('hostenv').PORT) != null ? ref : 9778;
    }

    /**
     * Get the Hostname
     * @method getHostname
     * @return {String}
     */
    getHostname() {
      var ref, ref1;
      return (ref = (ref1 = this.getConfig().hostname) != null ? ref1 : require('hostenv').HOSTNAME) != null ? ref : '0.0.0.0';
    }

    /**
     * Get address
     * @method getServerUrl
     * @param {Object} [opts={}]
     * @return {String}
     */
    getServerUrl(opts = {}) {
      var ref;
      if (opts.hostname == null) {
        opts.hostname = this.getHostname();
      }
      if (opts.port == null) {
        opts.port = this.getPort();
      }
      if (opts.simple == null) {
        opts.simple = false;
      }
      if (opts.simple === true && ((ref = opts.hostname) === '0.0.0.0' || ref === '::' || ref === '::1')) {
        return `http://127.0.0.1:${opts.port}`;
      } else {
        return `http://${opts.hostname}:${opts.port}`;
      }
    }

    /**
     * Get simple server URL (changes 0.0.0.0, ::, and ::1 to 127.0.0.1)
     * @method getSimpleServerUrl
     * @param {Object} [opts={}]
     * @param {Boolean} [opts.simple=true]
     * @return {String}
     */
    getSimpleServerUrl(opts = {}) {
      opts.simple = true;
      return this.getServerUrl(opts);
    }

    // =================================
    // Initialization Functions
    /**
     * Create our own custom TaskGroup instance for DocPad.
     * That will listen to tasks as they execute and provide debugging information.
     * @method createTaskGroup
     * @param {Object} opts
     * @return {TaskGroup}
     */
    createTaskGroup(...opts) {
      var docpad, progress, tasks;
      boundMethodCheck(this, DocPad);
      docpad = this;
      progress = docpad.progressInstance;
      tasks = TaskGroup.create(...opts);
      // Listen to executing tasks and output their progress
      tasks.on('running', function() {
        var config, name, totals;
        config = tasks.getConfig();
        name = tasks.getNames();
        if (progress) {
          totals = tasks.getItemTotals();
          return progress.update(name, totals);
        } else {
          return docpad.log('debug', name + ' > running');
        }
      });
      // Listen to executing tasks and output their progress
      tasks.on('item.add', function(item) {
        var config, name;
        config = tasks.getConfig();
        name = item.getNames();
        if (!progress) {
          docpad.log('debug', name + ' > added');
        }
        // Listen to executing tasks and output their progress
        item.on('started', function(item) {
          var totals;
          config = tasks.getConfig();
          name = item.getNames();
          if (progress) {
            totals = tasks.getItemTotals();
            return progress.update(name, totals);
          } else {
            return docpad.log('debug', name + ' > started');
          }
        });
        // Listen to executing tasks and output their progress
        return item.done(function(err) {
          var totals;
          config = tasks.getConfig();
          name = item.getNames();
          if (progress) {
            totals = tasks.getItemTotals();
            return progress.update(name, totals);
          } else {
            return docpad.log('debug', name + ' > done');
          }
        });
      });
      // Return
      return tasks;
    }

    /**
     * Constructor method. Sets up the DocPad instance.
     * next(err)
     * @method constructor
     * @param {Object} instanceConfig
     * @param {Function} next callback
     * @param {Error} next.err
     */
    constructor(instanceConfig, next) {
      var action, configEventContext, docpad, i, len, logger, loggerConsole, loggers, methodName, ref, ref1;
      super();
      this.createTaskGroup = this.createTaskGroup.bind(this);
      [instanceConfig, next] = extractOptsAndCallback(instanceConfig, next);
      docpad = this;
      ref = "action\nlog warn error fatal inspector notify checkRequest\nserverMiddlewareRouter serverMiddlewareHeader serverMiddleware404 serverMiddleware500\ndestroyWatchers".split(/\s+/);
      // Binders
      // Using this over coffescript's => on class methods, ensures that the method length is kept
      for (i = 0, len = ref.length; i < len; i++) {
        methodName = ref[i];
        this[methodName] = this[methodName].bind(this);
      }
      // Allow DocPad to have unlimited event listeners
      this.setMaxListeners(0);
      // Setup configuration event wrappers
      configEventContext = {docpad}; // here to allow the config event context to persist between event calls
      this.getEvents().forEach(function(eventName) {
        // Bind to the event
        return docpad.on(eventName, function(opts, next) {
          var args, eventHandler, ref1;
          eventHandler = (ref1 = docpad.getConfig().events) != null ? ref1[eventName] : void 0;
          // Fire the config event handler for this event, if it exists
          if (typeChecker.isFunction(eventHandler)) {
            args = [opts, next];
            return ambi(unbounded.binder.call(eventHandler, configEventContext), ...args);
          } else {
            // It doesn't exist, so lets continue
            return next();
          }
        });
      });
      // Create our action runner
      this.actionRunnerInstance = this.createTaskGroup('action runner', {
        abortOnError: false,
        destroyOnceDone: false
      }).whenDone(function(err) {
        var ref1;
        if ((ref1 = docpad.progressInstance) != null) {
          ref1.update('');
        }
        if (err) {
          return docpad.error(err);
        }
      });
      // Initialize the loggers
      if ((loggers = instanceConfig.loggers)) {
        delete instanceConfig.loggers;
      } else {
        // Create
        logger = require('caterpillar').create({
          lineOffset: 2
        });
        // console
        loggerConsole = logger.pipe(require('caterpillar-filter').create()).pipe(require('caterpillar-human').create());
        // Apply
        loggers = {
          logger,
          console: loggerConsole
        };
      }
      // Apply the loggers
      safefs.unlink(this.debugLogPath, function() {}); // Remove the old debug log file
      this.setLoggers(loggers); // Apply the logger streams
      this.setLogLevel((ref1 = instanceConfig.logLevel) != null ? ref1 : this.initialConfig.logLevel); // Set the default log level
      
      // Log to bubbled events
      this.on('log', function(...args) {
        return docpad.log.apply(this, args);
      });
      // Dereference and initialise advanced variables
      // we deliberately ommit initialTemplateData here, as it is setup in getTemplateData
      this.slowPlugins = {};
      this.loadedPlugins = {};
      this.exchange = {};
      this.pluginsTemplateData = {};
      this.instanceConfig = {};
      this.collections = [];
      this.blocks = {};
      this.filesByUrl = {};
      this.filesBySelector = {};
      this.filesByOutPath = {};
      this.database = new FilesCollection(null, {
        name: 'database'
      }).on('remove', function(model, options) {
        var j, len1, outPath, ref2, updatedModels, url;
        // Skip if we are not a writeable file
        if (model.get('write') === false) {
          return;
        }
        ref2 = model.get('urls') || [];
        // Delete the urls
        for (j = 0, len1 = ref2.length; j < len1; j++) {
          url = ref2[j];
          delete docpad.filesByUrl[url];
        }
        // Ensure we regenerate anything (on the next regeneration) that was using the same outPath
        outPath = model.get('outPath');
        if (outPath) {
          updatedModels = docpad.database.findAll({outPath});
          updatedModels.remove(model);
          updatedModels.each(function(model) {
            return model.set({
              'mtime': new Date()
            });
          });
          // Log
          docpad.log('debug', 'Updated mtime for these models due to remove of a similar one', updatedModels.pluck('relativePath'));
        }
        // Return safely
        return true;
      }).on('add change:urls', function(model) {
        var j, k, len1, len2, ref2, ref3, url;
        // Skip if we are not a writeable file
        if (model.get('write') === false) {
          return;
        }
        ref2 = model.previous('urls') || [];
        // Delete the old urls
        for (j = 0, len1 = ref2.length; j < len1; j++) {
          url = ref2[j];
          delete docpad.filesByUrl[url];
        }
        ref3 = model.get('urls');
        // Add the new urls
        for (k = 0, len2 = ref3.length; k < len2; k++) {
          url = ref3[k];
          docpad.filesByUrl[url] = model.cid;
        }
        // Return safely
        return true;
      }).on('add change:outPath', function(model) {
        var base, existingModel, existingModelId, existingModelPath, modelPath, outPath, previousModelId, previousModels, previousOutPath;
        // Skip if we are not a writeable file
        if (model.get('write') === false) {
          return;
        }
        // Check if we have changed our outPath
        previousOutPath = model.previous('outPath');
        if (previousOutPath) {
          // Ensure we regenerate anything (on the next regeneration) that was using the same outPath
          previousModels = docpad.database.findAll({
            outPath: previousOutPath
          });
          previousModels.remove(model);
          previousModels.each(function(model) {
            return model.set({
              'mtime': new Date()
            });
          });
          // Log
          docpad.log('debug', 'Updated mtime for these models due to addition of a similar one', previousModels.pluck('relativePath'));
          // Update the cache entry with another file that has the same outPath or delete it if there aren't any others
          previousModelId = docpad.filesByOutPath[previousOutPath];
          if (previousModelId === model.id) {
            if (previousModels.length) {
              docpad.filesByOutPath[previousOutPath] = previousModelId;
            } else {
              delete docpad.filesByOutPath[previousOutPath];
            }
          }
        }
        // Update the cache entry and fetch the latest if it was already set
        if ((outPath = model.get('outPath'))) {
          existingModelId = (base = docpad.filesByOutPath)[outPath] != null ? base[outPath] : base[outPath] = model.id;
          if (existingModelId !== model.id) {
            existingModel = docpad.database.get(existingModelId);
            if (existingModel) {
              // We have a conflict, let the user know
              modelPath = model.get('fullPath') || (model.get('relativePath') + ':' + model.id);
              existingModelPath = existingModel.get('fullPath') || (existingModel.get('relativePath') + ':' + existingModel.id);
              docpad.warn(util.format(docpad.getLocale().outPathConflict, outPath, modelPath, existingModelPath));
            } else {
              // There reference was old, update it with our new one
              docpad.filesByOutPath[outPath] = model.id;
            }
          }
        }
        // Return safely
        return true;
      });
      this.userConfig = extendr.dereferenceJSON(this.userConfig);
      this.initialConfig = extendr.dereferenceJSON(this.initialConfig);
      // Extract action
      if (instanceConfig.action != null) {
        action = instanceConfig.action;
      } else {
        action = 'load ready';
      }
      // Check if we want to perform an action
      if (action) {
        this.action(action, instanceConfig, function(err) {
          if (next != null) {
            return next(err, docpad);
          } else if (err) {
            return docpad.fatal(err);
          }
        });
      } else {
        if (typeof next === "function") {
          next(null, docpad);
        }
      }
      // Chain
      this;
    }

    /**
     * Destructor. Destroy the DocPad instance
     * This is an action, and should be called as such
     * E.g. docpad.action('destroy', next)
     * @method destroy
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     */
    destroy(opts, next) {
      var docpad;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      // Destroy Regenerate Timer
      docpad.destroyRegenerateTimer();
      // Wait one second to wait for any logging to complete
      docpadUtil.wait(1000, function() {
        // Destroy Plugins
        return docpad.emitSerial('docpadDestroy', function(err) {
          if (err) {
            return typeof next === "function" ? next(err) : void 0;
          }
          // Destroy Plugins
          docpad.destroyPlugins();
          // Destroy Server
          docpad.destroyServer();
          // Destroy Watchers
          docpad.destroyWatchers();
          // Destroy Blocks
          docpad.destroyBlocks();
          // Destroy Collections
          docpad.destroyCollections();
          // Destroy Database
          docpad.destroyDatabase();
          // Destroy progress
          docpad.updateProgress(false);
          // Destroy Logging
          docpad.destroyLoggers();
          // Destroy Process Listners
          process.removeListener('uncaughtException', docpad.error);
          // Destroy DocPad Listeners
          docpad.removeAllListeners();
          return typeof next === "function" ? next() : void 0;
        });
      });
      return this;
    }

    /**
     * Emit event, serial
     * @private
     * @method emitSerial
     * @param {String} eventName
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     */
    emitSerial(eventName, opts, next) {
      var docpad, locale;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      locale = docpad.getLocale();
      // Log
      docpad.log('debug', util.format(locale.emittingEvent, eventName));
      // Emit
      super.emitSerial(eventName, opts, function(err) {
        if (err) {
          // Check
          return next(err);
        }
        // Log
        docpad.log('debug', util.format(locale.emittedEvent, eventName));
        // Forward
        return next(err);
      });
      return this;
    }

    /**
     * Emit event, parallel
     * @private
     * @method emitParallel
     * @param {String} eventName
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     */
    emitParallel(eventName, opts, next) {
      var docpad, locale;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      locale = docpad.getLocale();
      // Log
      docpad.log('debug', util.format(locale.emittingEvent, eventName));
      // Emit
      super.emitParallel(eventName, opts, function(err) {
        if (err) {
          // Check
          return next(err);
        }
        // Log
        docpad.log('debug', util.format(locale.emittedEvent, eventName));
        // Forward
        return next(err);
      });
      return this;
    }

    // =================================
    // Helpers
    /**
     * Get the ignore options for the DocPad project
     * @method getIgnoreOpts
     * @return {Array} string array of ignore options
     */
    getIgnoreOpts() {
      return pick(this.config, ['ignorePaths', 'ignoreHiddenFiles', 'ignoreCommonPatterns', 'ignoreCustomPatterns']);
    }

    /**
     * Is the supplied path ignored?
     * @method isIgnoredPath
     * @param {String} path
     * @param {Object} [opts={}]
     * @return {Boolean}
     */
    isIgnoredPath(path, opts = {}) {
      opts = extendr.extend(this.getIgnoreOpts(), opts);
      return ignorefs.isIgnoredPath(path, opts);
    }

    /**
     * Scan directory
     * @method scandir
     * @param {Object} [opts={}]
     */
    //NB: How does this work? What is returned?
    //Does it require a callback (next) passed as
    //one of the options
    scandir(opts = {}) {
      opts = extendr.extend(this.getIgnoreOpts(), opts);
      return scandir(opts);
    }

    /**
     * Watch Directory. Wrapper around the Bevry watchr
     * module (https://github.com/bevry/watchr). Used
     * internally by DocPad to watch project documents
     * and files and then activate the regeneration process
     * when any of those items are updated.
     * @private
     * @method watchdir
     * @param {String} path - the path to watch
     * @param {Object} listeners - listeners to attach to the watcher
     * @param {Function} next - completion callback accepting error
     * @return {Object} the watcher
     */
    watchdir(path, listeners, next) {
      var key, opts, stalker, value;
      opts = extendr.extend(this.getIgnoreOpts(), this.config.watchOptions || {});
      stalker = require('watchr').create(path);
      for (key in listeners) {
        if (!hasProp.call(listeners, key)) continue;
        value = listeners[key];
        stalker.on(key, value);
      }
      stalker.setConfig(opts);
      stalker.watch(next);
      return stalker;
    }

    /**
     * Watch Directories. Wrapper around watchdir.
     * @private
     * @method watchdirs
     * @param {Array} paths - the paths to watch
     * @param {Object} listeners - listeners to attach to the watcher
     * @param {Function} next - completion callback accepting error and watchers/stalkers
     */
    watchdirs(paths, listeners, next) {
      var docpad, stalkers, tasks;
      docpad = this;
      stalkers = [];
      tasks = new TaskGroup('watching directories').setConfig({
        concurrency: 0
      }).done(function(err) {
        var i, len, stalker;
        if (err) {
          for (i = 0, len = stalkers.length; i < len; i++) {
            stalker = stalkers[i];
            stalker.close();
          }
          return next(err);
        } else {
          return next(err, stalkers);
        }
      });
      paths.forEach(function(path) {
        return tasks.addTask(`watching ${path}`, function(done) {
          // check if the dir exists first as reloadPaths may not apparently
          return safefs.exists(path, function(exists) {
            if (!exists) {
              return done();
            }
            return stalkers.push(docpad.watchdir(path, listeners, done));
          });
        });
      });
      tasks.run();
      return this;
    }

    // =================================
    // Setup and Loading
    /**
     * DocPad is ready. Peforms the tasks needed after DocPad construction
     * and DocPad has loaded. Triggers the docpadReady event.
     * next(err,docpadInstance)
     * @private
     * @method ready
     * @param {Object} [opts]
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.docpadInstance
     */
    ready(opts, next) {
      var config, docpad, instanceConfig, locale, pluginName, pluginsList, tasks;
      // Prepare
      [instanceConfig, next] = extractOptsAndCallback(instanceConfig, next);
      docpad = this;
      config = this.getConfig();
      locale = this.getLocale();
      // Render Single Extensions
      this.DocumentModel.prototype.defaults.renderSingleExtensions = config.renderSingleExtensions;
      // Version Check
      this.compareVersion();
      // Fetch the plugins
      if (this.getDebugging()) {
        pluginsList = ((function() {
          var i, len, ref, results;
          ref = Object.keys(this.loadedPlugins).sort();
          results = [];
          for (i = 0, len = ref.length; i < len; i++) {
            pluginName = ref[i];
            results.push(`${pluginName} v${this.loadedPlugins[pluginName].version}`);
          }
          return results;
        }).call(this)).join(', ');
      } else {
        pluginsList = Object.keys(this.loadedPlugins).sort().join(', ');
      }
      // Welcome Output
      docpad.log('info', util.format(locale.welcome, this.getVersionString()));
      docpad.log('notice', locale.welcomeDonate);
      docpad.log('info', locale.welcomeContribute);
      docpad.log('info', util.format(locale.welcomePlugins, pluginsList));
      docpad.log('info', util.format(locale.welcomeEnvironment, this.getEnvironment()));
      // Prepare
      tasks = this.createTaskGroup('ready tasks', {
        next: function(err) {
          if (err) {
            // Error?
            return docpad.error(err);
          }
          return typeof next === "function" ? next(null, docpad) : void 0;
        }
      });
      // kept here in case plugins use it
      tasks.addTask('welcome event', function(complete) {
        if (!config.welcome) {
          // No welcome
          return complete();
        }
        // Welcome
        return docpad.emitSerial('welcome', {docpad}, complete);
      });
      tasks.addTask('emit docpadReady', function(complete) {
        return docpad.emitSerial('docpadReady', {docpad}, complete);
      });
      // Run tasks
      tasks.run();
      return this;
    }

    /**
     * Performs the merging of the passed configuration objects
     * @private
     * @method mergeConfigurations
     * @param {Object} configPackages
     * @param {Object} configsToMerge
     */
    mergeConfigurations(configPackages, configsToMerge) {
      var configPackage, env, envConfig, envs, i, j, len, len1, ref;
      // Prepare
      envs = this.getEnvironments();
// Figure out merging
      for (i = 0, len = configPackages.length; i < len; i++) {
        configPackage = configPackages[i];
        if (!configPackage) {
          continue;
        }
        configsToMerge.push(configPackage);
        for (j = 0, len1 = envs.length; j < len1; j++) {
          env = envs[j];
          envConfig = (ref = configPackage.environments) != null ? ref[env] : void 0;
          if (envConfig) {
            configsToMerge.push(envConfig);
          }
        }
      }
      // Merge
      extendr.deep(...configsToMerge);
      return this;
    }

    /**
     * Set the instance configuration
     * by merging the properties of the passed object
     * with the existing DocPad instanceConfig object
     * @private
     * @method setInstanceConfig
     * @param {Object} instanceConfig
     */
    setInstanceConfig(instanceConfig) {
      var logLevel;
      // Merge in the instance configurations
      if (instanceConfig) {
        logLevel = this.getLogLevel();
        extendr.deepDefaults(this.instanceConfig, instanceConfig);
        if (this.config) { // @TODO document why there is the if
          extendr.deepDefaults(this.config, instanceConfig);
        }
        if (instanceConfig.logLevel && instanceConfig.logLevel !== logLevel) {
          this.setLogLevel(instanceConfig.logLevel);
        }
      }
      return this;
    }

    /**
     * Set the DocPad configuration object.
     * Performs a number of tasks, including
     * merging the pass instanceConfig with DocPad's
     * other config objects.
     * next(err,config)
     * @private
     * @method setConfig
     * @param {Object} instanceConfig
     * @param {Object} next
     * @param {Error} next.err
     * @param {Object} next.config
     */
    setConfig(instanceConfig, next) {
      var configPackages, configsToMerge, docpad, i, j, k, key, len, len1, len2, len3, locale, m, postTasks, ref, ref1, type, typePath, typePaths;
      // Prepare
      [instanceConfig, next] = extractOptsAndCallback(instanceConfig, next);
      docpad = this;
      locale = this.getLocale();
      if (instanceConfig) {
        // Apply the instance configuration, generally we won't have it at this level
        // as it would have been applied earlier the load step
        this.setInstanceConfig(instanceConfig);
      }
      // Apply the environment
      // websitePackageConfig.env is left out of the detection here as it is usually an object
      // that is already merged with our process.env by the environment runner
      // rather than a string which is the docpad convention
      this.config.env = this.instanceConfig.env || this.websiteConfig.env || this.initialConfig.env || process.env.NODE_ENV;
      // Merge configurations
      configPackages = [this.initialConfig, this.userConfig, this.websiteConfig, this.instanceConfig];
      configsToMerge = [this.config];
      docpad.mergeConfigurations(configPackages, configsToMerge);
      // Extract and apply the server
      this.setServer(extendr.defaults({
        serverHttp: this.config.serverHttp,
        serverExpress: this.config.serverExpress
      }, this.config.server || {}));
      // Extract and apply the logger
      this.setLogLevel(this.config.logLevel);
      // Update the progress bar configuration
      this.updateProgress();
      // Resolve any paths
      this.config.rootPath = pathUtil.resolve(this.config.rootPath);
      this.config.outPath = pathUtil.resolve(this.config.rootPath, this.config.outPath);
      this.config.srcPath = pathUtil.resolve(this.config.rootPath, this.config.srcPath);
      this.config.databaseCachePath = pathUtil.resolve(this.config.rootPath, this.config.databaseCachePath);
      this.config.packagePath = pathUtil.resolve(this.config.rootPath, this.config.packagePath);
      ref = ['documents', 'files', 'layouts'];
      // Resolve Documents, Files, Layouts paths
      for (i = 0, len = ref.length; i < len; i++) {
        type = ref[i];
        typePaths = this.config[type + 'Paths'];
        for (key = j = 0, len1 = typePaths.length; j < len1; key = ++j) {
          typePath = typePaths[key];
          typePaths[key] = pathUtil.resolve(this.config.srcPath, typePath);
        }
      }
      ref1 = ['plugins'];
      // Resolve Plugins paths
      for (k = 0, len2 = ref1.length; k < len2; k++) {
        type = ref1[k];
        typePaths = this.config[type + 'Paths'];
        for (key = m = 0, len3 = typePaths.length; m < len3; key = ++m) {
          typePath = typePaths[key];
          typePaths[key] = pathUtil.resolve(this.config.rootPath, typePath);
        }
      }
      // Bind the error handler, so we don't crash on errors
      process.removeListener('uncaughtException', this.error);
      this.removeListener('error', this.error);
      if (this.config.catchExceptions) {
        process.setMaxListeners(0);
        process.on('uncaughtException', this.error);
        this.on('error', this.error);
      }
      // Prepare the Post Tasks
      postTasks = this.createTaskGroup('setConfig post tasks', {
        next: function(err) {
          return next(err, docpad.config);
        }
      });
      /*
      postTasks.addTask 'lazy depedencnies: encoding', (complete) =>
      	return complete()  unless @config.detectEncoding
      	return lazyRequire 'encoding', {cwd:corePath, stdio:'inherit'}, (err) ->
      		docpad.warn(locale.encodingLoadFailed)  if err
      		return complete()
      */
      postTasks.addTask('load plugins', function(complete) {
        return docpad.loadPlugins(complete);
      });
      postTasks.addTask('extend collections', function(complete) {
        return docpad.extendCollections(complete);
      });
      postTasks.addTask('fetch plugins templateData', function(complete) {
        return docpad.emitSerial('extendTemplateData', {
          templateData: docpad.pluginsTemplateData
        }, complete);
      });
      postTasks.addTask('fire the docpadLoaded event', function(complete) {
        return docpad.emitSerial('docpadLoaded', complete);
      });
      // Fire post tasks
      postTasks.run();
      return this;
    }

    /**
     * Load the various configuration files from the
     * file system. Set the instanceConfig.
     * next(err,config)
     * @private
     * @method load
     * @param {Object} instanceConfig
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.config
     */
    load(instanceConfig, next) {
      var docpad, locale, preTasks;
      // Prepare
      [instanceConfig, next] = extractOptsAndCallback(instanceConfig, next);
      docpad = this;
      locale = this.getLocale();
      instanceConfig || (instanceConfig = {});
      // Reset non persistant configurations
      this.websitePackageConfig = {};
      this.websiteConfig = {};
      this.config = {};
      // Merge in the instance configurations
      this.setInstanceConfig(instanceConfig);
      // Prepare the Load Tasks
      preTasks = this.createTaskGroup('load tasks', {
        next: (err) => {
          if (err) {
            return next(err);
          }
          return this.setConfig(next);
        }
      });
      preTasks.addTask('normalize the userConfigPath', (complete) => {
        return safeps.getHomePath((err, homePath) => {
          var dropboxPath;
          if (err) {
            return complete(err);
          }
          dropboxPath = pathUtil.resolve(homePath, 'Dropbox');
          return safefs.exists(dropboxPath, (dropboxPathExists) => {
            var userConfigDirPath;
            // @TODO: Implement checks here for
            // https://github.com/bevry/docpad/issues/799
            userConfigDirPath = dropboxPathExists ? dropboxPath : homePath;
            this.userConfigPath = pathUtil.resolve(userConfigDirPath, this.userConfigPath);
            return complete();
          });
        });
      });
      preTasks.addTask("load the user's configuration", (complete) => {
        var configPath;
        configPath = this.userConfigPath;
        docpad.log('debug', util.format(locale.loadingUserConfig, configPath));
        return this.loadConfigPath({configPath}, (err, data) => {
          if (err) {
            return complete(err);
          }
          // Apply loaded data
          this.userConfig = this.userConfig;
          // Done
          docpad.log('debug', util.format(locale.loadingUserConfig, configPath));
          return complete();
        });
      });
      preTasks.addTask("load the website's package data", (complete) => {
        var configPath, rootPath;
        rootPath = pathUtil.resolve(this.instanceConfig.rootPath || this.initialConfig.rootPath);
        configPath = pathUtil.resolve(rootPath, this.instanceConfig.packagePath || this.initialConfig.packagePath);
        docpad.log('debug', util.format(locale.loadingWebsitePackageConfig, configPath));
        return this.loadConfigPath({configPath}, (err, data) => {
          if (err) {
            return complete(err);
          }
          data || (data = {});
          // Apply loaded data
          this.websitePackageConfig = data;
          // Done
          docpad.log('debug', util.format(locale.loadedWebsitePackageConfig, configPath));
          return complete();
        });
      });
      preTasks.addTask("read the .env file if it exists", (complete) => {
        var configPath, rootPath;
        rootPath = pathUtil.resolve(this.instanceConfig.rootPath || this.websitePackageConfig.rootPath || this.initialConfig.rootPath);
        configPath = pathUtil.resolve(rootPath, '.env');
        docpad.log('debug', util.format(locale.loadingEnvConfig, configPath));
        return safefs.exists(configPath, function(exists) {
          if (!exists) {
            return complete();
          }
          return require('envfile').parseFile(configPath, function(err, data) {
            var key, value;
            if (err) {
              return complete(err);
            }
            for (key in data) {
              if (!hasProp.call(data, key)) continue;
              value = data[key];
              process.env[key] = value;
            }
            docpad.log('debug', util.format(locale.loadingEnvConfig, configPath));
            return complete();
          });
        });
      });
      preTasks.addTask("load the website's configuration", (complete) => {
        var configPath, configPaths, i, index, len, rootPath;
        docpad.log('debug', util.format(locale.loadingWebsiteConfig));
        rootPath = pathUtil.resolve(this.instanceConfig.rootPath || this.initialConfig.rootPath);
        configPaths = this.instanceConfig.configPaths || this.initialConfig.configPaths;
        for (index = i = 0, len = configPaths.length; i < len; index = ++i) {
          configPath = configPaths[index];
          configPaths[index] = pathUtil.resolve(rootPath, configPath);
        }
        return this.loadConfigPath({configPaths}, (err, data) => {
          if (err) {
            return complete(err);
          }
          data || (data = {});
          // Apply loaded data
          extendr.extend(this.websiteConfig, data);
          // Done
          docpad.log('debug', util.format(locale.loadedWebsiteConfig));
          return complete();
        });
      });
      // Run the load tasks synchronously
      preTasks.run();
      return this;
    }

    // =================================
    // Configuration
    /**
     * Update user configuration with the passed data
     * @method updateUserConfig
     * @param {Object} [data={}]
     * @param {Function} next
     * @param {Error} next.err
     */
    updateUserConfig(data = {}, next) {
      var docpad, userConfigPath;
      // Prepare
      [data, next] = extractOptsAndCallback(data, next);
      docpad = this;
      userConfigPath = this.userConfigPath;
      if (data) {
        // Apply back to our loaded configuration
        // does not apply to @config as we would have to reparse everything
        // and that appears to be an imaginary problem
        extendr.extend(this.userConfig, data);
      }
      // Convert to CSON
      CSON.createCSONString(this.userConfig, function(err, userConfigString) {
        if (err) {
          err.context = "Failed to create the CSON string for the user configuration";
          return next(err);
        }
        // Write it
        return safefs.writeFile(userConfigPath, userConfigString, 'utf8', function(err) {
          // Forward
          return next(err);
        });
      });
      return this;
    }

    /**
     * Load a configuration url.
     * @method loadConfigUrl
     * @param {String} configUrl
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.parsedData
     */
    loadConfigUrl(configUrl, next) {
      var docpad, locale;
      // Prepare
      docpad = this;
      locale = this.getLocale();
      // Log
      docpad.log('debug', util.format(locale.loadingConfigUrl, configUrl));
      // Read the URL
      superAgent.get(configUrl).timeout(30 * 1000).end(function(err, res) {
        if (err) {
          // Check
          return next(err);
        }
        // Read the string using CSON
        return CSON.parseCSONString(res.text, next);
      });
      return this;
    }

    /**
     * Load the configuration from a file path
     * passed as one of the options (opts.configPath) or
     * from DocPad's configPaths
     * @private
     * @method loadConfigPath
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.parsedData
     */
    loadConfigPath(opts, next) {
      var docpad, load, locale;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      locale = this.getLocale();
      // Prepare
      load = function(configPath) {
        if (!configPath) {
          // Check
          return next();
        }
        // Log
        docpad.log('debug', util.format(locale.loadingConfigPath, configPath));
        // Check that it exists
        return safefs.exists(configPath, function(exists) {
          var csonOptions;
          if (!exists) {
            return next();
          }
          // Prepare CSON Options
          csonOptions = {
            cson: true,
            json: true,
            coffeescript: true,
            javascript: true
          };
          // Read the path using CSON
          return CSON.requireFile(configPath, csonOptions, function(err, data) {
            if (err) {
              err.context = util.format(locale.loadingConfigPathFailed, configPath);
              return next(err);
            }
            // Check if the data is a function, if so, then execute it as one
            while (typeChecker.isFunction(data)) {
              try {
                data = data(docpad);
              } catch (error) {
                err = error;
                return next(err);
              }
            }
            if (!typeChecker.isObject(data)) {
              err = new Error(`Loading the configuration ${docpad.inspector(configPath)} returned an invalid result ${docpad.inspector(data)}`);
              if (err) {
                return next(err);
              }
            }
            // Return the data
            return next(null, data);
          });
        });
      };
      // Check
      if (opts.configPath) {
        load(opts.configPath);
      } else {
        this.getConfigPath(opts, function(err, configPath) {
          return load(configPath);
        });
      }
      return this;
    }

    /**
     * Get config paths and check that those
     * paths exist
     * @private
     * @method getConfigPath
     * @param {Object} opts
     * @param {Object} next
     * @param {Error} next.err
     * @param {String} next.path
     */
    getConfigPath(opts, next) {
      var config, docpad, result, tasks;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      config = this.getConfig();
      result = null;
      // Ensure array
      if (opts.configPaths == null) {
        opts.configPaths = config.configPaths;
      }
      if (!typeChecker.isArray(opts.configPaths)) {
        opts.configPaths = [opts.configPaths];
      }
      tasks = this.createTaskGroup('getConfigPath tasks', {
        next: function(err) {
          return next(err, result);
        }
      });
      // Determine our configuration path
      opts.configPaths.forEach(function(configPath) {
        return tasks.addTask(`Checking if [${configPath}] exists`, function(complete) {
          if (result) {
            return complete();
          }
          return safefs.exists(configPath, function(exists) {
            if (exists) {
              result = configPath;
              tasks.clear();
              return complete();
            } else {
              return complete();
            }
          });
        });
      });
      // Run them synchronously
      tasks.run();
      return this;
    }

    /**
     * Extend collecitons. Create DocPad's
     * standard (documents, files
     * layouts) and special (generate, referencesOthers,
     * hasLayout, html, stylesheet) collections. Set blocks
     * @private
     * @method extendCollections
     * @param {Function} next
     * @param {Error} next.err
     */
    extendCollections(next) {
      var database, docpad, docpadConfig, locale, tasks;
      // Prepare
      docpad = this;
      docpadConfig = this.getConfig();
      locale = this.getLocale();
      database = this.getDatabase();
      // Standard Collections
      this.setCollections({
        // Standard Collections
        documents: database.createLiveChildCollection().setQuery('isDocument', {
          render: true,
          write: true
        }).on('add', function(model) {
          return docpad.log('debug', util.format(locale.addingDocument, model.getFilePath()));
        }),
        files: database.createLiveChildCollection().setQuery('isFile', {
          render: false,
          write: true
        }).on('add', function(model) {
          return docpad.log('debug', util.format(locale.addingFile, model.getFilePath()));
        }),
        layouts: database.createLiveChildCollection().setQuery('isLayout', {
          $or: {
            isLayout: true,
            fullPath: {
              $startsWith: docpadConfig.layoutsPaths
            }
          }
        }).on('add', function(model) {
          docpad.log('debug', util.format(locale.addingLayout, model.getFilePath()));
          return model.setDefaults({
            isLayout: true,
            render: false,
            write: false
          });
        }),
        // Special Collections
        generate: database.createLiveChildCollection().setQuery('generate', {
          dynamic: false,
          ignored: false
        }).on('add', function(model) {
          return docpad.log('debug', util.format(locale.addingGenerate, model.getFilePath()));
        }),
        referencesOthers: database.createLiveChildCollection().setQuery('referencesOthers', {
          dynamic: false,
          ignored: false,
          referencesOthers: true
        }).on('add', function(model) {
          return docpad.log('debug', util.format(locale.addingReferencesOthers, model.getFilePath()));
        }),
        hasLayout: database.createLiveChildCollection().setQuery('hasLayout', {
          dynamic: false,
          ignored: false,
          layout: {
            $exists: true
          }
        }).on('add', function(model) {
          return docpad.log('debug', util.format(locale.addingHasLayout, model.getFilePath()));
        }),
        html: database.createLiveChildCollection().setQuery('isHTML', {
          write: true,
          outExtension: 'html'
        }).on('add', function(model) {
          return docpad.log('debug', util.format(locale.addingHtml, model.getFilePath()));
        }),
        stylesheet: database.createLiveChildCollection().setQuery('isStylesheet', {
          write: true,
          outExtension: 'css'
        })
      });
      // Blocks
      this.setBlocks({
        meta: new MetaCollection(),
        scripts: new ScriptsCollection(),
        styles: new StylesCollection()
      });
      // Custom Collections Group
      tasks = this.createTaskGroup("extendCollections tasks", {
        concurrency: 0,
        next: function(err) {
          if (err) {
            docpad.error(err);
          }
          return docpad.emitSerial('extendCollections', next);
        }
      });
      // Cycle through Custom Collections
      eachr(docpadConfig.collections || {}, function(fn, name) {
        var err;
        if (!name || !typeChecker.isString(name)) {
          err = new Error(`Inside your DocPad configuration you have a custom collection with an invalid name of: ${docpad.inspector(name)}`);
          docpad.error(err);
          return;
        }
        if (!fn || !typeChecker.isFunction(fn)) {
          err = new Error(`Inside your DocPad configuration you have a custom collection called ${docpad.inspector(name)} with an invalid method of: ${docpad.inspector(fn)}`);
          docpad.error(err);
          return;
        }
        return tasks.addTask(`creating the custom collection: ${name}`, function(complete) {
          // Init
          return ambi(unbounded.binder.call(fn, docpad), database, function(err, collection) {
            // Check for error
            if (err) {
              docpad.error(err);
              return complete();
            // Check the type of the collection
            } else if (!(collection instanceof QueryCollection)) {
              docpad.warn(util.format(locale.errorInvalidCollection, name));
              return complete();
            }
            if (collection) {
              // Make it a live collection
              collection.live(true);
            }
            // Apply the collection
            docpad.setCollection(name, collection);
            return complete();
          });
        });
      });
      // Run Custom collections
      tasks.run();
      return this;
    }

    /**
     * Reset collections. Perform a complete clean of our collections
     * @private
     * @method resetCollections
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     */
    resetCollections(opts, next) {
      var database, docpad, meta, scripts, styles;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      database = docpad.getDatabase();
      // Make it as if we have never generated before
      docpad.generated = false;
      // Perform a complete clean of our collections
      database.reset([]);
      meta = this.getBlock('meta').reset([]);
      scripts = this.getBlock('scripts').reset([]);
      styles = this.getBlock('styles').reset([]);
      if (docpad.getConfig().poweredByDocPad !== false) {
        // ^ Backbone.js v1.1 changes the return values of these, however we change that in our Element class
        // because if we didn't, all our skeletons would fail

        // Add default block entries
        meta.add(`<meta name="generator" content="DocPad v${docpad.getVersion()}" />`);
      }
      // Reset caches
      this.filesByUrl = {};
      this.filesBySelector = {};
      this.filesByOutPath = {};
      // Chain
      next();
      return this;
    }

    /**
     * Initialise git repo
     * @private
     * @method initGitRepo
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.results
     */
    initGitRepo(opts, next) {
      var config, docpad;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      config = this.getConfig();
      // Extract
      if (opts.cwd == null) {
        opts.cwd = config.rootPath;
      }
      if (opts.output == null) {
        opts.output = this.getDebugging();
      }
      // Forward
      safeps.initGitRepo(opts, next);
      return this;
    }

    /**
     * Init node modules
     * @private
     * @method initNodeModules
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.results
     */
    initNodeModules(opts, next) {
      var config, docpad;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      config = this.getConfig();
      // Extract
      if (opts.cwd == null) {
        opts.cwd = config.rootPath;
      }
      if (opts.output == null) {
        opts.output = docpad.getDebugging();
      }
      if (opts.force == null) {
        opts.force = config.offline ? false : true;
      }
      // ^ @todo this line causes --force to be added, when it shouldn't be
      if (opts.args == null) {
        opts.args = [];
      }
      if (config.force) {
        opts.args.push('--force');
      }
      if (config.offline) {
        opts.args.push('--no-registry');
      }
      if (opts.output) {
        // Log
        docpad.log('info', 'npm install');
      }
      // Forward
      safeps.initNodeModules(opts, next);
      return this;
    }

    /**
     * Fix node package versions
     * Combat to https://github.com/npm/npm/issues/4587#issuecomment-35370453
     * @private
     * @method fixNodePackageVersions
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     */
    fixNodePackageVersions(opts, next) {
      var config, docpad;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      config = this.getConfig();
      // Extract
      if (opts.packagePath == null) {
        opts.packagePath = config.packagePath;
      }
      // Read and replace
      safefs.readFile(opts.packagePath, function(err, buffer) {
        var data;
        data = buffer.toString();
        data = data.replace(/("docpad(?:.*?)": ")\^/g, '$1~');
        return safefs.writeFile(opts.packagePath, data, function(err) {
          return next(err);
        });
      });
      return this;
    }

    /**
     * Install node module. Same as running
     * 'npm install' through the command line
     * @private
     * @method installNodeModule
     * @param {Array} names
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.result
     */
    installNodeModule(names, opts, next) {
      var command, config, docpad;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      config = this.getConfig();
      // Extract
      if (opts.cwd == null) {
        opts.cwd = config.rootPath;
      }
      if (opts.args == null) {
        opts.args = [];
      }
      if (docpad.getDebugging()) {
        if (opts.stdio == null) {
          opts.stdio = 'inherit';
        }
      }
      if (opts.global == null) {
        opts.global = false;
      }
      if (opts.global === true) {
        opts.global = ['--global'];
      }
      if (opts.global && Array.isArray(opts.global) === false) {
        opts.global = [opts.global];
      }
      if (opts.save == null) {
        opts.save = !opts.global;
      }
      if (opts.save === true) {
        opts.save = ['--save'];
      }
      if (opts.save && Array.isArray(opts.save) === false) {
        opts.save = [opts.save];
      }
      // Command
      command = ['npm', 'install'];
      if (!typeChecker.isArray(names)) {
        // Names
        names = names.split(/[,\s]+/);
      }
      names.forEach(function(name) {
        // Check
        if (!name) {
          return;
        }
        if (name.indexOf('@') === -1) {
          // Ensure latest if version isn't specfied
          name += '@latest';
        }
        // Push the name to the commands
        return command.push(name);
      });
      // Arguments
      command.push(...opts.args);
      if (config.force) {
        command.push('--force');
      }
      if (config.offline) {
        command.push('--no-registry');
      }
      if (opts.save) {
        command.push(...opts.save);
      }
      if (opts.global) {
        command.push(...opts.global);
      }
      if (opts.output) {
        // Log
        docpad.log('info', command.join(' '));
      }
      // Forward
      safeps.spawn(command, opts, next);
      return this;
    }

    /**
     * Uninstall node module. Same as running
     * 'npm uninstall' through the command line
     * @private
     * @method uninstallNodeModule
     * @param {Array} names
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.result
     */
    uninstallNodeModule(names, opts, next) {
      var command, config, docpad;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      config = this.getConfig();
      // Extract
      if (opts.cwd == null) {
        opts.cwd = config.rootPath;
      }
      if (opts.output == null) {
        opts.output = docpad.getDebugging();
      }
      if (opts.args == null) {
        opts.args = [];
      }
      if (opts.global == null) {
        opts.global = false;
      }
      if (opts.global === true) {
        opts.global = ['--global'];
      }
      if (opts.global && Array.isArray(opts.global) === false) {
        opts.global = [opts.global];
      }
      if (opts.save == null) {
        opts.save = !opts.global;
      }
      if (opts.save === true) {
        opts.save = ['--save', '--save-dev'];
      }
      if (opts.save && Array.isArray(opts.save) === false) {
        opts.save = [opts.save];
      }
      // Command
      command = ['npm', 'uninstall'];
      if (!typeChecker.isArray(names)) {
        // Names
        names = names.split(/[,\s]+/);
      }
      command.push(...names);
      // Arguments
      command.push(...opts.args);
      if (opts.save) {
        command.push(...opts.save);
      }
      if (opts.global) {
        command.push(...opts.global);
      }
      if (opts.output) {
        // Log
        docpad.log('info', command.join(' '));
      }
      // Forward
      safeps.spawn(command, opts, next);
      return this;
    }

    // =================================
    // Logging
    /**
     * Set the log level
     * @private
     * @method setLogLevel
     * @param {Number} level
     */
    setLogLevel(level) {
      var loggers;
      this.getLogger().setConfig({level});
      if (level === 7) {
        loggers = this.getLoggers();
        if ((loggers.debug != null) === false) {
          loggers.debug = loggers.logger.pipe(require('caterpillar-human').create({
            color: false
          })).pipe(require('fs').createWriteStream(this.debugLogPath));
        }
      }
      return this;
    }

    /**
     * Get the log level
     * @method getLogLevel
     * @return {Number} the log level
     */
    getLogLevel() {
      return this.getConfig().logLevel;
    }

    /**
     * Are we debugging?
     * @method getDebugging
     * @return {Boolean}
     */
    getDebugging() {
      return this.getLogLevel() === 7;
    }

    /**
     * Handle a fatal error
     * @private
     * @method fatal
     * @param {Object} err
     */
    fatal(err) {
      var config, docpad;
      docpad = this;
      config = this.getConfig();
      if (!err) {
        // Check
        return this;
      }
      // Handle
      this.error(err);
      // Even though the error would have already been logged by the above
      // Ensure it is definitely outputted in the case the above fails
      docpadUtil.writeError(err);
      // Destroy DocPad
      this.destroy();
      return this;
    }

    /**
     * Inspect. Converts object to JSON string. Wrapper around nodes util.inspect method.
     * Can't use the inspect namespace as for some silly reason it destroys everything
     * @method inspector
     * @param {Object} obj
     * @param {Object} opts
     * @return {String} JSON string of passed object
     */
    inspector(obj, opts) {
      if (opts == null) {
        opts = {};
      }
      if (opts.colors == null) {
        opts.colors = this.getConfig().color;
      }
      return docpadUtil.inspect(obj, opts);
    }

    /**
     * Log arguments to
     * @property {Object} log
     * @param {Mixed} args...
     */
    log(...args) {
      var logger;
      // Log
      logger = this.getLogger();
      if ((logger != null ? logger.log : void 0) != null) {
        logger.log.apply(logger, args);
      } else {
        // logger doesn't exist, this is probably because it was destroyed
        // in which case handle the minimum case
        // that we want to log everything except debug, unless we are debugging

        // @todo add an option so this becomes an opt-in error case, however for that to happen, there needs to be a way
        // to prevent the following log message after destroy from causing an failure state
        //   debug action runner ➞  runner task for action: destroy > done

        // @todo an alternative approach is to destroy the loggers in a timeout after completion callback has fired

        // @todo one could use caterpillar's log level handling to determine accurate log levels, however
        // that is does not resolve that anything here should not, in theory, occur, as such this is a murky situation

        // as such, considering all the above, output the message, if
        // - message is not a debug message, regardless of log level
        // - message is a debug message, and log level is 7
        // this is done this way, as without proper log level parsing, and with the murky information
        // we can only reasonably assume 'debug' as the first argument is a debug message, anything else could be the message
        if (args[0] !== 'debug' || this.getLogLevel() === 7) {
          console.log(...args);
        }
      }
      return this;
    }

    /**
     * Create an error object
     * @method createError
     * @param {Object} err
     * @param {Object} opts
     * @return {Object} the error
     */
    createError(err, opts) {
      var key, ref, ref1, ref2, ref3, ref4, value;
      // Prepare
      if (opts == null) {
        opts = {};
      }
      if (opts.level == null) {
        opts.level = (ref = err.level) != null ? ref : 'error';
      }
      if (opts.log == null) {
        opts.log = (ref1 = err.log) != null ? ref1 : true;
      }
      if (opts.logged == null) {
        opts.logged = (ref2 = err.logged) != null ? ref2 : false;
      }
      if (opts.notify == null) {
        opts.notify = (ref3 = err.notify) != null ? ref3 : true;
      }
      if (opts.notified == null) {
        opts.notified = (ref4 = err.notified) != null ? ref4 : false;
      }
      if (err.context != null) {
        if (opts.context == null) {
          opts.context = err.context;
        }
      }
      if (!err.stack) {
        // Ensure we have an error object
        err = new Error(err);
      }
      for (key in opts) {
        if (!hasProp.call(opts, key)) continue;
        value = opts[key];
        if (err[key] == null) {
          err[key] = value;
        }
      }
      // Return the error
      return err;
    }

    /**
     * Create an error and log it
     * @method error
     * @param {Object} err
     * @param {Object} [level='err']
     */
    error(err, level = 'err') {
      var docpad;
      // Prepare
      docpad = this;
      // Create the error
      err = this.createError(err, {level});
      // Log the error
      this.logError(err);
      // Notify the error
      this.notifyError(err);
      return this;
    }

    /**
     * Log an error
     * @method logError
     * @param {Object} err
     */
    logError(err) {
      var docpad, locale, message, occured, ref;
      // Prepare
      docpad = this;
      locale = this.getLocale();
      // Log
      if (err && err.log !== false && err.logged !== true) {
        err = this.createError(err, {
          logged: true
        });
        occured = (ref = err.level) === 'warn' || ref === 'warning' ? locale.warnOccured : locale.errorOccured;
        message = err.context ? err.context + locale.errorFollows : occured;
        message += '\n\n' + err.stack.toString().trim();
        message += '\n\n' + locale.errorSubmission;
        docpad.log(err.level, message);
      }
      return this;
    }

    /**
     * Notify error
     * @private
     * @method notifyError
     * @param {Object} err
     */
    notifyError(err) {
      var docpad, locale, occured, ref;
      // Prepare
      docpad = this;
      locale = this.getLocale();
      // Check
      if (err.notify !== false && err.notified !== true) {
        err.notified = true;
        occured = (ref = err.level) === 'warn' || ref === 'warning' ? locale.warnOccured : locale.errorOccured;
        docpad.notify(err.message, {
          title: occured
        });
      }
      return this;
    }

    /**
     * Log an error of level 'warn'
     * @method warn
     * @param {String} message
     * @param {Object} err
     * @return {Object} description
     */
    warn(message, err) {
      // Handle
      if (err) {
        err.context = message;
        err.level = 'warn';
        this.error(err);
      } else {
        err = message instanceof Error ? message : new Error(message);
        err.level = 'warn';
        this.error(err);
      }
      return this;
    }

    /**
     * Send a notify event to plugins (like growl)
     * @method notify
     * @param {String} message
     * @param {Object} [opts={}]
     */
    notify(message, opts = {}) {
      var docpad;
      // Prepare
      docpad = this;
      // Emit
      docpad.emitSerial('notify', {message, opts}, function(err) {
        if (err) {
          return docpad.error(err);
        }
      });
      return this;
    }

    /**
     * Check Request
     * @private
     * @method checkRequest
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.res
     */
    checkRequest(next) {
      if (next == null) {
        next = this.error.bind(this);
      }
      return function(err, res) {
        var ref, ref1;
        if (err) {
          // Check
          return next(err, res);
        }
        // Check
        if (((ref = res.body) != null ? ref.success : void 0) === false || ((ref1 = res.body) != null ? ref1.error : void 0)) {
          err = new Error(res.body.error || 'unknown request error'); // @TODO localise this
          return next(err, res);
        }
        // Success
        return next(null, res);
      };
    }

    // =================================
    // Models and Collections

    // ---------------------------------
    // b/c compat functions
    /**
     * Create file model. Calls
     * {{#crossLink "DocPad/createModel:method"}}{{/crossLink}}
     * with the 'file' modelType.
     * @method createFile
     * @param {Object} [attrs={}]
     * @param {Object} [opts={}]
     * @return {Object} FileModel
     */
    createFile(attrs = {}, opts = {}) {
      opts.modelType = 'file';
      return this.createModel(attrs, opts);
    }

    /**
     * Create document model. Calls
     * {{#crossLink "DocPad/createModel:method"}}{{/crossLink}}
     * with the 'document' modelType.
     * @method createDocument
     * @param {Object} [attrs={}]
     * @param {Object} [opts={}]
     * @return {Object} DocumentModel
     */
    createDocument(attrs = {}, opts = {}) {
      opts.modelType = 'document';
      return this.createModel(attrs, opts);
    }

    /**
     * Parse the files directory and
     * return a files collection to
     * the passed callback
     * @method parseFileDirectory
     * @param {Object} [opts={}]
     * @param {Function} next callback
     * @param {Error} next.err
     * @param {Object} next.files files collection
     */
    parseFileDirectory(opts, next) {
      [opts, next] = extractOptsAndCallback(opts, next);
      if (opts.modelType == null) {
        opts.modelType = 'file';
      }
      if (opts.collection == null) {
        opts.collection = this.getDatabase();
      }
      return this.parseDirectory(opts, next);
    }

    /**
     * Parse the documents directory and
     * return a documents collection to
     * the passed callback.
     *
     * The partials plugin (https://github.com/docpad/docpad-plugin-partials)
     * uses this method to load a collection of
     * files from the partials directory.
     *
     * 	docpad.parseDocumentDirectory({path: config.partialsPath}, next)
     *
     * @method parseDocumentDirectory
     * @param {Object} [opts={}]
     * @param {String} [opts.modelType='document']
     * @param {Object} [opts.collection=docpad.database]
     * @param {Object} [opts.path]
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.files files collection of documents
     */
    parseDocumentDirectory(opts, next) {
      [opts, next] = extractOptsAndCallback(opts, next);
      if (opts.modelType == null) {
        opts.modelType = 'document';
      }
      if (opts.collection == null) {
        opts.collection = this.getDatabase();
      }
      return this.parseDirectory(opts, next);
    }

    // ---------------------------------
    // Standard functions
    /**
     * Attach events to a document model.
     * @private
     * @method attachModelEvents
     * @param {Object} model
     */
    attachModelEvents(model) {
      var docpad;
      // Prepare
      docpad = this;
      // Only attach events if we haven't already done so
      if (model.attachedDocumentEvents !== true) {
        model.attachedDocumentEvents = true;
        // Attach document events
        if (model.type === 'document') {
          // Clone
          model.on('clone', function(clonedModel) {
            return docpad.attachModelEvents(clonedModel);
          });
          // Render
          model.on('render', function(...args) {
            return docpad.emitSerial('render', ...args);
          });
          // Render document
          model.on('renderDocument', function(...args) {
            return docpad.emitSerial('renderDocument', ...args);
          });
          // Fetch a layout
          model.on('getLayout', function(opts, next) {
            var layout;
            opts.collection = docpad.getCollection('layouts');
            layout = docpad.getFileBySelector(opts.selector, opts);
            return next(null, {layout});
          });
        }
        // Remove
        //model.on 'remove', (file) ->
        //	docpad.getDatabase().remove(file)
        // ^ Commented out as for some reason this stops layouts from working

        // Error
        model.on('error', function(...args) {
          return docpad.error(...args);
        });
        // Log
        model.on('log', function(...args) {
          var ref, ref1;
          if (args.length === 2) {
            if ((ref = args[0]) === 'err' || ref === 'error') {
              docpad.error(args[1]);
              return;
            }
            if ((ref1 = args[0]) === 'warn' || ref1 === 'warning') {
              docpad.warn(args[1]);
              return;
            }
          }
          return docpad.log(...args);
        });
      }
      return this;
    }

    /**
     * Add supplied model to the DocPad database. If the passed
     * model definition is a plain object of properties, a new
     * model will be created prior to adding to the database.
     * Calls {{#crossLink "DocPad/createModel:method"}}{{/crossLink}}
     * before adding the model to the database.
     *
     *	# Override the stat's mtime to now
     *	# This is because renames will not update the mtime
     *	fileCurrentStat?.mtime = new Date()
     *
     *	# Create the file object
     *	file = docpad.addModel({fullPath:filePath, stat:fileCurrentStat})
     *
     * @method addModel
     * @param {Object} model either a plain object defining the required properties, in particular
     * the file path or an actual model object
     * @param {Object} opts
     * @return {Object} the model
     */
    addModel(model, opts) {
      model = this.createModel(model, opts);
      this.getDatabase().add(model);
      return model;
    }

    /**
     * Add the supplied collection of models to the DocPad database.
     * Calls {{#crossLink "DocPad/createModels:method"}}{{/crossLink}}
     * before adding the models to the database.
     *
     * 	databaseData = JSON.parse data.toString()
     *	models = docpad.addModels(databaseData.models)
     *
     * @method addModels
     * @param {Object} models DocPad collection of models
     * @param {Object} opts
     * @return {Object} the models
     */
    addModels(models, opts) {
      models = this.createModels(models, opts);
      this.getDatabase().add(models);
      return models;
    }

    /**
     * Create a collection of models from the supplied collection
     * ensuring that the collection is suitable for adding to the
     * DocPad database. The method calls {{#crossLink "DocPad/createModel"}}{{/crossLink}}
     * for each model in the models array.
     * @private
     * @method createModels
     * @param {Object} models DocPad collection of models
     * @param {Object} opts
     * @return {Object} the models
     */
    createModels(models, opts) {
      var i, len, model, results;
      results = [];
      for (i = 0, len = models.length; i < len; i++) {
        model = models[i];
        results.push(this.createModel(model, opts));
      }
      return results;
    }

    // return the for loop results
    /**
     * Creates either a file or document model.
     * The model type to be created can be passed
     * as an opts property, if not, the method will
     * attempt to determing the model type by checking
     * if the file is in one of the documents or
     * layout paths.
     *
     * Ensures a duplicate model is not created
     * and all required attributes are present and
     * events attached.
     *
     * Generally it is not necessary for an application
     * to manually create a model via creatModel as DocPad
     * will handle this process when watching a project's
     * file and document directories. However, it is possible
     * that a plugin might have a requirement to do so.
     *
     * 	model = @docpad.createModel({fullPath:fullPath})
     *   model.load()
     *   @docpad.getDatabase().add(model)
     *
     * @method createModel
     * @param {Object} [attrs={}]
     * @param {String} attrs.fullPath the full path to the file
     * @param {Object} [opts={}]
     * @param {String} opts.modelType either 'file' or 'document'
     * @return {Object} the file or document model
     */
    createModel(attrs = {}, opts = {}) {
      var config, database, dirPath, docpad, fileFullPath, i, j, len, len1, model, ref, ref1, result;
      // Check
      if (attrs instanceof FileModel) {
        return attrs;
      }
      // Prepare
      docpad = this;
      config = this.getConfig();
      database = this.getDatabase();
      fileFullPath = attrs.fullPath || null;
      // Find or create
      // This functionality use to be inside ensureModel
      // But that caused duplicates in some instances
      // So now we will always check
      if (attrs.fullPath) {
        result = database.findOne({
          fullPath: attrs.fullPath
        });
        if (result) {
          return result;
        }
      }
      // -----------------------------
      // Try and determine the model type

      // If the type hasn't been specified try and detemrine it based on the full path
      if (fileFullPath) {
        // Check if we have a document or layout
        if (!opts.modelType) {
          ref = config.documentsPaths.concat(config.layoutsPaths);
          for (i = 0, len = ref.length; i < len; i++) {
            dirPath = ref[i];
            if (fileFullPath.indexOf(dirPath) === 0) {
              attrs.relativePath || (attrs.relativePath = fileFullPath.replace(dirPath, '').replace(/^[\/\\]/, ''));
              opts.modelType = 'document';
              break;
            }
          }
        }
        // Check if we have a file
        if (!opts.modelType) {
          ref1 = config.filesPaths;
          for (j = 0, len1 = ref1.length; j < len1; j++) {
            dirPath = ref1[j];
            if (fileFullPath.indexOf(dirPath) === 0) {
              attrs.relativePath || (attrs.relativePath = fileFullPath.replace(dirPath, '').replace(/^[\/\\]/, ''));
              opts.modelType = 'file';
              break;
            }
          }
        }
      }
      // -----------------------------
      // Create the appropriate emodel

      // Extend the opts with things we need
      opts = extendr.extend({
        detectEncoding: config.detectEncoding,
        rootOutDirPath: config.outPath,
        locale: this.getLocale(),
        createTaskGroup: this.createTaskGroup // @TODO this a bit dodgy, but works well enough
      }, opts);
      if (opts.modelType === 'file') {
        // Create a file model
        model = new FileModel(attrs, opts);
      } else {
        // Create document model
        model = new DocumentModel(attrs, opts);
      }
      // -----------------------------
      // Finish up

      // Attach Events
      this.attachModelEvents(model);
      // Return
      return model;
    }

    /**
     * Parse a directory and return a
     * files collection
     * @method parseDirectory
     * @param {Object} [opts={}]
     * @param {Object} next
     * @param {Error} next.err
     * @param {Object} next.files files collection
     */
    parseDirectory(opts, next) {
      var createFunction, docpad, files, locale, path;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      locale = this.getLocale();
      // Extract
      ({path, createFunction} = opts);
      if (createFunction == null) {
        createFunction = this.createModel;
      }
      files = opts.collection || new FilesCollection();
      // Check if the directory exists
      safefs.exists(path, function(exists) {
        var tasks;
        // Check
        if (!exists) {
          // Log
          docpad.log('debug', util.format(locale.renderDirectoryNonexistant, path));
          // Forward
          return next();
        }
        // Log
        docpad.log('debug', util.format(locale.renderDirectoryParsing, path));
        // Tasks
        tasks = new TaskGroup('parse directory').setConfig({
          concurrency: 0
        }).done(function(err) {
          if (err) {
            // Check
            return next(err);
          }
          // Log
          docpad.log('debug', util.format(locale.renderDirectoryParsed, path));
          // Forward
          return next(null, files);
        });
        // Files
        return docpad.scandir({
          // Path
          path: path,
          // File Action
          fileAction: function(fileFullPath, fileRelativePath, filename, fileStat) {
            var data, file;
            // Prepare
            data = {
              fullPath: fileFullPath,
              relativePath: fileRelativePath,
              stat: fileStat
            };
            // Create file
            file = createFunction.call(docpad, data, opts);
            // Create a task to load the file
            tasks.addTask(`load the file ${fileRelativePath}`, function(complete) {
              // Update the file's stat
              // To ensure changes files are handled correctly in generation
              return file.action('load', function(err) {
                if (err) {
                  // Error?
                  return complete(err);
                }
                // Add the file to the collection
                files.add(file);
                // Next
                return complete();
              });
            });
          },
          // Next
          // Return
          next: function(err) {
            if (err) {
              return next(err);
            }
            return tasks.run();
          }
        });
      });
      return this;
    }

    // =================================
    // Plugins
    /**
     * Get a plugin by it's name
     * @method getPlugin
     * @param {Object} pluginName
     * @return {Object} a DocPad plugin
     */
    getPlugin(pluginName) {
      return this.loadedPlugins[pluginName];
    }

    /**
     * Check if we have any plugins
     * @method hasPlugins
     * @return {Boolean}
     */
    hasPlugins() {
      return typeChecker.isEmptyObject(this.loadedPlugins) === false;
    }

    /**
     * Destructor. Destroy plugins
     * @private
     * @method destroyPlugins
     */
    destroyPlugins() {
      var name, plugin, ref;
      ref = this.loadedPlugins;
      for (name in ref) {
        if (!hasProp.call(ref, name)) continue;
        plugin = ref[name];
        plugin.destroy();
        this.loadedPlugins[name] = null;
      }
      return this;
    }

    /**
     * Load plugins from the file system
     * next(err)
     * @private
     * @method loadPlugins
     * @param {Function} next
     * @param {Error} next.err
     */
    loadPlugins(next) {
      var docpad, locale, snore, tasks;
      // Prepare
      docpad = this;
      locale = this.getLocale();
      // Snore
      this.slowPlugins = {};
      snore = balUtil.createSnore(function() {
        return docpad.log('notice', util.format(locale.pluginsSlow, Object.keys(docpad.slowPlugins).join(', ')));
      });
      // Async
      tasks = this.createTaskGroup("loadPlugins tasks", {
        concurrency: 0,
        next: function(err) {
          docpad.slowPlugins = {};
          snore.clear();
          return next(err);
        }
      });
      // Load website plugins
      (this.config.pluginsPaths || []).forEach(function(pluginsPath) {
        return tasks.addTask(`load the website's plugins at: ${pluginsPath}`, function(complete) {
          return safefs.exists(pluginsPath, function(exists) {
            if (!exists) {
              return complete();
            }
            return docpad.loadPluginsIn(pluginsPath, complete);
          });
        });
      });
      // Load specific plugins
      (this.config.pluginPaths || []).forEach(function(pluginPath) {
        return tasks.addTask(`load custom plugins at: ${pluginPath}`, function(complete) {
          return safefs.exists(pluginPath, function(exists) {
            if (!exists) {
              return complete();
            }
            return docpad.loadPlugin(pluginPath, complete);
          });
        });
      });
      // Execute the loading asynchronously
      tasks.run();
      return this;
    }

    /**
     * Checks if a plugin was loaded succesfully.
     * @method loadedPlugin
     * @param {String} pluginName
     * @param {Function} next
     * @param {Error} next.err
     * @param {Boolean} next.loaded
     */
    loadedPlugin(pluginName, next) {
      var docpad, loaded;
      // Prepare
      docpad = this;
      // Check
      loaded = docpad.loadedPlugins[pluginName] != null;
      next(null, loaded);
      return this;
    }

    /**
     * Load a plugin from its full file path
     * _next(err)
     * @private
     * @method loadPlugin
     * @param {String} fileFullPath
     * @param {Function} _next
     * @param {Error} _next.err
     * @return {Object} description
     */
    loadPlugin(fileFullPath, _next) {
      var config, docpad, enabled, loader, locale, next, pluginName;
      // Prepare
      docpad = this;
      config = this.getConfig();
      locale = this.getLocale();
      next = function(err) {
        // Remove from slow plugins
        delete docpad.slowPlugins[pluginName];
        // Forward
        return _next(err);
      };
      // Prepare variables
      loader = new PluginLoader({
        dirPath: fileFullPath,
        docpad: this,
        BasePlugin: BasePlugin
      });
      pluginName = loader.pluginName;
      enabled = (config.enableUnlistedPlugins && (config.enabledPlugins[pluginName] != null) === false) || config.enabledPlugins[pluginName] === true;
      // If we've already been loaded, then exit early as there is no use for us to load again
      if (docpad.loadedPlugins[pluginName] != null) {
        // However we probably want to reload the configuration as perhaps the user or environment configuration has changed
        docpad.loadedPlugins[pluginName].setConfig();
        // Complete
        return _next();
      }
      // Add to loading stores
      docpad.slowPlugins[pluginName] = true;
      // Check
      if (!enabled) {
        // Skip
        docpad.log('debug', util.format(locale.pluginSkipped, pluginName));
        return next();
      } else {
        // Load
        docpad.log('debug', util.format(locale.pluginLoading, pluginName));
        // Check existance
        loader.exists(function(err, exists) {
          if (err || !exists) {
            // Error or doesn't exist?
            return next(err);
          }
          // Check support
          return loader.unsupported(function(err, unsupported) {
            if (err) {
              // Error?
              return next(err);
            }
            // Unsupported?
            if (unsupported) {
              // Version?
              if ((unsupported === 'version-docpad' || unsupported === 'version-plugin') && config.skipUnsupportedPlugins === false) {
                docpad.log('warn', util.format(locale.pluginContinued, pluginName));
              } else {
                // Type?
                if (unsupported === 'type') {
                  docpad.log('debug', util.format(locale.pluginSkippedDueTo, pluginName, unsupported));
                } else {
                  // Something else?
                  docpad.log('warn', util.format(locale.pluginSkippedDueTo, pluginName, unsupported));
                }
                return next();
              }
            }
            // Load the class
            return loader.load(function(err) {
              if (err) {
                return next(err);
              }
              // Create an instance
              return loader.create({}, function(err, pluginInstance) {
                if (err) {
                  return next(err);
                }
                // Add to plugin stores
                docpad.loadedPlugins[loader.pluginName] = pluginInstance;
                // Log completion
                docpad.log('debug', util.format(locale.pluginLoaded, pluginName));
                // Forward
                return next();
              });
            });
          });
        });
      }
      return this;
    }

    /**
     * Load plugins from a directory path
     * @private
     * @method loadPluginsIn
     * @param {String} pluginsPath
     * @param {Function} next
     * @param {Error} next.err
     */
    loadPluginsIn(pluginsPath, next) {
      var docpad, locale, tasks;
      // Prepare
      docpad = this;
      locale = this.getLocale();
      // Tasks
      tasks = new TaskGroup('load plugins').setConfig({
        concurrency: 0
      }).done(function(err) {
        docpad.log('debug', util.format(locale.pluginsLoadedFor, pluginsPath));
        return next(err);
      });
      // Load Plugins
      docpad.log('debug', util.format(locale.pluginsLoadingFor, pluginsPath));
      docpad.scandir({
        // Path
        path: pluginsPath,
        // Skip files
        fileAction: false,
        // Handle directories
        dirAction: function(fileFullPath, fileRelativePath, pluginName) {
          // Prepare
          pluginName = pathUtil.basename(fileFullPath);
          // Add the task to load the oplugin to the queue
          return tasks.addTask(`load the plugin ${pluginName}`, function(complete) {
            // Otherwise, it is a plugin directory, so load the plugin
            return docpad.loadPlugin(fileFullPath, function(err) {
              // Warn about the plugin load error if there is one
              if (err) {
                docpad.warn(util.format(locale.pluginFailedToLoad, pluginName, fileFullPath), err);
              }
              // All done and don't recurse into this directory
              return complete();
            });
          });
        },
        // Next
        next: function(err) {
          if (err) {
            return next(err);
          }
          return tasks.run();
        }
      });
      return this;
    }

    // =================================
    // Utilities

    // ---------------------------------
    // Utilities: Misc
    /**
     * Compare current DocPad version to the latest
     * and print out the result to the console.
     * Used at startup.
     * @private
     * @method compareVersion
     */
    compareVersion() {
      var config, docpad, locale;
      // Prepare
      docpad = this;
      config = this.getConfig();
      locale = this.getLocale();
      if (config.offline || !config.checkVersion) {
        // Check
        return this;
      }
      // Check
      balUtil.packageCompare({
        local: this.packagePath,
        remote: config.helperUrl + 'latest',
        newVersionCallback: function(details) {
          var currentVersion, isLocalInstallation, latestVersion, message, messageFilled, upgradeUrl;
          isLocalInstallation = docpadUtil.isLocalDocPadExecutable();
          message = (isLocalInstallation ? locale.versionOutdatedLocal : locale.versionOutdatedGlobal);
          currentVersion = 'v' + details.local.version;
          latestVersion = 'v' + details.remote.version;
          upgradeUrl = details.local.upgradeUrl || details.remote.installUrl || details.remote.homepage;
          messageFilled = util.format(message, currentVersion, latestVersion, upgradeUrl);
          docpad.notify(latestVersion, {
            title: locale.versionOutdatedNotification
          });
          return docpad.log('notice', messageFilled);
        }
      });
      return this;
    }

    // ---------------------------------
    // Utilities: Exchange
    /**
     * Get DocPad's exchange data
     * Requires internet access
     * next(err,exchange)
     * @private
     * @method getExchange
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.exchange docpad.exchange
     */
    getExchange(next) {
      var config, docpad, exchangeUrl, locale;
      // Prepare
      docpad = this;
      config = this.getConfig();
      locale = this.getLocale();
      if (typeChecker.isEmptyObject(docpad.exchange) === false) {
        // Check if it is stored locally
        return next(null, docpad.exchange);
      }
      if (config.offline) {
        // Offline?
        return next(null, null);
      }
      // Log
      docpad.log('info', locale.exchangeUpdate + ' ' + locale.pleaseWait);
      // Otherwise fetch it from the exchangeUrl
      exchangeUrl = config.helperUrl + '?method=exchange&version=' + this.version;
      docpad.loadConfigUrl(exchangeUrl, function(err, parsedData) {
        // Check
        if (err) {
          locale = docpad.getLocale();
          docpad.warn(locale.exchangeError, err);
          return next();
        }
        // Log
        docpad.log('info', locale.exchangeUpdated);
        // Success
        docpad.exchange = parsedData;
        return next(null, parsedData);
      });
      return this;
    }

    // ---------------------------------
    // Utilities: Files
    /**
     * Contextualize files.
     * Contextualizing is the process of adding layouts and
     * awareness of other documents to our document. The
     * contextualizeBefore and contextualizeAfter events
     * are emitted here.
     * @private
     * @method contextualizeFiles
     * @param {Object} [opts={}]
     * @param {Function} next
     * @param {Error} next.err
     */
    contextualizeFiles(opts, next) {
      var collection, config, docpad, locale, slowFilesObject, slowFilesTimer, templateData;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      ({collection, templateData} = opts);
      docpad = this;
      config = this.getConfig();
      locale = this.getLocale();
      slowFilesObject = {};
      slowFilesTimer = null;
      // Log
      docpad.log('debug', util.format(locale.contextualizingFiles, collection.length));
      // Start contextualizing
      docpad.emitSerial('contextualizeBefore', {collection, templateData}, function(err) {
        var tasks;
        if (err) {
          // Prepare
          return next(err);
        }
        // Completion callback
        tasks = docpad.createTaskGroup("contextualizeFiles", {
          concurrency: 0,
          next: function(err) {
            // Kill the timer
            clearInterval(slowFilesTimer);
            slowFilesTimer = null;
            if (err) {
              // Check
              return next(err);
            }
            // After
            return docpad.emitSerial('contextualizeAfter', {collection}, function(err) {
              if (err) {
                // Check
                return next(err);
              }
              // Log
              docpad.log('debug', util.format(locale.contextualizedFiles, collection.length));
              // Forward
              return next();
            });
          }
        });
        // Add contextualize tasks
        collection.forEach(function(file, index) {
          var filePath;
          filePath = file.getFilePath();
          slowFilesObject[file.id] = file.get('relativePath') || file.id;
          return tasks.addTask(`conextualizing: ${filePath}`, function(complete) {
            return file.action('contextualize', function(err) {
              delete slowFilesObject[file.id];
              return complete(err);
            });
          });
        });
        // Setup the timer
        slowFilesTimer = setInterval(function() {
          var key, slowFilesArray, value;
          slowFilesArray = (function() {
            var results;
            results = [];
            for (key in slowFilesObject) {
              if (!hasProp.call(slowFilesObject, key)) continue;
              value = slowFilesObject[key];
              results.push(value || key);
            }
            return results;
          })();
          return docpad.log('info', util.format(locale.slowFiles, 'contextualizeFiles') + ' \n' + slowFilesArray.join('\n'));
        }, config.slowFilesDelay);
        // Run tasks
        return tasks.run();
      });
      return this;
    }

    /**
     * Render the DocPad project's files.
     * The renderCollectionBefore, renderCollectionAfter,
     * renderBefore, renderAfter events are all emitted here.
     * @private
     * @method renderFiles
     * @param {Object} [opts={}]
     * @param {Function} next
     * @param {Error} next.err
     */
    renderFiles(opts, next) {
      var collection, config, docpad, locale, renderCollection, renderFile, renderPasses, slowFilesObject, slowFilesTimer, templateData;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      ({collection, templateData, renderPasses} = opts);
      docpad = this;
      config = this.getConfig();
      locale = this.getLocale();
      slowFilesObject = {};
      slowFilesTimer = null;
      // Log
      docpad.log('debug', util.format(locale.renderingFiles, collection.length));
      // Render File
      // next(null, outContent, file)
      renderFile = function(file, next) {
        // Render
        if (file.get('render') === false || !file.get('relativePath')) {
          file.attributes.rtime = new Date();
          next(null, file.getOutContent(), file);
        } else {
          file.action('render', {templateData}, next);
        }
        // Return
        return file;
      };
      // Render Collection
      renderCollection = function(collectionToRender, {renderPass}, next) {
        // Plugin Event
        return docpad.emitSerial('renderCollectionBefore', {
          collection: collectionToRender,
          renderPass
        }, function(err) {
          var subTasks;
          if (err) {
            // Prepare
            return next(err);
          }
          subTasks = docpad.createTaskGroup(`renderFiles: Pass ${renderPass}]: renderCollection: ${collectionToRender.options.name}`, {
            concurrency: 0,
            next: function(err) {
              if (err) {
                // Prepare
                return next(err);
              }
              // Plugin Event
              return docpad.emitSerial('renderCollectionAfter', {
                collection: collectionToRender,
                renderPass
              }, next);
            }
          });
          // Cycle
          collectionToRender.forEach(function(file) {
            var filePath;
            filePath = file.getFilePath();
            slowFilesObject[file.id] = file.get('relativePath');
            return subTasks.addTask(`rendering: ${filePath}`, function(complete) {
              return renderFile(file, function(err) {
                delete slowFilesObject[file.id] || file.id;
                return complete(err);
              });
            });
          });
          // Return
          subTasks.run();
          return collectionToRender;
        });
      };
      // Plugin Event
      docpad.emitSerial('renderBefore', {collection, templateData}, function(err) {
        var initialCollection, subsequentCollection, tasks;
        if (err) {
          // Prepare
          return next(err);
        }
        // Async
        tasks = docpad.createTaskGroup("renderFiles: renderCollection: renderBefore", {
          next: function(err) {
            // Kill the timer
            clearInterval(slowFilesTimer);
            slowFilesTimer = null;
            if (err) {
              // Check
              return next(err);
            }
            // After
            return docpad.emitSerial('renderAfter', {collection}, function(err) {
              if (err) {
                // Check
                return next(err);
              }
              // Log
              docpad.log('debug', util.format(locale.renderedFiles, collection.length));
              // Forward
              return next();
            });
          }
        });
        // Queue the initial render
        initialCollection = collection.findAll({
          'referencesOthers': false
        });
        subsequentCollection = null;
        tasks.addTask("rendering the initial collection", function(complete) {
          return renderCollection(initialCollection, {
            renderPass: 1
          }, function(err) {
            if (err) {
              return complete(err);
            }
            subsequentCollection = collection.findAll({
              'referencesOthers': true
            });
            return renderCollection(subsequentCollection, {
              renderPass: 2
            }, complete);
          });
        });
        // Queue the subsequent renders
        if (renderPasses > 1) {
          (function() {
            var results = [];
            for (var i = 3; 3 <= renderPasses ? i <= renderPasses : i >= renderPasses; 3 <= renderPasses ? i++ : i--){ results.push(i); }
            return results;
          }).apply(this).forEach(function(renderPass) {
            return tasks.addTask(`rendering the subsequent collection index ${renderPass}`, function(complete) {
              return renderCollection(subsequentCollection, {renderPass}, complete);
            });
          });
        }
        // Setup the timer
        slowFilesTimer = setInterval(function() {
          var key, slowFilesArray, value;
          slowFilesArray = (function() {
            var results;
            results = [];
            for (key in slowFilesObject) {
              if (!hasProp.call(slowFilesObject, key)) continue;
              value = slowFilesObject[key];
              results.push(value || key);
            }
            return results;
          })();
          return docpad.log('info', util.format(locale.slowFiles, 'renderFiles') + ' \n' + slowFilesArray.join('\n'));
        }, config.slowFilesDelay);
        // Run tasks
        return tasks.run();
      });
      return this;
    }

    /**
     * Write rendered files to the DocPad out directory.
     * The writeBefore and writeAfter events are emitted here.
     * @private
     * @method writeFiles
     * @param {Object} [opts={}]
     * @param {Function} next
     * @param {Error} next.err
     */
    writeFiles(opts, next) {
      var collection, config, docpad, locale, slowFilesObject, slowFilesTimer, templateData;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      ({collection, templateData} = opts);
      docpad = this;
      config = this.getConfig();
      locale = this.getLocale();
      slowFilesObject = {};
      slowFilesTimer = null;
      // Log
      docpad.log('debug', util.format(locale.writingFiles, collection.length));
      // Plugin Event
      docpad.emitSerial('writeBefore', {collection, templateData}, function(err) {
        var tasks;
        if (err) {
          // Prepare
          return next(err);
        }
        // Completion callback
        tasks = docpad.createTaskGroup("writeFiles", {
          concurrency: 0,
          next: function(err) {
            // Kill the timer
            clearInterval(slowFilesTimer);
            slowFilesTimer = null;
            if (err) {
              // Check
              return next(err);
            }
            // After
            return docpad.emitSerial('writeAfter', {collection}, function(err) {
              if (err) {
                // Check
                return next(err);
              }
              // docpad.log 'debug', util.format(locale.wroteFiles, collection.length)
              return next();
            });
          }
        });
        // Add write tasks
        collection.forEach(function(file, index) {
          var filePath;
          filePath = file.getFilePath();
          return tasks.addTask(`writing the file: ${filePath}`, function(complete) {
            var fileTasks;
            // Prepare
            slowFilesObject[file.id] = file.get('relativePath');
            // Create sub tasks
            fileTasks = docpad.createTaskGroup(`tasks for file write: ${filePath}`, {
              concurrency: 0,
              next: function(err) {
                delete slowFilesObject[file.id];
                return complete(err);
              }
            });
            // Write out
            if (file.get('write') !== false && file.get('dynamic') !== true && file.get('outPath')) {
              fileTasks.addTask("write out", function(complete) {
                return file.action('write', complete);
              });
            }
            // Write source
            if (file.get('writeSource') === true && file.get('fullPath')) {
              fileTasks.addTask("write source", function(complete) {
                return file.action('writeSource', complete);
              });
            }
            // Run sub tasks
            return fileTasks.run();
          });
        });
        // Setup the timer
        slowFilesTimer = setInterval(function() {
          var key, slowFilesArray, value;
          slowFilesArray = (function() {
            var results;
            results = [];
            for (key in slowFilesObject) {
              if (!hasProp.call(slowFilesObject, key)) continue;
              value = slowFilesObject[key];
              results.push(value || key);
            }
            return results;
          })();
          return docpad.log('info', util.format(locale.slowFiles, 'writeFiles') + ' \n' + slowFilesArray.join('\n'));
        }, config.slowFilesDelay);
        // Run tasks
        return tasks.run();
      });
      return this;
    }

    /**
     * Destructor. Destroy the regeneration timer.
     * @private
     * @method destroyRegenerateTimer
     */
    destroyRegenerateTimer() {
      var docpad;
      // Prepare
      docpad = this;
      // Clear Regenerate Timer
      if (docpad.regenerateTimer) {
        clearTimeout(docpad.regenerateTimer);
        docpad.regenerateTimer = null;
      }
      return this;
    }

    /**
     * Create the regeneration timer
     * @private
     * @method createRegenerateTimer
     */
    createRegenerateTimer() {
      var config, docpad, locale;
      // Prepare
      docpad = this;
      locale = docpad.getLocale();
      config = docpad.getConfig();
      // Create Regenerate Timer
      if (config.regenerateEvery) {
        docpad.regenerateTimer = setTimeout(function() {
          docpad.log('info', locale.renderInterval);
          return docpad.action('generate', config.regenerateEveryOptions);
        }, config.regenerateEvery);
      }
      return this;
    }

    /**
     * Set off DocPad's generation process.
     * The generated, populateCollectionsBefore, populateCollections, populateCollections
     * generateBefore and generateAfter events are emitted here
     * @method generate
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     */
    generate(opts, next) {
      var addGroup, addTask, config, database, docpad, lastGenerateStarted, locale, ref, tasks;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      config = docpad.getConfig();
      locale = docpad.getLocale();
      database = docpad.getDatabase();
      if (((ref = opts.collection) != null ? ref.length : void 0) === 0) {
        // Check
        return next();
      }
      // Update generating flag
      lastGenerateStarted = docpad.generateStarted;
      docpad.generateStarted = new Date();
      docpad.generateEnded = null;
      docpad.generating = true;
      if (database.models.length) {
        // Update the cached database
        docpad.databaseTempCache = new FilesCollection(database.models);
      }
      // Grab the template data we will use for rendering
      opts.templateData = docpad.getTemplateData(opts.templateData || {});
      // How many render passes will we require?
      // Can be over-written by API calls
      opts.renderPasses || (opts.renderPasses = config.renderPasses);
      // Destroy Regenerate Timer
      docpad.destroyRegenerateTimer();
      if (!docpad.hasPlugins()) {
        // Check plugin count
        docpad.log('notice', locale.renderNoPlugins);
      }
      // Log
      docpad.log('info', locale.renderGenerating);
      docpad.notify((new Date()).toLocaleTimeString(), {
        title: locale.renderGeneratingNotification
      });
      // Tasks
      tasks = this.createTaskGroup("generate tasks").done(function(err) {
        var howMany, ref1, seconds;
        // Update generating flag
        docpad.generating = false;
        docpad.generateEnded = new Date();
        // Update caches
        docpad.databaseTempCache = null;
        // Create Regenerate Timer
        docpad.createRegenerateTimer();
        if (err) {
          // Error?
          return next(err);
        }
        // Log success message
        seconds = (docpad.generateEnded - docpad.generateStarted) / 1000;
        howMany = `${((ref1 = opts.collection) != null ? ref1.length : void 0) || 0}/${database.length}`;
        docpad.log('info', util.format(locale.renderGenerated, howMany, seconds));
        docpad.notify((new Date()).toLocaleTimeString(), {
          title: locale.renderGeneratedNotification
        });
        // Generated
        if (opts.initial === true) {
          docpad.generated = true;
          return docpad.emitSerial('generated', opts, next);
        // Safety check if generated is false but initial was false too
        // https://github.com/bevry/docpad/issues/811
        } else if (docpad.generated === false) {
          return next(new Error('DocPad is in an invalid state, please report this. Reference 3360'));
        } else {
          return next();
        }
      });
      // Extract functions from tasks for simplicity
      // when dealing with nested tasks/groups
      addGroup = tasks.addGroup.bind(tasks);
      addTask = tasks.addTask.bind(tasks);
      // Setup a clean database
      addTask('Reset our collections', function(complete) {
        if (!(opts.reset === true || docpad.generated === false)) {
          // Skip if we are not a reset generation, or an initial generation (generated is false)
          return complete();
        }
        return docpad.resetCollections(opts, complete);
      });
      // Figure out the options
      // This is here as resetCollections could change our state
      // https://github.com/bevry/docpad/issues/811
      addTask('Figure out options', function() {
        // Mode: Cache
        // Shall we write to the database cache?
        // Set to true if the configuration option says we can, and we are the initial generation
        if (opts.cache == null) {
          opts.cache = config.databaseCache;
        }
        // Mode: Initial
        // Shall we do some basic initial checks
        // Set to the opts.reset value if specified, or whether are the initial generation
        if (opts.initial == null) {
          opts.initial = !docpad.generated;
        }
        // Mode: Reset
        // Shall we reset the database
        // Set to true if we are the initial generation
        if (opts.reset == null) {
          opts.reset = opts.initial;
        }
        // Mode: Populate
        // Shall we fetch in new data?
        // Set to the opts.reset value if specified, or the opts.initial value
        if (opts.populate == null) {
          opts.populate = opts.reset;
        }
        // Mode: Reload
        // Shall we rescan the file system for changes?
        // Set to the opts.reset value if specified, or the opts.initial value
        if (opts.reload == null) {
          opts.reload = opts.reset;
        }
        // Mode: Partial
        // Shall we perform a partial generation (false) or a completion generation (true)?
        // Set to false if we are the initial generation
        if (opts.partial == null) {
          opts.partial = !opts.reset;
        }
        // Log our opts
        return docpad.log('debug', 'Generate options:', pick(opts, ['cache', 'initial', 'reset', 'populate', 'reload', 'partial', 'renderPasses']));
      });
      // Check directory structure
      addTask('check source directory exists', function(complete) {
        if (opts.initial !== true) {
          // Skip if we are not the initial generation
          return complete();
        }
        // Continue if we are the initial generation
        return safefs.exists(config.srcPath, function(exists) {
          var err;
          // Check
          if (!exists) {
            err = new Error(locale.renderNonexistant);
            return complete(err);
          }
          // Forward
          return complete();
        });
      });
      addGroup('fetch data to render', function(addGroup, addTask) {
        var ref1;
        // Fetch new data
        // If we are a populate generation (by default an initial generation)
        if (opts.populate === true) {
          // This will pull in new data from plugins
          addTask('populateCollectionsBefore', function(complete) {
            return docpad.emitSerial('populateCollectionsBefore', opts, complete);
          });
          // Import the cached data
          // If we are the initial generation, and we have caching enabled
          if (opts.initial === true && ((ref1 = opts.cache) === true || ref1 === 'read')) {
            addTask('import data from cache', function(complete) {
              // Check if we do have a databae cache
              return safefs.exists(config.databaseCachePath, function(exists) {
                if (exists === false) {
                  return complete();
                }
                // Read the database cache if it exists
                return safefs.readFile(config.databaseCachePath, function(err, data) {
                  var addedModels, databaseData;
                  if (err) {
                    return complete(err);
                  }
                  // Parse it and apply the data values
                  databaseData = JSON.parse(data.toString());
                  opts.cache = true;
                  opts.initial = true;
                  opts.reset = false;
                  opts.populate = true;
                  opts.reload = true;
                  opts.partial = true;
                  lastGenerateStarted = new Date(databaseData.generateStarted);
                  addedModels = docpad.addModels(databaseData.models);
                  docpad.log('info', util.format(locale.databaseCacheRead, database.length, databaseData.models.length));
                  // @TODO we need a way of detecting deleted files between generations
                  return complete();
                });
              });
            });
          }
          // Rescan the file system
          // If we are a reload generation (by default an initial generation)
          // This is useful when the database is out of sync with the source files
          // For instance, someone shut down docpad, and made some changes, then ran docpad again
          // See https://github.com/bevry/docpad/issues/705#issuecomment-29243666 for details
          if (opts.reload === true) {
            addGroup('import data from file system', function(addGroup, addTask) {
              // Documents
              config.documentsPaths.forEach(function(documentsPath) {
                return addTask('import documents', function(complete) {
                  return docpad.parseDirectory({
                    modelType: 'document',
                    collection: database,
                    path: documentsPath,
                    next: complete
                  });
                });
              });
              // Files
              config.filesPaths.forEach(function(filesPath) {
                return addTask('import files', function(complete) {
                  return docpad.parseDirectory({
                    modelType: 'file',
                    collection: database,
                    path: filesPath,
                    next: complete
                  });
                });
              });
              // Layouts
              return config.layoutsPaths.forEach(function(layoutsPath) {
                return addTask('import layouts', function(complete) {
                  return docpad.parseDirectory({
                    modelType: 'document',
                    collection: database,
                    path: layoutsPath,
                    next: complete
                  });
                });
              });
            });
          }
          // This will pull in new data from plugins
          return addTask('populateCollections', function(complete) {
            return docpad.emitSerial('populateCollections', opts, complete);
          });
        }
      });
      addGroup('determine files to render', function(addGroup, addTask) {
        // Perform a complete regeneration
        // If we are a reset generation (by default an initial non-cached generation)
        if (opts.partial === false) {
          // Use Entire Collection
          return addTask('Add all database models to render queue', function() {
            return opts.collection != null ? opts.collection : opts.collection = new FilesCollection().add(docpad.getCollection('generate').models);
          });
        } else {
          // Use Partial Collection
          // Perform a partial regeneration
          // If we are not a reset generation (by default any non-initial generation)
          return addTask('Add only changed models to render queue', function() {
            var changedQuery;
            changedQuery = {
              $or: {
                // Get changed files
                mtime: {
                  $gte: lastGenerateStarted
                },
                // Get new files
                $and: {
                  wtime: null,
                  write: true
                }
              }
            };
            return opts.collection != null ? opts.collection : opts.collection = new FilesCollection().add(docpad.getCollection('generate').findAll(changedQuery).models);
          });
        }
      });
      addTask('generateBefore', function(complete) {
        // If we have nothing to generate
        if (opts.collection.length === 0) {
          // then there is no need to execute further tasks
          tasks.clear();
          return complete();
        } else {
          // Otherwise continue down the task loop
          return docpad.emitSerial('generateBefore', opts, complete);
        }
      });
      addTask('prepare files', function(complete) {
        var addLayoutChildren, allStandalone, model, standalones;
        // Log the files to generate if we are in debug mode
        docpad.log('debug', 'Files to generate at', lastGenerateStarted, '\n', (function() {
          var i, len, ref1, results;
          ref1 = opts.collection.models;
          results = [];
          for (i = 0, len = ref1.length; i < len; i++) {
            model = ref1[i];
            results.push({
              id: model.id,
              path: model.getFilePath(),
              mtime: model.get('mtime'),
              wtime: model.get('wtime'),
              dynamic: model.get('dynamic'),
              ignored: model.get('ignored'),
              write: model.get('write')
            });
          }
          return results;
        })());
        // Add anything that references other documents (e.g. partials, listing, etc)
        // This could eventually be way better
        standalones = opts.collection.pluck('standalone');
        allStandalone = standalones.indexOf(false) === -1;
        if (allStandalone === false) {
          opts.collection.add(docpad.getCollection('referencesOthers').models);
        }
        // Deeply/recursively add the layout children
        addLayoutChildren = function(collection) {
          return collection.forEach(function(file) {
            var layoutChildrenCollection, layoutChildrenQuery;
            if (file.get('isLayout') === true) {
              // Find
              layoutChildrenQuery = {
                layoutRelativePath: file.get('relativePath')
              };
              layoutChildrenCollection = docpad.getCollection('hasLayout').findAll(layoutChildrenQuery);
              // Log the files to generate if we are in debug mode
              docpad.log('debug', 'Layout children to generate at', lastGenerateStarted, '\n', (function() {
                var i, len, ref1, results;
                ref1 = layoutChildrenCollection.models;
                results = [];
                for (i = 0, len = ref1.length; i < len; i++) {
                  model = ref1[i];
                  results.push({
                    id: model.id,
                    path: model.getFilePath(),
                    mtime: model.get('mtime'),
                    wtime: model.get('wtime'),
                    write: model.get('write')
                  });
                }
                return results;
              })(), '\n', layoutChildrenQuery);
              // Recurse
              addLayoutChildren(layoutChildrenCollection);
              // Add
              return opts.collection.add(layoutChildrenCollection.models);
            }
          });
        };
        addLayoutChildren(opts.collection);
        // Filter out ignored, and no-render no-write files
        opts.collection.reset(opts.collection.reject(function(file) {
          return file.get('render') === false && file.get('write') === false;
        }));
        // Log the files to generate if we are in debug mode
        docpad.log('debug', 'Files to generate at', lastGenerateStarted, '\n', (function() {
          var i, len, ref1, results;
          ref1 = opts.collection.models;
          results = [];
          for (i = 0, len = ref1.length; i < len; i++) {
            model = ref1[i];
            results.push({
              id: model.id,
              path: model.getFilePath(),
              mtime: model.get('mtime'),
              wtime: model.get('wtime'),
              dynamic: model.get('dynamic'),
              ignored: model.get('ignored'),
              write: model.get('write')
            });
          }
          return results;
        })());
        // Forward
        return complete();
      });
      addGroup('process file', function(addGroup, addTask) {
        addTask('contextualizeFiles', {
          args: [opts]
        }, docpad.contextualizeFiles.bind(docpad));
        addTask('renderFiles', {
          args: [opts]
        }, docpad.renderFiles.bind(docpad));
        return addTask('writeFiles', {
          args: [opts]
        }, docpad.writeFiles.bind(docpad));
      });
      addTask('generateAfter', function(complete) {
        return docpad.emitSerial('generateAfter', opts, complete);
      });
      // Write the cache file
      addTask('Write the database cache', function(complete) {
        var databaseData, databaseDataDump, model, ref1;
        if ((ref1 = opts.cache) !== true && ref1 !== 'write') {
          // Skip if we do not care for writing the cache
          return complete();
        }
        // Write the cache
        databaseData = {
          generateStarted: docpad.generateStarted,
          generateEnded: docpad.generateEnded,
          models: (function() {
            var i, len, ref2, results;
            ref2 = database.models;
            results = [];
            for (i = 0, len = ref2.length; i < len; i++) {
              model = ref2[i];
              results.push(model.getAttributes());
            }
            return results;
          })()
        };
        databaseDataDump = JSON.stringify(databaseData, null, '  ');
        docpad.log('info', util.format(locale.databaseCacheWrite, databaseData.models.length));
        return safefs.writeFile(config.databaseCachePath, databaseDataDump, complete);
      });
      // Run
      tasks.run();
      return this;
    }

    // ---------------------------------
    // Render
    /**
     * Load a document
     * @private
     * @method loadDocument
     * @param {Object} document
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.document
     */
    loadDocument(document, opts, next) {
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      // Load
      // @TODO: don't load if already loaded
      document.action('load contextualize', opts, next);
      return this;
    }

    /**
     * Load and render a document
     * @method loadAndRenderDocument
     * @param {Object} document
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.document
     */
    loadAndRenderDocument(document, opts, next) {
      var docpad;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      // Load
      docpad.loadDocument(document, opts, function(err) {
        if (err) {
          return next(err);
        }
        // Render
        return docpad.renderDocument(document, opts, next);
      });
      return this;
    }

    /**
     * Render a document
     * @method renderDocument
     * @param {Object} document
     * @param {Object} opts
     * @param {Object} next
     * @param {Error} next.err
     * @param {Object} next.result
     * @param {Object} next.document
     */
    renderDocument(document, opts, next) {
      var clone;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      // Render
      clone = document.clone().action('render', opts, function(err) {
        var result;
        result = clone.getOutContent();
        return next(err, result, document);
      });
      return this;
    }

    /**
     * Render a document at a file path
     * next(err,result)
     * @method renderPath
     * @param {String} path
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.result the rendered document
     */
    renderPath(path, opts, next) {
      var attributes, document;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      attributes = extendr.extend({
        fullPath: path
      }, opts.attributes || {});
      // Handle
      document = this.createDocument(attributes);
      this.loadAndRenderDocument(document, opts, next);
      return this;
    }

    /**
     * Render the passed content data as a
     * document. Required option, filename
     * (opts.filename)
     * next(err,result)
     * @method renderData
     * @param {String} content
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.result the rendered document
     */
    renderData(content, opts, next) {
      var attributes, document;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      attributes = extendr.extend({
        filename: opts.filename,
        data: content
      }, opts.attributes || {});
      // Handle
      document = this.createDocument(attributes);
      this.loadAndRenderDocument(document, opts, next);
      return this;
    }

    // Render Text
    // Doesn't extract meta information, or render layouts
    // TODO: Why not? Why not just have renderData?
    /**
     * Render the passed text data as a
     * document. Required option, filename
     * (opts.filename)
     * next(err,result)
     * @private
     * @method renderText
     * @param {String} text
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     * @param {Object} next.result the rendered content
     * @param {Object} next.document the rendered document model
     */
    renderText(text, opts, next) {
      var attributes, clone, document;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      if (opts.actions == null) {
        opts.actions = ['renderExtensions', 'renderDocument'];
      }
      attributes = extendr.extend({
        filename: opts.filename,
        data: text,
        body: text,
        content: text
      }, opts.attributes || {});
      // Handle
      document = this.createDocument(attributes);
      // Render
      clone = document.clone().action('normalize contextualize render', opts, function(err) {
        var result;
        result = clone.getOutContent();
        return next(err, result, document);
      });
      return this;
    }

    /**
     * Render action
     * next(err,document,result)
     * @method render
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     */
    render(opts, next) {
      var err, locale, path;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      locale = this.getLocale();
      // Extract document
      if (opts.document) {
        this.renderDocument(opts.document, opts, next);
      } else if (opts.data) {
        this.renderData(opts.data, opts, next);
      } else if (opts.text) {
        this.renderText(opts.text, opts, next);
      } else {
        path = opts.path || opts.fullPath || opts.filename || null;
        if (path) {
          this.renderPath(path, opts, next);
        } else {
          // Check
          err = new Error(locale.renderInvalidOptions);
          return next(err);
        }
      }
      return this;
    }

    /**
     * Destructor. Destroy the watchers used
     * by DocPad
     * @private
     * @method destroyWatchers
     */
    destroyWatchers() {
      var docpad, i, len, ref, watcher;
      // Prepare
      docpad = this;
      // Check
      if (docpad.watchers) {
        ref = docpad.watchers;
        // Close each of them
        for (i = 0, len = ref.length; i < len; i++) {
          watcher = ref[i];
          watcher.close();
        }
        // Reset the array
        docpad.watchers = [];
      }
      return this;
    }

    /**
     * Start up file watchers used by DocPad
     * @private
     * @method watch
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     */
    watch(opts, next) {
      var changeHandler, config, database, docpad, locale, performGenerate, queueRegeneration, regenerateTimer, restartWatchers;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      config = this.getConfig();
      locale = this.getLocale();
      database = this.getDatabase();
      if (this.watchers == null) {
        this.watchers = [];
      }
      // Restart our watchers
      restartWatchers = function(next) {
        var regeneratePaths, reloadPaths, srcPath, tasks;
        // Close our watchers
        docpad.destroyWatchers();
        // Start a group
        tasks = docpad.createTaskGroup("watch tasks", {
          concurrency: 0,
          next
        });
        // Watch reload paths
        reloadPaths = union(config.reloadPaths, config.configPaths);
        tasks.addTask("watch reload paths", function(complete) {
          return docpad.watchdirs(reloadPaths, {
            'log': docpad.log,
            'error': docpad.error,
            'change': function() {
              docpad.log('info', util.format(locale.watchReloadChange, new Date().toLocaleTimeString()));
              return docpad.action('load', function(err) {
                if (err) {
                  return docpad.fatal(err);
                }
                return performGenerate({
                  reset: true
                });
              });
            }
          }, function(err, _watchers) {
            var i, len, watcher;
            if (err) {
              docpad.warn("Watching the reload paths has failed:\n" + docpad.inspector(reloadPaths), err);
              return complete();
            }
            for (i = 0, len = _watchers.length; i < len; i++) {
              watcher = _watchers[i];
              docpad.watchers.push(watcher);
            }
            return complete();
          });
        });
        // Watch regenerate paths
        regeneratePaths = config.regeneratePaths;
        tasks.addTask("watch regenerate paths", function(complete) {
          return docpad.watchdirs(regeneratePaths, {
            'log': docpad.log,
            'error': docpad.error,
            'change': function() {
              return performGenerate({
                reset: true
              });
            }
          }, function(err, _watchers) {
            var i, len, watcher;
            if (err) {
              docpad.warn("Watching the regenerate paths has failed:\n" + docpad.inspector(regeneratePaths), err);
              return complete();
            }
            for (i = 0, len = _watchers.length; i < len; i++) {
              watcher = _watchers[i];
              docpad.watchers.push(watcher);
            }
            return complete();
          });
        });
        // Watch the source
        srcPath = config.srcPath;
        tasks.addTask("watch the source path", function(complete) {
          return docpad.watchdirs([srcPath], {
            'log': docpad.log,
            'error': docpad.error,
            'change': changeHandler
          }, function(err, _watchers) {
            var i, len, watcher;
            if (err) {
              docpad.warn("Watching the src path has failed: " + srcPath, err);
              return complete();
            }
            for (i = 0, len = _watchers.length; i < len; i++) {
              watcher = _watchers[i];
              docpad.watchers.push(watcher);
            }
            return complete();
          });
        });
        // Run
        tasks.run();
        return this;
      };
      // Timer
      regenerateTimer = null;
      queueRegeneration = function() {
        // Reset the wait
        if (regenerateTimer) {
          clearTimeout(regenerateTimer);
          regenerateTimer = null;
        }
        // Regenerat after a while
        return regenerateTimer = setTimeout(performGenerate, config.regenerateDelay);
      };
      performGenerate = function(opts = {}) {
        // Q: Should we also pass over the collection?
        // A: No, doing the mtime query in generate is more robust

        // Log
        docpad.log(util.format(locale.watchRegenerating, new Date().toLocaleTimeString()));
        // Afterwards, re-render anything that should always re-render
        return docpad.action('generate', opts, function(err) {
          if (err) {
            docpad.error(err);
          }
          return docpad.log(util.format(locale.watchRegenerated, new Date().toLocaleTimeString()));
        });
      };
      // Change event handler
      changeHandler = function(changeType, filePath, fileCurrentStat, filePreviousStat) {
        var err, file, fileEitherStat, isDirectory, isIgnored;
        // Prepare
        fileEitherStat = fileCurrentStat || filePreviousStat;
        // For some reason neither of the stats may exist, this will cause errors as this is an invalid state
        // as we depend on at least one stat existing, otherwise, what on earth is going on?
        // Whatever the case, this should be fixed within watchr, not docpad
        // as watchr should not be giving us invalid data
        // https://github.com/bevry/docpad/issues/792
        if (!fileEitherStat) {
          err = new Error("DocPad has encountered an invalid state while detecting changes for your files.\nSo the DocPad team can fix this right away, please provide any information you can to:\nhttps://github.com/bevry/docpad/issues/792");
          return docpad.error(err);
        }
        // Log the change
        docpad.log('info', util.format(locale.watchChange, new Date().toLocaleTimeString()), changeType, filePath);
        // Check if we are a file we don't care about
        // This check should not be needed with v2.3.3 of watchr
        // however we've still got it here as it may still be an issue
        isIgnored = docpad.isIgnoredPath(filePath);
        if (isIgnored) {
          docpad.log('debug', util.format(locale.watchIgnoredChange, new Date().toLocaleTimeString()), filePath);
          return;
        }
        // Don't care if we are a directory
        isDirectory = fileEitherStat.isDirectory();
        if (isDirectory) {
          docpad.log('debug', util.format(locale.watchDirectoryChange, new Date().toLocaleTimeString()), filePath);
          return;
        }
        // Override the stat's mtime to now
        // This is because renames will not update the mtime
        if (fileCurrentStat != null) {
          fileCurrentStat.mtime = new Date();
        }
        // Create the file object
        file = docpad.addModel({
          fullPath: filePath,
          stat: fileCurrentStat
        });
        if (changeType === 'update') {
          file.setStat(fileCurrentStat);
        }
        // File was deleted, delete the rendered file, and remove it from the database
        if (changeType === 'delete') {
          database.remove(file);
          return file.action('delete', function(err) {
            if (err) {
              return docpad.error(err);
            }
            return queueRegeneration();
          });
        // File is new or was changed, update it's mtime by setting the stat
        } else if (changeType === 'create' || changeType === 'update') {
          return file.action('load', function(err) {
            if (err) {
              return docpad.error(err);
            }
            return queueRegeneration();
          });
        }
      };
      // Watch
      docpad.log(locale.watchStart);
      restartWatchers(function(err) {
        if (err) {
          return next(err);
        }
        docpad.log(locale.watchStarted);
        return next();
      });
      return this;
    }

    // ---------------------------------
    // Run Action
    /**
     * Run an action
     * @method run
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     */
    run(opts, next) {
      var config, docpad, locale, rootPath, run, srcPath;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      locale = this.getLocale();
      config = this.getConfig();
      ({srcPath, rootPath} = config);
      // Prepare
      run = function(complete) {
        return balUtil.flow({
          object: docpad,
          action: 'server generate watch',
          args: [opts],
          next: complete
        });
      };
      // Check if we have the docpad structure
      safefs.exists(srcPath, function(exists) {
        if (exists) {
          // Check if have the correct structure, if so let's proceed with DocPad
          return run(next);
        }
        // We don't have the correct structure
        // Check if we are running on an empty directory
        return safefs.readdir(rootPath, function(err, files) {
          if (err) {
            return next(err);
          }
          // Check if our directory is empty
          if (files.length) {
            // It isn't empty, display a warning
            docpad.warn(util.format(locale.skeletonNonexistant, rootPath));
            return next();
          } else {
            return docpad.skeleton(opts, function(err) {
              if (err) {
                // Check
                return next(err);
              }
              if (opts.global === true || docpad.getConfig().global === true) {
                // Keep in global?
                return run(next);
              }
              // Log
              docpad.log('notice', locale.startLocal);
              // Destroy our DocPad instance so we can boot the local one
              return docpad.destroy(function(err) {
                if (err) {
                  // Check
                  return next(err);
                }
                // Forward onto the local DocPad Instance now that it has been installed
                return docpadUtil.startLocalDocPadExecutable(next);
              });
            });
          }
        });
      });
      return this;
    }

    // ---------------------------------
    // Skeleton
    /**
     * Initialize the skeleton install process.
     * @private
     * @method initInstall
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     */
    initInstall(opts, next) {
      var config, docpad, tasks;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      config = this.getConfig();
      // Tasks
      tasks = this.createTaskGroup("initInstall tasks", {
        concurrency: 0,
        next
      });
      tasks.addTask("node modules", function(complete) {
        var path;
        path = pathUtil.join(config.rootPath, 'node_modules');
        return safefs.ensurePath(path, complete);
      });
      tasks.addTask("package", function(complete) {
        var path;
        // Exists?
        path = pathUtil.join(config.rootPath, 'package.json');
        return safefs.exists(path, function(exists) {
          var data;
          if (exists) {
            // Check
            return complete();
          }
          // Write
          data = JSON.stringify({
            name: 'no-skeleton.docpad',
            version: '0.1.0',
            description: 'New DocPad project without using a skeleton',
            dependencies: {
              docpad: '~' + docpad.getVersion()
            },
            main: 'node_modules/.bin/docpad-server',
            scripts: {
              start: 'node_modules/.bin/docpad-server'
            }
          }, null, '  ');
          return safefs.writeFile(path, data, complete);
        });
      });
      // Run
      tasks.run();
      return this;
    }

    /**
     * Uninstall a plugin.
     * @private
     * @method uninstall
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     */
    uninstall(opts, next) {
      var config, docpad, tasks;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      config = this.getConfig();
      // Tasks
      tasks = this.createTaskGroup("uninstall tasks", {next});
      // Uninstall a plugin
      if (opts.plugin) {
        tasks.addTask(`uninstall the plugin: ${opts.plugin}`, function(complete) {
          var plugin, plugins;
          plugins = (function() {
            var i, len, ref, results;
            ref = opts.plugin.split(/[,\s]+/);
            results = [];
            for (i = 0, len = ref.length; i < len; i++) {
              plugin = ref[i];
              if (plugin.indexOf('docpad-plugin-') !== 0) {
                plugin = `docpad-plugin-${plugin}`;
              }
              results.push(plugin);
            }
            return results;
          })();
          return docpad.uninstallNodeModule(plugins, {
            stdio: 'inherit',
            next: complete
          });
        });
      }
      // Re-load configuration
      tasks.addTask("re-load configuration", function(complete) {
        return docpad.load(complete);
      });
      // Run
      tasks.run();
      return this;
    }

    /**
     * Install a plugin
     * @private
     * @method install
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     */
    install(opts, next) {
      var config, docpad, tasks;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      config = this.getConfig();
      // Tasks
      tasks = this.createTaskGroup("install tasks", {next});
      tasks.addTask("init the installation", function(complete) {
        return docpad.initInstall(opts, complete);
      });
      // Install a plugin
      if (opts.plugin) {
        tasks.addTask(`install the plugin: ${opts.plugin}`, function(complete) {
          var plugin, plugins;
          plugins = (function() {
            var i, len, ref, results;
            ref = opts.plugin.split(/[,\s]+/);
            results = [];
            for (i = 0, len = ref.length; i < len; i++) {
              plugin = ref[i];
              if (plugin.indexOf('docpad-plugin-') !== 0) {
                plugin = `docpad-plugin-${plugin}`;
              }
              if (plugin.indexOf('@') === -1) {
                plugin += '@' + docpad.pluginVersion;
              }
              results.push(plugin);
            }
            return results;
          })();
          return docpad.installNodeModule(plugins, {
            stdio: 'inherit',
            next: complete
          });
        });
      }
      tasks.addTask("re-initialize the website's modules", function(complete) {
        return docpad.initNodeModules({
          stdio: 'inherit',
          next: complete
        });
      });
      tasks.addTask("fix node package versions", function(complete) {
        return docpad.fixNodePackageVersions(complete);
      });
      tasks.addTask("re-load the configuration", function(complete) {
        return docpad.load(complete);
      });
      // Run
      tasks.run();
      return this;
    }

    /**
     * Update global NPM and DocPad
     * @private
     * @method upgrade
     * @param {Object} opts
     * @param {Object} next
     * @param {Error} next.err
     * @return {Object} description
     */
    upgrade(opts, next) {
      // Update Global NPM and DocPad
      this.installNodeModule('npm docpad@6', {
        global: true,
        stdio: 'inherit',
        next: next
      });
      return this;
    }

    /**
     * Update the local DocPad and plugin dependencies
     * @private
     * @method update
     * @param {Object} opts
     * @param {Object} next
     * @param {Error} next.err
     */
    update(opts, next) {
      var config, dependencies, devDependencies, docpad, tasks;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      config = this.getConfig();
      // Tasks
      tasks = this.createTaskGroup("update tasks", {next});
      tasks.addTask("init the install", function(complete) {
        return docpad.initInstall(opts, complete);
      });
      // Update the local docpad and plugin dependencies
      // Grouped together to avoid npm dependency shortcuts that can cause missing dependencies
      // But don't update git/http/https dependencies, those are special for some reason
      // > https://github.com/bevry/docpad/pull/701
      dependencies = [];
      eachr(docpad.websitePackageConfig.dependencies, function(version, name) {
        if (/^docpad-plugin-/.test(name) === false || /:\/\//.test(version) === true) {
          return;
        }
        return dependencies.push(name + '@' + docpad.pluginVersion);
      });
      if (dependencies.length !== 0) {
        tasks.addTask("update plugins that are dependencies", function(complete) {
          return docpad.installNodeModule('docpad@6 ' + dependencies, {
            stdio: 'inherit',
            next: complete
          });
        });
      }
      // Update the plugin dev dependencies
      devDependencies = [];
      eachr(docpad.websitePackageConfig.devDependencies, function(version, name) {
        if (/^docpad-plugin-/.test(name) === false) {
          return;
        }
        return devDependencies.push(name + '@' + docpad.pluginVersion);
      });
      if (devDependencies.length !== 0) {
        tasks.addTask("update plugins that are dev dependencies", function(complete) {
          return docpad.installNodeModule(devDependencies, {
            save: '--save-dev',
            stdio: 'inherit',
            next: complete
          });
        });
      }
      tasks.addTask("fix node package versions", function(complete) {
        return docpad.fixNodePackageVersions(complete);
      });
      tasks.addTask("re-initialize the rest of the website's modules", function(complete) {
        return docpad.initNodeModules({
          stdio: 'inherit',
          next: complete
        });
      });
      // Run
      tasks.run();
      return this;
    }

    /**
     * DocPad cleanup tasks.
     * @private
     * @method clean
     * @param {Object} opts
     * @param {Object} next
     * @param {Error} next.err
     * @return {Object} description
     */
    clean(opts, next) {
      var config, docpad, locale, tasks;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      config = docpad.getConfig();
      locale = this.getLocale();
      // Log
      docpad.log('info', locale.renderCleaning);
      // Tasks
      tasks = this.createTaskGroup("clean tasks", {
        concurrency: 0
      }, {
        next: function(err) {
          if (err) {
            // Error?
            return next(err);
          }
          // Log
          docpad.log('info', locale.renderCleaned);
          // Forward
          return next();
        }
      });
      tasks.addTask('reset the collecitons', function(complete) {
        return docpad.resetCollections(opts, complete);
      });
      // Delete out path
      // but only if our outPath is not a parent of our rootPath
      tasks.addTask('delete out path', function(complete) {
        if (config.rootPath.indexOf(config.outPath) !== -1) {
          // Check if our outPath is higher than our root path, so do not remove files
          return complete();
        }
        // Our outPath is not related or lower than our root path, so do remove it
        return rimraf(config.outPath, complete);
      });
      // Delete database cache
      tasks.addTask('delete database cache file', function(complete) {
        return safefs.unlink(config.databaseCachePath, complete);
      });
      // Run tasks
      tasks.run();
      return this;
    }

    /**
     * Initialize a Skeleton into to a Directory
     * @private
     * @method initSkeleton
     * @param {Object} skeletonModel
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     */
    initSkeleton(skeletonModel, opts, next) {
      var config, docpad, tasks;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      config = this.getConfig();
      // Defaults
      if (opts.destinationPath == null) {
        opts.destinationPath = config.rootPath;
      }
      // Tasks
      tasks = this.createTaskGroup("initSkeleton tasks", {next});
      tasks.addTask("ensure the path we are writing to exists", function(complete) {
        return safefs.ensurePath(opts.destinationPath, complete);
      });
      // Clone out the repository if applicable
      if ((skeletonModel != null) && skeletonModel.id !== 'none') {
        tasks.addTask("clone out the git repo", function(complete) {
          return docpad.initGitRepo({
            cwd: opts.destinationPath,
            url: skeletonModel.get('repo'),
            branch: skeletonModel.get('branch'),
            remote: 'skeleton',
            stdio: 'inherit',
            next: complete
          });
        });
      } else {
        tasks.addTask("ensure src path exists", function(complete) {
          return safefs.ensurePath(config.srcPath, complete);
        });
        tasks.addGroup("initialize the website directory files", function() {
          this.setConfig({
            concurrency: 0
          });
          // README
          this.addTask("README.md", function(complete) {
            var path;
            // Exists?
            path = pathUtil.join(config.rootPath, 'README.md');
            return safefs.exists(path, function(exists) {
              var data;
              if (exists) {
                // Check
                return complete();
              }
              // Write
              data = `# Your [DocPad](http://docpad.org) Project\n\n## License\nCopyright &copy; ${(new Date()).getFullYear()}+ All rights reserved.`;
              return safefs.writeFile(path, data, complete);
            });
          });
          // Config
          this.addTask("docpad.coffee configuration file", function(complete) {
            // Exists?
            return docpad.getConfigPath(function(err, path) {
              var data;
              if (err || path) {
                // Check
                return complete(err);
              }
              path = pathUtil.join(config.rootPath, 'docpad.coffee');
              // Write
              data = "# DocPad Configuration File\n# http://docpad.org/docs/config\n\n# Define the DocPad Configuration\ndocpadConfig = {\n	# ...\n}\n\n# Export the DocPad Configuration\nmodule.exports = docpadConfig";
              return safefs.writeFile(path, data, complete);
            });
          });
          // Documents
          this.addTask("documents directory", function(complete) {
            return safefs.ensurePath(config.documentsPaths[0], complete);
          });
          // Layouts
          this.addTask("layouts directory", function(complete) {
            return safefs.ensurePath(config.layoutsPaths[0], complete);
          });
          // Files
          return this.addTask("files directory", function(complete) {
            return safefs.ensurePath(config.filesPaths[0], complete);
          });
        });
      }
      // Run
      tasks.run();
      return this;
    }

    /**
     * Install a Skeleton into a Directory
     * @private
     * @method installSkeleton
     * @param {Object} skeletonModel
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     */
    installSkeleton(skeletonModel, opts, next) {
      var docpad;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      // Defaults
      if (opts.destinationPath == null) {
        opts.destinationPath = this.getConfig().rootPath;
      }
      // Initialize and install the skeleton
      docpad.initSkeleton(skeletonModel, opts, function(err) {
        if (err) {
          // Check
          return next(err);
        }
        // Forward
        return docpad.install(opts, next);
      });
      return this;
    }

    /**
     * Use a Skeleton
     * @private
     * @method useSkeleton
     * @param {Object} skeletonModel
     * @param {Object} opts
     * @param {Object} next
     * @param {Error} next.err
     * @return {Object} description
     */
    useSkeleton(skeletonModel, opts, next) {
      var docpad, locale, skeletonId, skeletonName;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      locale = this.getLocale();
      // Defaults
      if (opts.destinationPath == null) {
        opts.destinationPath = this.getConfig().rootPath;
      }
      // Extract
      skeletonId = (skeletonModel != null ? skeletonModel.id : void 0) || 'none';
      skeletonName = (skeletonModel != null ? skeletonModel.get('name') : void 0) || locale.skeletonNoneName;
      // Log
      docpad.log('info', util.format(locale.skeletonInstall, skeletonName, opts.destinationPath) + ' ' + locale.pleaseWait);
      // Install Skeleton
      docpad.installSkeleton(skeletonModel, opts, function(err) {
        if (err) {
          // Error?
          return next(err);
        }
        // Log
        docpad.log('info', locale.skeletonInstalled);
        // Forward
        return next(err);
      });
      return this;
    }

    /**
     * Select a Skeleton
     * @private
     * @method selectSkeleton
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     * @param {Error} next.skeletonModel
     */
    selectSkeleton(opts, next) {
      var docpad;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      if (opts.selectSkeletonCallback == null) {
        opts.selectSkeletonCallback = null;
      }
      // Get the available skeletons
      docpad.getSkeletons(function(err, skeletonsCollection) {
        if (err) {
          // Check
          return next(err);
        }
        // Provide selection to the interface
        return opts.selectSkeletonCallback(skeletonsCollection, next);
      });
      return this;
    }

    /**
     * Skeleton Empty?
     * @private
     * @method skeletonEmpty
     * @param {Object} path
     * @param {Function} next
     * @param {Error} next.err
     */
    skeletonEmpty(path, next) {
      var locale;
      // Prepare
      locale = this.getLocale();
      // Defaults
      if (path == null) {
        path = this.getConfig().rootPath;
      }
      // Check the destination path is empty
      safefs.exists(pathUtil.join(path, 'package.json'), function(exists) {
        var err;
        // Check
        if (exists) {
          err = new Error(locale.skeletonExists);
          return next(err);
        }
        // Success
        return next();
      });
      return this;
    }

    /**
     * Initialize the project directory
     * with the basic skeleton.
     * @private
     * @method skeleton
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     */
    skeleton(opts, next) {
      var docpad;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      if (opts.selectSkeletonCallback == null) {
        opts.selectSkeletonCallback = null;
      }
      // Init the directory with the basic skeleton
      this.skeletonEmpty(null, function(err) {
        if (err) {
          // Check
          return next(err);
        }
        // Select Skeleton
        return docpad.selectSkeleton(opts, function(err, skeletonModel) {
          if (err) {
            // Check
            return next(err);
          }
          // Use Skeleton
          return docpad.useSkeleton(skeletonModel, next);
        });
      });
      return this;
    }

    /**
     * Initialize the project directory
     * with the basic skeleton.
     * @private
     * @method init
     * @param {Object} opts
     * @param {Object} next
     * @param {Error} next.err
     * @return {Object} description
     */
    init(opts, next) {
      var docpad;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      // Init the directory with the basic skeleton
      this.skeletonEmpty(null, function(err) {
        if (err) {
          // Check
          return next(err);
        }
        // Basic Skeleton
        return docpad.useSkeleton(null, next);
      });
      return this;
    }

    // ---------------------------------
    // Server
    /**
     * Serve a document
     * @private
     * @method serveDocument
     * @param {Object} opts
     * @param {Function} next
     * @param {Error} next.err
     */
    serveDocument(opts, next) {
      var collection, config, content, ctime, docpad, document, dynamic, err, etag, locale, mtime, req, res, stat, templateData;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      ({document, err, req, res} = opts);
      docpad = this;
      config = this.getConfig();
      locale = this.getLocale();
      // If no document, then exit early
      if (!document) {
        if (opts.statusCode != null) {
          return res.send(opts.statusCode);
        } else {
          return next();
        }
      }
      // Prepare
      if (res.setHeaderIfMissing == null) {
        res.setHeaderIfMissing = function(name, value) {
          if (!res.getHeader(name)) {
            return res.setHeader(name, value);
          }
        };
      }
      if (config.maxAge) {
        // Cache-Control (max-age)
        res.setHeaderIfMissing('Cache-Control', `public, max-age=${config.maxAge}`);
      }
      // Send
      dynamic = document.get('dynamic');
      if (dynamic) {
        docpad.log('warn', util.format(locale.documentDynamicError, document.getFilePath()));
        // If you are debugging why a dynamic document isn't rendering
        // it could be that you don't have cleanurls installed
        // e.g. if index.html is dynamic, and you are accessing it via /
        // then this code will not be reached, as we don't register that url
        // where if we have the cleanurls plugin installed, then do register that url
        // against the document, so this is reached
        collection = new FilesCollection([document], {
          name: 'dynamic collection'
        });
        templateData = extendr.extend({}, req.templateData || {}, {req, err});
        docpad.action('generate', {collection, templateData}, function(err) {
          var content;
          content = document.getOutContent();
          if (err) {
            docpad.error(err);
            return next(err);
          } else {
            if (opts.statusCode != null) {
              return res.send(opts.statusCode, content);
            } else {
              return res.send(content);
            }
          }
        });
      } else {
        // Content Type + Encoding/Charset
        res.setHeaderIfMissing('Content-Type', document.getContentTypeHeader());
        // ETag: `"<size>-<mtime>"`
        ctime = document.get('date'); // use the date or mtime, it should always exist
        mtime = document.get('wtime'); // use the last generate time, it may not exist though
        stat = document.getStat();
        if (mtime && stat) {
          etag = stat.size + '-' + Number(mtime);
        }
        if (etag) {
          res.setHeaderIfMissing('ETag', '"' + etag + '"');
        }
        if ((ctime != null ? ctime.toUTCString : void 0) != null) {
          // Date
          res.setHeaderIfMissing('Date', ctime.toUTCString());
        }
        if ((mtime != null ? mtime.toUTCString : void 0) != null) {
          res.setHeaderIfMissing('Last-Modified', mtime.toUTCString());
        }
        // @TODO:
        // The above .toUTCString? check is a workaround because sometimes the date object
        // isn't really a date object, this needs to be fixed properly
        // https://github.com/bevry/docpad/pull/781

        // Send
        if (etag && etag === (req.get('If-None-Match') || '').replace(/^"|"$/g, '')) {
          res.send(304); // not modified
        } else {
          content = document.getOutContent();
          if (content) {
            if (opts.statusCode != null) {
              res.send(opts.statusCode, content);
            } else {
              res.send(content);
            }
          } else {
            if (opts.statusCode != null) {
              res.send(opts.statusCode);
            } else {
              next();
            }
          }
        }
      }
      return this;
    }

    /**
     * Server Middleware: Header
     * @private
     * @method serverMiddlewareHeader
     * @param {Object} req
     * @param {Object} res
     * @param {Object} next
     */
    serverMiddlewareHeader(req, res, next) {
      var docpad, tools;
      // Prepare
      docpad = this;
      // Handle
      // Always enable this until we get a complaint about not having it
      // For instance, Express.js also forces this
      tools = res.get('X-Powered-By').split(/[,\s]+/g);
      tools.push(`DocPad v${docpad.getVersion()}`);
      tools = tools.join(', ');
      res.set('X-Powered-By', tools);
      // Forward
      next();
      return this;
    }

    /**
     * Server Middleware: Router
     * @private
     * @method serverMiddlewareRouter
     * @param {Object} req
     * @param {Object} res
     * @param {Function} next
     * @param {Error} next.err
     */
    serverMiddlewareRouter(req, res, next) {
      var docpad;
      // Prepare
      docpad = this;
      // Get the file
      docpad.getFileByRoute(req.url, function(err, file) {
        var cleanUrl, url;
        if (err || (file != null) === false) {
          // Check
          return next(err);
        }
        // Check if we are the desired url
        // if we aren't do a permanent redirect
        url = file.get('url');
        cleanUrl = docpad.getUrlPathname(req.url);
        if ((url !== cleanUrl) && (url !== req.url)) {
          return res.redirect(301, url);
        }
        // Serve the file to the user
        return docpad.serveDocument({
          document: file,
          req,
          res,
          next
        });
      });
      return this;
    }

    /**
     * Server Middleware: 404
     * @private
     * @method serverMiddleware404
     * @param {Object} req
     * @param {Object} res
     * @param {Object} next
     */
    serverMiddleware404(req, res, next) {
      var database, docpad, document;
      // Prepare
      docpad = this;
      database = docpad.getDatabaseSafe();
      // Notify the user of a 404
      docpad.log('notice', "404 Not Found:", req.url);
      if (!database) {
        // Check
        return res.send(500);
      }
      // Serve the document to the user
      document = database.findOne({
        relativeOutPath: '404.html'
      });
      docpad.serveDocument({
        document,
        req,
        res,
        next,
        statusCode: 404
      });
      return this;
    }

    /**
     * Server Middleware: 500
     * @private
     * @method serverMiddleware500
     * @param {Object} err
     * @param {Object} req
     * @param {Object} res
     * @param {Function} next
     */
    serverMiddleware500(err, req, res, next) {
      var database, docpad, document;
      // Prepare
      docpad = this;
      database = docpad.getDatabaseSafe();
      if (!database) {
        // Check
        return res.send(500);
      }
      // Serve the document to the user
      document = database.findOne({
        relativeOutPath: '500.html'
      });
      docpad.serveDocument({
        document,
        err,
        req,
        res,
        next,
        statusCode: 500
      });
      return this;
    }

    /**
     * Configure and start up the DocPad web server.
     * Http and express server is created, extended with
     * middleware, started up and begins listening.
     * The events serverBefore, serverExtend and
     * serverAfter emitted here.
     * @private
     * @method server
     * @param {Object} opts
     * @param {Function} next
     */
    server(opts, next) {
      var config, docpad, express, hostname, http, locale, port, ref, ref1, ref2, servers, tasks;
      // Prepare
      [opts, next] = extractOptsAndCallback(opts, next);
      docpad = this;
      config = this.config;
      locale = this.getLocale();
      port = this.getPort();
      hostname = this.getHostname();
      // Require
      http = require('http');
      express = require('express');
      // Config
      servers = this.getServer(true);
      if (opts.serverExpress == null) {
        opts.serverExpress = servers.serverExpress;
      }
      if (opts.serverHttp == null) {
        opts.serverHttp = servers.serverHttp;
      }
      if (opts.middlewareBodyParser == null) {
        opts.middlewareBodyParser = (ref = config.middlewareBodyParser) != null ? ref : config.middlewareStandard;
      }
      if (opts.middlewareMethodOverride == null) {
        opts.middlewareMethodOverride = (ref1 = config.middlewareMethodOverride) != null ? ref1 : config.middlewareStandard;
      }
      if (opts.middlewareExpressRouter == null) {
        opts.middlewareExpressRouter = (ref2 = config.middlewareExpressRouter) != null ? ref2 : config.middlewareStandard;
      }
      if (opts.middleware404 == null) {
        opts.middleware404 = config.middleware404;
      }
      if (opts.middleware500 == null) {
        opts.middleware500 = config.middleware500;
      }
      // @TODO: Why do we do opts here instead of config???

      // Tasks
      tasks = this.createTaskGroup("server tasks", {next});
      // Before Plugin Event
      tasks.addTask("emit serverBefore", function(complete) {
        return docpad.emitSerial('serverBefore', complete);
      });
      if (!opts.serverExpress || !opts.serverHttp) {
        tasks.addTask("create server", function() {
          opts.serverExpress || (opts.serverExpress = express());
          opts.serverHttp || (opts.serverHttp = http.createServer(opts.serverExpress));
          return docpad.setServer(opts);
        });
      }
      // Extend the server with our middlewares
      if (config.extendServer === true) {
        tasks.addTask("extend the server", function(complete) {
          // Parse url-encoded and json encoded form data
          if (opts.middlewareBodyParser !== false) {
            opts.serverExpress.use(express.urlencoded());
            opts.serverExpress.use(express.json());
          }
          // Allow over-riding of the request type (e.g. GET, POST, PUT, DELETE)
          if (opts.middlewareMethodOverride !== false) {
            if (typeChecker.isString(opts.middlewareMethodOverride)) {
              opts.serverExpress.use(require('method-override')(opts.middlewareMethodOverride));
            } else {
              opts.serverExpress.use(require('method-override')());
            }
          }
          // Emit the serverExtend event
          // So plugins can define their routes earlier than the DocPad routes
          return docpad.emitSerial('serverExtend', {
            server: opts.serverExpress, // b/c
            express: opts.serverExpress, // b/c
            serverHttp: opts.serverHttp,
            serverExpress: opts.serverExpress
          }, function(err) {
            if (err) {
              return next(err);
            }
            // DocPad Header Middleware
            // Keep it after the serverExtend event
            opts.serverExpress.use(docpad.serverMiddlewareHeader);
            if (opts.middlewareExpressRouter !== false) {
              // Router Middleware
              // Keep it after the serverExtend event
              opts.serverExpress.use(opts.serverExpress.router);
            }
            // DocPad Router Middleware
            // Keep it after the serverExtend event
            opts.serverExpress.use(docpad.serverMiddlewareRouter);
            // Static
            // Keep it after the serverExtend event
            if (config.maxAge) {
              opts.serverExpress.use(express.static(config.outPath, {
                maxAge: config.maxAge
              }));
            } else {
              opts.serverExpress.use(express.static(config.outPath));
            }
            if (opts.middleware404 !== false) {
              // DocPad 404 Middleware
              // Keep it after the serverExtend event
              opts.serverExpress.use(docpad.serverMiddleware404);
            }
            if (opts.middleware500 !== false) {
              // DocPad 500 Middleware
              // Keep it after the serverExtend event
              opts.serverExpress.use(docpad.serverMiddleware500);
            }
            // Done
            return complete();
          });
        });
      }
      // Start Server
      tasks.addTask("start the server", function(complete) {
        // Catch
        opts.serverHttp.once('error', function(err) {
          // Friendlify the error message if it is what we suspect it is
          if (err.message.indexOf('EADDRINUSE') !== -1) {
            err = new Error(util.format(locale.serverInUse, port));
          }
          // Done
          return complete(err);
        });
        // Listen
        docpad.log('debug', util.format(locale.serverStart, hostname, port));
        return opts.serverHttp.listen(port, hostname, function() {
          var address, serverUrl, simpleServerUrl;
          // Log
          address = opts.serverHttp.address();
          serverUrl = docpad.getServerUrl({
            hostname: address.hostname,
            port: address.port
          });
          simpleServerUrl = docpad.getSimpleServerUrl({
            hostname: address.hostname,
            port: address.port
          });
          docpad.log('info', util.format(locale.serverStarted, serverUrl));
          if (serverUrl !== simpleServerUrl) {
            docpad.log('info', util.format(locale.serverBrowse, simpleServerUrl));
          }
          // Done
          return complete();
        });
      });
      // After Plugin Event
      tasks.addTask("emit serverAfter", function(complete) {
        return docpad.emitSerial('serverAfter', {
          server: opts.serverExpress, // b/c
          express: opts.serverExpress, // b/c
          serverHttp: opts.serverHttp,
          serverExpress: opts.serverExpress
        }, complete);
      });
      // Run the tasks
      tasks.run();
      return this;
    }

  };

  // =================================
  // Variables

  // ---------------------------------
  // Modules

  // ---------------------------------
  // Base
  /**
   * Events class
   * https://github.com/docpad/docpad/blob/master/src/lib/base.coffee
   * @property {Object} Events
   */
  DocPad.prototype.Events = Events;

  /**
   * Model class
   * Extension of the Backbone Model class
   * http://backbonejs.org/#Model
   * https://github.com/docpad/docpad/blob/master/src/lib/base.coffee
   * @property {Object} Model
   */
  DocPad.prototype.Model = Model;

  /**
   * Collection class
   * Extension of the Backbone Collection class
   * https://github.com/docpad/docpad/blob/master/src/lib/base.coffee
   * http://backbonejs.org/#Collection
   * @property {Object} Collection
   */
  DocPad.prototype.Collection = Collection;

  /**
   * QueryCollection class
   * Extension of the Query Engine QueryCollection class
   * https://github.com/docpad/docpad/blob/master/src/lib/base.coffee
   * https://github.com/bevry/query-engine/blob/master/src/documents/lib/query-engine.js.coffee
   * @property {Object} QueryCollection
   */
  DocPad.prototype.QueryCollection = QueryCollection;

  // ---------------------------------
  // Models
  /**
   * File Model class
   * Extension of the Model class
   * https://github.com/docpad/docpad/blob/master/src/lib/models/file.coffee
   * @property {Object} FileModel
   */
  DocPad.prototype.FileModel = FileModel;

  /**
   * Document Model class
   * Extension of the File Model class
   * https://github.com/docpad/docpad/blob/master/src/lib/models/document.coffee
   * @property {Object} DocumentModel
   */
  DocPad.prototype.DocumentModel = DocumentModel;

  // ---------------------------------
  // Collections
  /**
   * Collection of files in a DocPad project
   * Extension of the QueryCollection class
   * https://github.com/docpad/docpad/blob/master/src/lib/collections/files.coffee
   * @property {Object} FilesCollection
   */
  DocPad.prototype.FilesCollection = FilesCollection;

  /**
   * Collection of elements in a DocPad project
   * Extension of the Collection class
   * https://github.com/docpad/docpad/blob/master/src/lib/collections/elements.coffee
   * @property {Object} ElementsCollection
   */
  DocPad.prototype.ElementsCollection = ElementsCollection;

  /**
   * Collection of metadata in a DocPad project
   * Extension of the ElementsCollection class
   * https://github.com/docpad/docpad/blob/master/src/lib/collections/meta.coffee
   * @property {Object} MetaCollection
   */
  DocPad.prototype.MetaCollection = MetaCollection;

  /**
   * Collection of JS script files in a DocPad project
   * Extension of the ElementsCollection class
   * https://github.com/docpad/docpad/blob/master/src/lib/collections/scripts.coffee
   * @property {Object} ScriptsCollection
   */
  DocPad.prototype.ScriptsCollection = ScriptsCollection;

  /**
   * Collection of CSS style files in a DocPad project
   * Extension of the ElementsCollection class
   * https://github.com/docpad/docpad/blob/master/src/lib/collections/styles.coffee
   * @property {Object} StylesCollection
   */
  DocPad.prototype.StylesCollection = StylesCollection;

  /**
   * Plugin Loader class
   * https://github.com/docpad/docpad/blob/master/src/lib/plugin-loader.coffee
   * Loads the DocPad plugins from the file system into
   * a DocPad project
   * @property {Object} PluginLoader
   */
  DocPad.prototype.PluginLoader = PluginLoader;

  /**
   * Base class for all DocPad plugins
   * https://github.com/docpad/docpad/blob/master/src/lib/plugin.coffee
   * @property {Object} BasePlugin
   */
  DocPad.prototype.BasePlugin = BasePlugin;

  // ---------------------------------
  // DocPad
  /**
   * DocPad's version number
   * @private
   * @property {Number} version
   */
  DocPad.prototype.version = null;

  /**
   * The plugin version requirements
   * @property {String} pluginVersion
   */
  DocPad.prototype.pluginVersion = '2';

  /**
   * The express.js server instance bound to DocPad.
   * http://expressjs.com
   * @private
   * @property {Object} serverExpress
   */
  DocPad.prototype.serverExpress = null;

  /**
   * The Node.js http server instance bound to DocPad
   * https://nodejs.org/api/http.html
   * @private
   * @property {Object} serverHttp
   */
  DocPad.prototype.serverHttp = null;

  // @TODO figure out how to destroy the express server

  /**
   * Internal property. The caterpillar logger instances bound to DocPad
   * @private
   * @property {Object} loggerInstances
   */
  DocPad.prototype.loggerInstances = null;

  /**
   * Instance of progress-title
   * @private
   * @property {Progress} progressInstance
   */
  DocPad.prototype.progressInstance = null;

  /**
   * The action runner instance bound to docpad
   * @private
   * @property {Object} actionRunnerInstance
   */
  DocPad.prototype.actionRunnerInstance = null;

  /**
   * Event Listing. String array of event names.
   * Whenever an event is created, it must be applied here to be available to plugins and configuration files
   * Events must be sorted by the order of execution, not for a functional need, but for a documentation need
   * Whenever this array changes, also update: https://docpad.org/docs/events/
   * @private
   * @property {Array} string array of event names
   */
  DocPad.prototype.events = [
    'extendCollections', // fired each load
    'extendTemplateData', // fired each load
    'docpadLoaded', // fired multiple times, first time command line configuration hasn't been applied yet
    'docpadReady', // fired only once
    'docpadDestroy', // fired once on shutdown
    'consoleSetup', // fired once
    'generateBefore',
    'populateCollectionsBefore',
    'populateCollections',
    'contextualizeBefore',
    'contextualizeAfter',
    'renderBefore',
    'renderCollectionBefore',
    'render', // fired for each extension conversion
    'renderDocument', // fired for each document render, including layouts and render passes
    'renderCollectionAfter',
    'renderAfter',
    'writeBefore',
    'writeAfter',
    'generateAfter',
    'generated',
    'serverBefore',
    'serverExtend',
    'serverAfter',
    'notify'
  ];

  // ---------------------------------
  // Collections

  // Database collection
  /**
   * QueryEngine collection
   * @private
   * @property {Object} database
   */
  DocPad.prototype.database = null;

  /**
   * A FilesCollection of models updated
   * from the DocPad database after each regeneration.
   * @private
   * @property {Object} databaseTempCache FileCollection of models
   */
  DocPad.prototype.databaseTempCache = null;

  /**
   * Files by url. Used to speed up fetching
   * @private
   * @property {Object} filesByUrl
   */
  DocPad.prototype.filesByUrl = null;

  /**
   * Files by Selector. Used to speed up fetching
   * @private
   * @property {Object} filesBySelector
   */
  DocPad.prototype.filesBySelector = null;

  /**
   * Files by Out Path. Used to speed up conflict detection. Do not use for anything else
   * @private
   * @property {Object} filesByOutPath
   */
  DocPad.prototype.filesByOutPath = null;

  /**
   * Blocks
   * @private
   * @property {Object} blocks
   */
  DocPad.prototype.blocks = null;

  /**
   * The DocPad collections
   * @private
   * @property {Object} collections
   */
  DocPad.prototype.collections = null;

  // ---------------------------------
  // Skeletons
  /**
   * Skeletons Collection
   * @private
   * @property {Object} skeletonsCollection
   */
  DocPad.prototype.skeletonsCollection = null;

  // ---------------------------------
  // Plugins
  /**
   * Plugins that are loading really slow
   * @property {Object} slowPlugins
   */
  DocPad.prototype.slowPlugins = null; // {}

  /**
   * Loaded plugins indexed by name
   * @property {Object} loadedPlugins
   */
  DocPad.prototype.loadedPlugins = null; // {}

  /**
   * A listing of all the available extensions for DocPad
   * @property {Object} exchange
   */
  DocPad.prototype.exchange = null; // {}

  
  // -----------------------------
  // Paths
  /**
   * The DocPad directory
   * @property {String} corePath
   */
  DocPad.prototype.corePath = corePath;

  /**
   * The DocPad library directory
   * @private
   * @property {String} libPath
   */
  DocPad.prototype.libPath = __dirname;

  /**
   * The main DocPad file
   * @property {String} mainPath
   */
  DocPad.prototype.mainPath = pathUtil.resolve(__dirname, 'docpad');

  /**
   * The DocPad package.json path
   * @property {String} packagePath
   */
  DocPad.prototype.packagePath = pathUtil.resolve(__dirname, '..', '..', 'package.json');

  /**
   * The DocPad locale path
   * @property {String} localePath
   */
  DocPad.prototype.localePath = pathUtil.resolve(__dirname, '..', '..', 'locale');

  /**
   * The DocPad debug log path (docpad-debug.log)
   * @property {String} debugLogPath
   */
  DocPad.prototype.debugLogPath = pathUtil.join(process.cwd(), 'docpad-debug.log');

  /**
   * The User's configuration path (.docpad.cson)
   * @property {String} userConfigPath
   */
  DocPad.prototype.userConfigPath = '.docpad.cson';

  // -----------------------------
  // Template Data
  /**
   * Description for initialTemplateData
   * @private
   * @property {Object} initialTemplateData
   */
  DocPad.prototype.initialTemplateData = null; // {}

  /**
   * Plugin's Extended Template Data
   * @private
   * @property {Object} pluginsTemplateData
   */
  DocPad.prototype.pluginsTemplateData = null; // {}

  // -----------------------------
  // Locales
  /**
   * Determined locale
   * @private
   * @property {Object} locale
   */
  DocPad.prototype.locale = null;

  // -----------------------------
  // Configuration
  /**
   * Hash Key
   * The key that we use to hash some data before sending it to our statistic server
   * @private
   * @property {String} string constant
   */
  DocPad.prototype.hashKey = '7>9}$3hP86o,4=@T'; // const

  /**
   * Website Package Configuration
   * @private
   * @property {Object} websitePackageConfig
   */
  DocPad.prototype.websitePackageConfig = null; // {}

  /**
   * Merged Configuration
   * Merged in the order of:
   * - initialConfig
   * - userConfig
   * - websiteConfig
   * - instanceConfig
   * - environmentConfig
   * Use getConfig to retrieve this value
   * @private
   * @property {Object} config
   */
  DocPad.prototype.config = null; // {}

  /**
   * Instance Configuration

   * @private
   * @property {Object} instanceConfig
   */
  DocPad.prototype.instanceConfig = null; // {}

  /**
   * Website Configuration
   * Merged into the config property
   * @private
   * @property {Object} websiteConfig
   */
  DocPad.prototype.websiteConfig = null; // {}

  /**
   * User Configuraiton
   * Merged into the config property
   * @private
   * @property {Object} userConfig
   */
  DocPad.prototype.userConfig = null; // {}

  /**
   * Initial Configuration. The default docpadConfig
   * settings that can be overridden in a project's docpad.coffee file.
   * Merged into the config property
   * @private
   * @property {Object} initialConfig
   */
  DocPad.prototype.initialConfig = {
    // -----------------------------
    // Plugins

    // Force re-install of all plugin dependencies
    force: false,
    // Whether or not we should use the global docpad instance
    global: false,
    // Whether or not we should enable plugins that have not been listed or not
    enableUnlistedPlugins: true,
    // Plugins which should be enabled or not pluginName: pluginEnabled
    enabledPlugins: {},
    // Whether or not we should skip unsupported plugins
    skipUnsupportedPlugins: true,
    // Whether or not to warn about uncompiled private plugins
    warnUncompiledPrivatePlugins: true,
    // Configuration to pass to any plugins pluginName: pluginConfiguration
    plugins: {},
    // -----------------------------
    // Project Paths

    // The project directory
    rootPath: process.cwd(),
    // The project's database cache path
    databaseCachePath: '.docpad.db',
    // The project's package.json path
    packagePath: 'package.json',
    // The project's configuration paths
    // Reads only the first one that exists
    // If you want to read multiple configuration paths, then point it to a coffee|js file that requires
    // the other paths you want and exports the merged config
    configPaths: ['docpad.js', 'docpad.coffee', 'docpad.json', 'docpad.cson'],
    // Plugin directories to load
    pluginPaths: [],
    // The project's plugins directory
    pluginsPaths: ['node_modules', 'plugins'],
    // Paths that we should watch for reload changes in
    reloadPaths: [],
    // Paths that we should watch for regeneration changes in
    regeneratePaths: [],
    // The time to wait after a source file has changed before using it to regenerate
    regenerateDelay: 100,
    // The time to wait before outputting the files we are waiting on
    slowFilesDelay: 20 * 1000,
    // The project's out directory
    outPath: 'out',
    // The project's src directory
    srcPath: 'src',
    // The project's documents directories
    // relative to the srcPath
    documentsPaths: ['documents', 'render'],
    // The project's files directories
    // relative to the srcPath
    filesPaths: ['files', 'static', 'public'],
    // The project's layouts directory
    // relative to the srcPath
    layoutsPaths: ['layouts'],
    // Ignored file patterns during directory parsing
    ignorePaths: false,
    ignoreHiddenFiles: false,
    ignoreCommonPatterns: true,
    ignoreCustomPatterns: false,
    // Watch options
    watchOptions: null,
    // -----------------------------
    // Server

    // Port
    // The port that the server should use
    // Defaults to these environment variables:
    // - PORT — Heroku, Nodejitsu, Custom
    // - VCAP_APP_PORT — AppFog
    // - VMC_APP_PORT — CloudFoundry
    port: null,
    // Hostname
    // The hostname we wish to listen to
    // Defaults to these environment variables:
    // HOSTNAME — Generic
    // Do not set to "localhost" it does not work on heroku
    hostname: null,
    // Max Age
    // The caching time limit that is sent to the client
    maxAge: 86400000,
    // Server
    // The Express.js server that we want docpad to use
    serverExpress: null,
    // The HTTP server that we want docpad to use
    serverHttp: null,
    // Extend Server
    // Whether or not we should extend the server with extra middleware and routing
    extendServer: true,
    // Which middlewares would you like us to activate
    // The standard middlewares (bodyParser, methodOverride, express router)
    middlewareStandard: true,
    // The standard bodyParser middleware
    middlewareBodyParser: true,
    // The standard methodOverride middleware
    middlewareMethodOverride: true,
    // The standard express router middleware
    middlewareExpressRouter: true,
    // Our own 404 middleware
    middleware404: true,
    // Our own 500 middleware
    middleware500: true,
    // -----------------------------
    // Logging

    // Log Level
    // Which level of logging should we actually output
    logLevel: ((indexOf.call(process.argv, '-d') >= 0) ? 7 : 6),
    // Catch uncaught exceptions
    catchExceptions: true,
    // Color
    // Whether or not our terminal output should have color
    // `null` will default to what the terminal supports
    color: null,
    // -----------------------------
    // Other

    // Utilise the database cache
    databaseCache: false, // [false, true, 'write']
    
    // Detect Encoding
    // Should we attempt to auto detect the encoding of our files?
    // Useful when you are using foreign encoding (e.g. GBK) for your files
    detectEncoding: false,
    // Render Single Extensions
    // Whether or not we should render single extensions by default
    renderSingleExtensions: false,
    // Render Passes
    // How many times should we render documents that reference other documents?
    renderPasses: 1,
    // Offline
    // Whether or not we should run in offline mode
    // Offline will disable the following:
    // - checkVersion
    // - fetching skeletons
    offline: false,
    // Check Version
    // Whether or not to check for newer versions of DocPad
    checkVersion: false,
    // Prompts
    // Whether or not we should display any prompts
    prompts: false,
    // Progress
    // Whether or not we should display the progress in the terminal title bar
    // Requires prompts being true, and log level 6 or above
    progress: true,
    // Powered By DocPad
    // Whether or not we should include DocPad in the Powered-By meta header
    // Please leave this enabled as it is a standard practice and promotes DocPad in the web eco-system
    poweredByDocPad: true,
    // Helper Url
    // Helper's source-code can be found at: https://github.com/docpad/helper
    helperUrl: true ? 'http://helper.docpad.org/' : 'http://localhost:8000/',
    // Template Data
    // What data would you like to expose to your templates
    templateData: {},
    // Collections
    // A hash of functions that create collections
    collections: {},
    // Events
    // A hash of event handlers
    events: {},
    // Regenerate Every
    // Performs a regenerate every x milliseconds, useful for always having the latest data
    regenerateEvery: false,
    // Regerenate Every Options
    // The generate options to use on the regenerate every call
    regenerateEveryOptions: {
      populate: true,
      partial: false
    },
    // -----------------------------
    // Environment Configuration

    // Locale Code
    // The code we shall use for our locale (e.g. en, fr, etc)
    localeCode: null,
    // Environment
    // Whether or not we are in production or development
    // Separate environments using a comma or a space
    env: null,
    // Environments
    // Environment specific configuration to over-ride the global configuration
    environments: {
      development: {
        // Always refresh from server
        maxAge: false,
        // Only do these if we are running standalone (aka not included in a module)
        checkVersion: isUser,
        welcome: isUser,
        prompts: isUser
      }
    }
  };

  /**
   * Regenerate Timer
   * When config.regenerateEvery is set to a value, we create a timer here
   * @private
   * @property {Object} regenerateTimer
   */
  DocPad.prototype.regenerateTimer = null;

  // ---------------------------------
  // Generate

  // Generate Helpers
  /**
   * Has DocPad's generation process started?
   * @private
   * @property {Boolean} generateStarted
   */
  DocPad.prototype.generateStarted = null;

  /**
   * Has DocPad's generation process ended?
   * @private
   * @property {Boolean} generateEnded
   */
  DocPad.prototype.generateEnded = null;

  /**
   * Is DocPad currently generating?
   * @private
   * @property {Boolean} generating
   */
  DocPad.prototype.generating = false;

  /**
   * Has DocPad done at least one generation?
   * True once the first generation has occured.
   * @private
   * @property {Object} generated
   */
  DocPad.prototype.generated = false;

  // ---------------------------------
  // Watch
  /**
   * Array of file watchers
   * @private
   * @property {Array} watchers
   */
  DocPad.prototype.watchers = null;

  return DocPad;

}).call(this);

// ---------------------------------
// Export
module.exports = DocPad;
