// Generated by CoffeeScript 2.3.1
  // =====================================
  // Requires

  // Standard Library
var ConsoleInterface, TaskGroup, docpadUtil, extendr, pathUtil, promptly, safefs, safeps,
  hasProp = {}.hasOwnProperty;

pathUtil = require('path');

// External
safefs = require('safefs');

safeps = require('safeps');

({TaskGroup} = require('taskgroup'));

extendr = require('extendr');

promptly = require('promptly');

// Local
docpadUtil = require('../util');

// =====================================
// Classes
/**
 * Console Interface
 * @constructor
 */
ConsoleInterface = class ConsoleInterface {
  /**
   * Constructor method. Setup the CLI
   * @private
   * @method constructor
   * @param {Object} opts
   * @param {Function} next
   */
  constructor(opts, next) {
    var commander, consoleInterface, docpad, locale;
    // =================================
    // Helpers
    /**
     * Start the CLI
     * @method start
     * @param {Array} argv
     */
    this.start = this.start.bind(this);
    /**
     * Get the commander
     * @method getCommander
     * @return the commander instance
     */
    this.getCommander = this.getCommander.bind(this);
    /**
     * Destructor.
     * @method destroy
     * @param {Object} err
     */
    this.destroy = this.destroy.bind(this);
    /**
     * Wrap Action
     * @method wrapAction
     * @param {Object} action
     * @param {Object} config
     */
    this.wrapAction = this.wrapAction.bind(this);
    /**
     * Perform Action
     * @method performAction
     * @param {Object} action
     * @param {Object} args
     * @param {Object} [config={}]
     */
    this.performAction = this.performAction.bind(this);
    /**
     * Extract Configuration
     * @method extractConfig
     * @param {Object} [customConfig={}]
     * @return {Object} the DocPad config
     */
    this.extractConfig = this.extractConfig.bind(this);
    /**
     * Select a skeleton
     * @method selectSkeletonCallback
     * @param {Object} skeletonsCollection
     * @param {Function} next
     */
    this.selectSkeletonCallback = this.selectSkeletonCallback.bind(this);
    // =================================
    // Actions
    /**
     * Do action
     * @method action
     * @param {Function} next
     * @param {Object} opts
     */
    this.action = this.action.bind(this);
    /**
     * Action initialise
     * @method init
     * @param {Function} next
     */
    this.init = this.init.bind(this);
    /**
     * Generate action
     * @method generate
     * @param {Function} next
     */
    this.generate = this.generate.bind(this);
    /**
     * Help method
     * @method help
     * @param {Function} next
     */
    this.help = this.help.bind(this);
    /**
     * Info method
     * @method info
     * @param {Function} next
     */
    this.info = this.info.bind(this);
    /**
     * Update method
     * @method update
     * @param {Function} next
     * @param {Object} opts
     */
    this.update = this.update.bind(this);
    /**
     * Upgrade method
     * @method upgrade
     * @param {Function} next
     * @param {Object} opts
     */
    this.upgrade = this.upgrade.bind(this);
    /**
     * Install method
     * @method install
     * @param {Function} next
     * @param {Object} opts
     */
    this.install = this.install.bind(this);
    /**
     * Uninstall method
     * @method uninstall
     * @param {Function} next
     * @param {Object} opts
     */
    this.uninstall = this.uninstall.bind(this);
    /**
     * Render method
     * @method render
     * @param {Function} next
     * @param {Object} opts
     */
    this.render = this.render.bind(this);
    /**
     * Run method
     * @method run
     * @param {Function} next
     */
    this.run = this.run.bind(this);
    /**
     * Server method
     * @method server
     * @param {Function} next
     */
    this.server = this.server.bind(this);
    /**
     * Clean method
     * @method clean
     * @param {Function} next
     */
    this.clean = this.clean.bind(this);
    /**
     * Watch method
     * @method watch
     * @param {Function} next
     */
    this.watch = this.watch.bind(this);
    // Prepare
    consoleInterface = this;
    this.docpad = docpad = opts.docpad;
    this.commander = commander = require('commander');
    locale = docpad.getLocale();
    // -----------------------------
    // Global config
    commander.version(docpad.getVersionString()).option('-o, --out <outPath>', locale.consoleOptionOut).option('-c, --config <configPath>', locale.consoleOptionConfig).option('-e, --env <environment>', locale.consoleOptionEnv).option('-d, --debug [logLevel]', locale.consoleOptionDebug, parseInt).option('-g, --global', locale.consoleOptionGlobal).option('-f, --force', locale.consoleOptionForce).option('--no-color', locale.consoleOptionNoColor).option('-p, --port <port>', locale.consoleOptionPort, parseInt).option('--cache', locale.consoleOptionCache).option('--silent', locale.consoleOptionSilent).option('--skeleton <skeleton>', locale.consoleOptionSkeleton).option('--offline', locale.consoleOptionOffline); // commander translates this to the `color` option for us
    // -----------------------------
    // Commands

    // actions
    commander.command('action <actions>').description(locale.consoleDescriptionRun).action(consoleInterface.wrapAction(consoleInterface.action));
    // init
    commander.command('init').description(locale.consoleDescriptionInit).action(consoleInterface.wrapAction(consoleInterface.init));
    // run
    commander.command('run').description(locale.consoleDescriptionRun).action(consoleInterface.wrapAction(consoleInterface.run, {
      _stayAlive: true
    }));
    // server
    commander.command('server').description(locale.consoleDescriptionServer).action(consoleInterface.wrapAction(consoleInterface.server, {
      _stayAlive: true
    }));
    // render
    commander.command('render [path]').description(locale.consoleDescriptionRender).action(consoleInterface.wrapAction(consoleInterface.render, {
      // Disable anything unnecessary or that could cause extra output we don't want
      logLevel: 3, // 3:error, 2:critical, 1:alert, 0:emergency
      checkVersion: false,
      welcome: false,
      prompts: false
    }));
    // generate
    commander.command('generate').description(locale.consoleDescriptionGenerate).action(consoleInterface.wrapAction(consoleInterface.generate));
    // watch
    commander.command('watch').description(locale.consoleDescriptionWatch).action(consoleInterface.wrapAction(consoleInterface.watch, {
      _stayAlive: true
    }));
    // update
    commander.command('update').description(locale.consoleDescriptionUpdate).action(consoleInterface.wrapAction(consoleInterface.update));
    // upgrade
    commander.command('upgrade').description(locale.consoleDescriptionUpgrade).action(consoleInterface.wrapAction(consoleInterface.upgrade));
    // install
    commander.command('install [pluginName]').description(locale.consoleDescriptionInstall).action(consoleInterface.wrapAction(consoleInterface.install));
    // uninstall
    commander.command('uninstall <pluginName>').description(locale.consoleDescriptionUninstall).action(consoleInterface.wrapAction(consoleInterface.uninstall));
    // clean
    commander.command('clean').description(locale.consoleDescriptionClean).action(consoleInterface.wrapAction(consoleInterface.clean));
    // info
    commander.command('info').description(locale.consoleDescriptionInfo).action(consoleInterface.wrapAction(consoleInterface.info));
    // help
    commander.command('help').description(locale.consoleDescriptionHelp).action(consoleInterface.wrapAction(consoleInterface.help));
    // unknown
    commander.command('*').description(locale.consoleDescriptionUnknown).action(consoleInterface.wrapAction(consoleInterface.help));
    // -----------------------------
    // Finish Up

    // Plugins
    docpad.emitSerial('consoleSetup', {consoleInterface, commander}, function(err) {
      if (err) {
        return consoleInterface.destroyWithError(err);
      }
      return next(null, consoleInterface);
    });
    // Chain
    this;
  }

  start(argv) {
    this.commander.parse(argv || process.argv);
    return this;
  }

  getCommander() {
    return this.commander;
  }

  destroy(err) {
    var docpad, locale, logLevel;
    // Prepare
    docpad = this.docpad;
    locale = docpad.getLocale();
    logLevel = docpad.getLogLevel();
    if (err) {
      // Error?
      docpadUtil.writeError(err);
    }
    // Log Shutdown
    docpad.log('info', locale.consoleShutdown);
    // Handle any errors that occur when stdin is closed
    // https://github.com/docpad/docpad/pull/1049
    process.stdin.on('error', function(err) {
      if (6 <= logLevel) {
        return console.error(err);
      }
    });
    // Close stdin
    process.stdin.end();
    // Destroy docpad
    docpad.destroy(function(err) {
      var activeHandles, activeRequests;
      if (err) {
        // Error?
        docpadUtil.writeError(err);
      }
      // Output if we are not in silent mode
      if (6 <= logLevel) {
        // Note any requests that are still active
        activeRequests = process._getActiveRequests();
        if (activeRequests != null ? activeRequests.length : void 0) {
          console.log(`Waiting on the requests:\n${docpadUtil.inspect(activeRequests)}`);
        }
        // Note any handles that are still active
        activeHandles = process._getActiveHandles();
        if (activeHandles != null ? activeHandles.length : void 0) {
          return console.log(`Waiting on the handles:\n${docpadUtil.inspect(activeHandles)}`);
        }
      }
    });
    return this;
  }

  wrapAction(action, config) {
    var consoleInterface;
    consoleInterface = this;
    return function(...args) {
      return consoleInterface.performAction(action, args, config);
    };
  }

  performAction(action, args, config = {}) {
    var completeAction, consoleInterface, docpad, opts, stayAlive;
    // Prepare
    consoleInterface = this;
    docpad = this.docpad;
    // Special Opts
    stayAlive = false;
    if (config._stayAlive) {
      stayAlive = config._stayAlive;
      delete config._stayAlive;
    }
    // Create
    opts = {};
    opts.commander = args.slice(-1)[0];
    opts.args = args.slice(0, -1);
    opts.instanceConfig = extendr.deepDefaults({}, this.extractConfig(opts.commander), config);
    // Complete Action
    completeAction = function(err) {
      var locale;
      // Prepare
      locale = docpad.getLocale();
      // Handle the error
      if (err) {
        docpad.log('error', locale.consoleSuccess);
        return docpad.fatal(err);
      }
      // Success
      docpad.log('info', locale.consoleSuccess);
      if (stayAlive === false) {
        // Shutdown
        return consoleInterface.destroy();
      }
    };
    // Load
    docpad.action('load ready', opts.instanceConfig, function(err) {
      if (err) {
        // Check
        return completeAction(err);
      }
      // Action
      return action(completeAction, opts); // this order for interface actions for b/c
    });
    return this;
  }

  extractConfig(customConfig = {}) {
    var commanderConfig, config, configPath, key, outPath, sourceConfig, value;
    // Prepare
    config = {};
    commanderConfig = this.commander;
    sourceConfig = this.docpad.initialConfig;
    // debug -> logLevel
    if (commanderConfig.debug) {
      if (commanderConfig.debug === true) {
        commanderConfig.debug = 7;
      }
      commanderConfig.logLevel = commanderConfig.debug;
    }
    // silent -> prompt
    if (commanderConfig.silent != null) {
      commanderConfig.prompts = !commanderConfig.silent;
    }
    // cache -> databaseCache
    if (commanderConfig.silent != null) {
      commanderConfig.databaseCache = commanderConfig.cache;
    }
    // config -> configPaths
    if (commanderConfig.config) {
      configPath = pathUtil.resolve(process.cwd(), commanderConfig.config);
      commanderConfig.configPaths = [configPath];
    }
    // out -> outPath
    if (commanderConfig.out) {
      outPath = pathUtil.resolve(process.cwd(), commanderConfig.out);
      commanderConfig.outPath = outPath;
    }
    for (key in commanderConfig) {
      if (!hasProp.call(commanderConfig, key)) continue;
      value = commanderConfig[key];
      if (typeof sourceConfig[key] !== 'undefined') {
        config[key] = value;
      }
    }
    for (key in customConfig) {
      if (!hasProp.call(customConfig, key)) continue;
      value = customConfig[key];
      if (typeof sourceConfig[key] !== 'undefined') {
        config[key] = value;
      }
    }
    // Return config object
    return config;
  }

  selectSkeletonCallback(skeletonsCollection, next) {
    var commander, consoleInterface, docpad, err, locale, skeletonModel, skeletonNames;
    // Prepare
    consoleInterface = this;
    commander = this.commander;
    docpad = this.docpad;
    locale = docpad.getLocale();
    skeletonNames = [];
    // Already selected?
    if (this.commander.skeleton) {
      skeletonModel = skeletonsCollection.get(this.commander.skeleton);
      if (skeletonModel) {
        next(null, skeletonModel);
      } else {
        err = new Error(`Couldn't fetch the skeleton with id ${this.commander.skeleton}`);
        next(err);
      }
      return this;
    }
    // Show
    docpad.log('info', locale.skeletonSelectionIntroduction + '\n');
    skeletonsCollection.forEach(function(skeletonModel) {
      var skeletonDescription, skeletonName;
      skeletonName = skeletonModel.get('name');
      skeletonDescription = skeletonModel.get('description').replace(/\n/g, '\n\t');
      skeletonNames.push(skeletonName);
      return console.log(`  ${skeletonModel.get('position') + 1}.\t${skeletonName}\n  \t${skeletonDescription}\n`);
    });
    // Select
    consoleInterface.choose(locale.skeletonSelectionPrompt, skeletonNames, {}, function(err, choice) {
      var index;
      if (err) {
        return next(err);
      }
      index = skeletonNames.indexOf(choice);
      return next(null, skeletonsCollection.at(index));
    });
    return this;
  }

  /**
   * Prompt for input
   * @method prompt
   * @param {String} message
   * @param {Object} [opts={}]
   * @param {Function} next
   */
  prompt(message, opts = {}, next) {
    if (opts.default) {
      // Default
      message += ` [${opts.default}]`;
    }
    // Options
    opts = extendr.extend({
      trim: true,
      retry: true,
      silent: false
    }, opts);
    // Log
    promptly.prompt(message, opts, next);
    return this;
  }

  /**
   * Confirm an option
   * @method confirm
   * @param {String} message
   * @param {Object} [opts={}]
   * @param {Function} next
   */
  confirm(message, opts = {}, next) {
    // Default
    if (opts.default === true) {
      message += " [Y/n]";
    } else if (opts.default === false) {
      message += " [y/N]";
    }
    // Options
    opts = extendr.extend({
      trim: true,
      retry: true,
      silent: false
    }, opts);
    // Log
    promptly.confirm(message, opts, next);
    return this;
  }

  /**
   * Choose something
   * @method choose
   * @param {String} message
   * @param {Object} choices
   * @param {Object} [opts={}]
   * @param {Function} next
   */
  choose(message, choices, opts = {}, next) {
    var choice, i, index, indexes, j, len, prompt;
    // Default
    message += ` [1-${choices.length}]`;
    indexes = [];
    for (i = j = 0, len = choices.length; j < len; i = ++j) {
      choice = choices[i];
      index = i + 1;
      indexes.push(index);
      message += `\n  ${index}.\t${choice}`;
    }
    // Options
    opts = extendr.extend({
      trim: true,
      retry: true,
      silent: false
    }, opts);
    // Prompt
    prompt = '> ';
    if (opts.default) {
      prompt += ` [${opts.default}]`;
    }
    // Log
    console.log(message);
    promptly.choose(prompt, indexes, opts, function(err, index) {
      if (err) {
        return next(err);
      }
      choice = choices[index - 1];
      return next(null, choice);
    });
    return this;
  }

  action(next, opts) {
    var actions;
    actions = opts.args[0];
    this.docpad.log('info', 'Performing the actions:', actions);
    this.docpad.action(actions, next);
    return this;
  }

  init(next) {
    this.docpad.action('init', next);
    return this;
  }

  generate(next) {
    this.docpad.action('generate', next);
    return this;
  }

  help(next) {
    var help;
    help = this.commander.helpInformation();
    console.log(help);
    next();
    return this;
  }

  info(next) {
    var docpad, info;
    docpad = this.docpad;
    info = docpad.inspector(docpad.config);
    console.log(info);
    next();
    return this;
  }

  update(next, opts) {
    // Act
    this.docpad.action('clean update', next);
    return this;
  }

  upgrade(next, opts) {
    // Act
    this.docpad.action('upgrade', next);
    return this;
  }

  install(next, opts) {
    var plugin;
    // Extract
    plugin = opts.args[0] || null;
    // Act
    this.docpad.action('install', {plugin}, next);
    return this;
  }

  uninstall(next, opts) {
    var plugin;
    // Extract
    plugin = opts.args[0] || null;
    // Act
    this.docpad.action('uninstall', {plugin}, next);
    return this;
  }

  render(next, opts) {
    var basename, commander, data, docpad, filename, renderDocument, renderOpts, stdin, timeout, useStdin;
    // Prepare
    docpad = this.docpad;
    commander = this.commander;
    renderOpts = {};
    // Extract
    filename = opts.args[0] || null;
    basename = pathUtil.basename(filename);
    renderOpts.filename = filename;
    renderOpts.renderSingleExtensions = 'auto';
    // Prepare text
    data = '';
    // Render
    useStdin = true;
    renderDocument = function(complete) {
      // Perform the render
      return docpad.action('render', renderOpts, function(err, result) {
        if (err) {
          return complete(err);
        }
        // Path
        if (commander.out != null) {
          return safefs.writeFile(commander.out, result, complete);
        } else {
          // Stdout
          process.stdout.write(result);
          return complete();
        }
      });
    };
    // Timeout if we don't have stdin
    timeout = docpadUtil.wait(1000, function() {
      // Clear timeout
      timeout = null;
      if (data.replace(/\s+/, '')) {
        // Skip if we are using stdin
        return next();
      }
      // Close stdin as we are not using it
      useStdin = false;
      stdin.pause();
      // Render the document
      return renderDocument(next);
    });
    // Read stdin
    stdin = process.stdin;
    stdin.resume();
    stdin.setEncoding('utf8');
    stdin.on('data', function(_data) {
      return data += _data.toString();
    });
    process.stdin.on('end', function() {
      if (!useStdin) {
        return;
      }
      if (timeout) {
        clearTimeout(timeout);
        timeout = null;
      }
      renderOpts.data = data;
      return renderDocument(next);
    });
    return this;
  }

  run(next) {
    this.docpad.action('run', {
      selectSkeletonCallback: this.selectSkeletonCallback,
      next: next
    });
    return this;
  }

  server(next) {
    this.docpad.action('server generate', next);
    return this;
  }

  clean(next) {
    this.docpad.action('clean', next);
    return this;
  }

  watch(next) {
    this.docpad.action('generate watch', next);
    return this;
  }

};

// =====================================
// Export
module.exports = ConsoleInterface;
