// Generated by CoffeeScript 2.3.1
  // ---------------------------------
  // Requires

  // Standard Library
var PluginLoader, pathUtil, safefs, semver, util,
  indexOf = [].indexOf;

pathUtil = require('path');

util = require('util');

// External
semver = require('semver');

safefs = require('safefs');

PluginLoader = (function() {
  // ---------------------------------
  // Classes

  // Define Plugin Loader
  /**
   * The Plugin Loader class
   * @class PluginLoader
   * @constructor
   */
  class PluginLoader {
    // ---------------------------------
    // Functions
    /**
     * Constructor method
     * @method constructor
     * @param {Object} opts
     * @param {Object} opts.docpad The docpad instance that we are loading plugins for
     * @param {String} opts.dirPath The directory path of the plugin
     * @param {Object} opts.BasePlugin The base plugin class
     */
    constructor({
        docpad: docpad1,
        dirPath,
        BasePlugin
      }) {
      var docpad;
      this.docpad = docpad1;
      this.dirPath = dirPath;
      this.BasePlugin = BasePlugin;
      // Prepare
      docpad = this.docpad;
      // Apply
      this.pluginName = pathUtil.basename(this.dirPath).replace(/^docpad-plugin-/, '');
      this.pluginClass = {};
      this.packageData = {};
      this.nodeModulesPath = pathUtil.resolve(this.dirPath, 'node_modules');
    }

    /**
     * Loads the package.json file and extracts the main path
     * next(err,exists)
     * @method exists
     * @param {Function} next
     */
    exists(next) {
      var failure, packagePath, success;
      // Prepare
      packagePath = this.packagePath || pathUtil.resolve(this.dirPath, "package.json");
      failure = function(err = null) {
        return next(err, false);
      };
      success = function() {
        return next(null, true);
      };
      // Check the package
      safefs.exists(packagePath, (exists) => {
        if (!exists) {
          return failure();
        }
        // Apply
        this.packagePath = packagePath;
        // Read the package
        return safefs.readFile(packagePath, (err, data) => {
          var pluginPath, pluginVersion;
          if (err) {
            return failure(err);
          }
          try {
            // Parse the package
            this.packageData = JSON.parse(data.toString());
          } catch (error) {
            err = error;
            return failure(err);
          } finally {
            if (!this.packageData) {
              return failure();
            }
          }
          // Extract the version and main
          pluginVersion = this.packageData.version;
          pluginPath = this.packageData.main && pathUtil.join(this.dirPath, this.packageData.main);
          if (!pluginVersion) {
            // Check defined
            return failure();
          }
          if (!pluginPath) {
            return failure();
          }
          // Success
          this.pluginVersion = pluginVersion;
          this.pluginPath = pluginPath;
          return success();
        });
      });
      return this;
    }

    /**
     * Check if this plugin is unsupported
     * Boolean value returned as a parameter
     * in the passed callback
     * next(err,supported)
     * @method unsupported
     * @param {Function} next
     */
    unsupported(next) {
      var docpad, engines, keywords, peerDependencies, platforms, ref, unsupported, version;
      // Prepare
      docpad = this.docpad;
      // Extract
      version = this.packageData.version;
      keywords = this.packageData.keywords || [];
      platforms = this.packageData.platforms || [];
      engines = this.packageData.engines || {};
      peerDependencies = this.packageData.peerDependencies || {};
      // Check
      // Check type
      // Check version
      // Check platform
      // Check node engine
      // Check docpad engine
      // the .replace is to support version flags, such as -beta
      // Check docpad peerDependencies
      // the .replace is to support version flags, such as -beta
      // Supported
      unsupported = indexOf.call(keywords, 'docpad-plugin') < 0 ? 'type' : version && !semver.satisfies(version, docpad.pluginVersion) ? 'version-plugin' : platforms.length && (ref = process.platform, indexOf.call(platforms, ref) < 0) ? 'platform' : (engines.node != null) && !semver.satisfies(process.version, engines.node) ? 'engine-node' : (engines.docpad != null) && !semver.satisfies(docpad.getVersion().replace(/-.+/, ''), engines.docpad) ? 'version-docpad' : (peerDependencies.docpad != null) && !semver.satisfies(docpad.getVersion().replace(/-.+/, ''), peerDependencies.docpad) ? 'version-docpad' : false;
      // Supported
      next(null, unsupported);
      return this;
    }

    /**
     * Installs the plugins node modules.
     * next(err)
     * @private
     * @method install
     * @param {Function} next
     */
    install(next) {
      var docpad;
      // Prepare
      docpad = this.docpad;
      // Only install if we have a package path
      if (this.packagePath) {
        // Install npm modules
        docpad.initNodeModules({
          path: this.dirPath,
          next: function(err, results) {
            // Forward
            return next(err);
          }
        });
      } else {
        // Continue
        next();
      }
      return this;
    }

    /**
     * Load in the pluginClass from the plugin file.
     * The plugin class that has been loaded is returned
     * in the passed callback
     * next(err,pluginClass)
     * @method load
     * @param {Function} next
     */
    load(next) {
      var base, docpad, err, locale, pluginPrototypeName, ref, ref1, ref2, validPluginName;
      // Prepare
      docpad = this.docpad;
      locale = docpad.getLocale();
      // Ensure we still have deprecated support for old-style uncompiled plugins
      if (pathUtil.extname(this.pluginPath) === '.coffee') {
        // Warn the user they are trying to include an uncompiled plugin (if they want to be warned)
        // They have the option of opting out of warnings for private plugins
        if (!(this.packageData.private === true && docpad.getConfig().warnUncompiledPrivatePlugins === false)) {
          docpad.warn(util.format(locale.pluginUncompiled, this.pluginName, ((ref = this.packageData.bugs) != null ? ref.url : void 0) || locale.pluginIssueTracker));
        }
        try {
          // Attempt to include the coffee-script register extension
          // coffee-script is an external party dependency (docpad doesn't depend on it, so we don't install it)
          // so we may not have it, hence the try catch
          require('coffeescript/register');
        } catch (error) {
          err = error;
          // Including coffee-script has failed, so let the user know, and exit
          err.context = util.format(locale.pluginUncompiledFailed, this.pluginName, ((ref1 = this.packageData.bugs) != null ? ref1.url : void 0) || locale.pluginIssueTracker);
          return next(err);
          this;
        }
      }
      try {
        // Attempt to load the plugin
        this.pluginClass = require(this.pluginPath)(this.BasePlugin);
      } catch (error) {
        err = error;
        // Loading the plugin has failed, so let the user know, and exit
        err.context = util.format(locale.pluginLoadFailed, this.pluginName, ((ref2 = this.packageData.bugs) != null ? ref2.url : void 0) || locale.pluginIssueTracker);
        return next(err);
        this;
      }
      // Plugin loaded, inject it's version and grab its name
      if ((base = this.pluginClass.prototype).version == null) {
        base.version = this.pluginVersion;
      }
      pluginPrototypeName = this.pluginClass.prototype.name;
      // Check Alphanumeric Name
      if (/^[a-z0-9]+$/.test(this.pluginName) === false) {
        validPluginName = this.pluginName.replace(/[^a-z0-9]/, '');
        docpad.warn(util.format(locale.pluginNamingConventionInvalid, this.pluginName, validPluginName));
      }
      // Check for Empty Name
      if (pluginPrototypeName === null) {
        this.pluginClass.prototype.name = this.pluginName;
        docpad.warn(util.format(locale.pluginPrototypeNameUndefined, this.pluginName));
      // Check for Same Name
      } else if (pluginPrototypeName !== this.pluginName) {
        docpad.warn(util.format(locale.pluginPrototypeNameDifferent, this.pluginName, pluginPrototypeName));
      }
      // Return our plugin
      next(null, this.pluginClass);
      return this;
    }

    /**
     * Create an instance of a plugin
     * defined by the passed config.
     * The plugin instance is returned in
     * the passed callback.
     * next(err,pluginInstance)
     * @method create
     * @param {Object} config
     * @param {Function} next
     */
    create(config, next) {
      var docpad, err, pluginInstance;
      try {
        // Create instance with merged configuration
        // Load
        docpad = this.docpad;
        pluginInstance = new this.pluginClass({docpad, config});
      } catch (error) {
        err = error;
        // An error occured, return it
        return next(err, null);
      }
      // Return our instance
      return next(null, pluginInstance);
      return this;
    }

  };

  // ---------------------------------
  // Constructed
  /**
   * The DocPad Instance
   * @private
   * @property {Object} docpad
   */
  PluginLoader.prototype.docpad = null;

  /**
   * The BasePlugin Class
   * @private
   * @property {Object}
   */
  PluginLoader.prototype.BasePlugin = null;

  /**
   * The full path of the plugin's directory
   * @private
   * @property {String}
   */
  PluginLoader.prototype.dirPath = null;

  // ---------------------------------
  // Loaded
  /**
   * The full path of the plugin's package.json file
   * @private
   * @property {String}
   */
  PluginLoader.prototype.packagePath = null;

  /**
   * The parsed contents of the plugin's package.json file
   * @private
   * @property {Object}
   */
  PluginLoader.prototype.packageData = {};

  /**
   * The full path of the plugin's main file
   * @private
   * @property {String}
   */
  PluginLoader.prototype.pluginPath = null;

  /**
   * The parsed content of the plugin's main file
   * @private
   * @property {Object}
   */
  PluginLoader.prototype.pluginClass = {};

  /**
   * The plugin name
   * @private
   * @property {String}
   */
  PluginLoader.prototype.pluginName = null;

  /**
   * The plugin version
   * @private
   * @property {String}
   */
  PluginLoader.prototype.pluginVersion = null;

  /**
   * Node modules path
   * @private
   * @property {String}
   */
  PluginLoader.prototype.nodeModulesPath = null;

  return PluginLoader;

}).call(this);

// ---------------------------------
// Export
module.exports = PluginLoader;
