// Generated by CoffeeScript 2.3.1
// ---------------------------------
// Requires

// External
var BasePlugin, eachr, extendr, typeChecker, unbounded;

extendr = require('extendr');

typeChecker = require('typechecker');

eachr = require('eachr');

unbounded = require('unbounded');

BasePlugin = (function() {
  // ---------------------------------
  // Classes

  // Define Plugin
  /**
   * The base class for all DocPad plugins
   * @class BasePlugin
   * @constructor
   */
  class BasePlugin {
    /**
     * Constructor method for the plugin
     * @method constructor
     * @param {Object} opts
     */
    constructor(opts) {
      var config, docpad, e, me;
      /**
       * Set Configuration
       * @private
       * @method {Object} setConfig
       * @param {Object} [instanceConfig=null]
       */
      this.setConfig = this.setConfig.bind(this);
      /**
       * Get the Configuration
       * @private
       * @method {Object}
       */
      this.getConfig = this.getConfig.bind(this);
      // Prepare
      me = this;
      ({docpad, config} = opts);
      this.docpad = docpad;
      // Bind listeners
      this.bindListeners();
      // Setup the configuration so that there are not reference errors
      this.config = extendr.clone(this.config);
      this.instanceConfig = extendr.clone(this.instanceConfig);
      if (typeChecker.isEmptyObject(this.initialConfig)) {
        this.initialConfig = extendr.clone(this.config);
      } else if (typeChecker.isEmptyObject(this.config)) {
        try {
          this.initialConfig = extendr.clone(this.initialConfig);
        } catch (error) {
          e = error;
        }
      } else {
        // ignore, as must be exposed via a getter, in which case it won't have reference issues
        throw new Error(`Plugin ${this.name} is misconfigured, it has both @config and @initialConfig defined, it can only have one or the other`);
      }
      // Apply the configuration
      this.setConfig(config);
      if (this.isEnabled() === false) {
        // Return early if we are disabled
        return this;
      }
      // Listen to events
      this.addListeners();
      // Chain
      this;
    }

    /**
     * Set Instance Configuration
     * @private
     * @method setInstanceConfig
     * @param {Object} instanceConfig
     */
    setInstanceConfig(instanceConfig) {
      // Merge in the instance configurations
      if (instanceConfig) {
        extendr.deepDefaults(this.instanceConfig, instanceConfig);
        if (this.config) {
          extendr.deepDefaults(this.config, instanceConfig);
        }
      }
      return this;
    }

    setConfig(instanceConfig = null) {
      var configPackages, configsToMerge, docpad, userConfig;
      // Prepare
      docpad = this.docpad;
      userConfig = this.docpad.config.plugins[this.name];
      this.config = this.docpad.config.plugins[this.name] = {};
      if (instanceConfig) {
        // Instance config
        this.setInstanceConfig(instanceConfig);
      }
      // Merge configurations
      configPackages = [this.initialConfig, userConfig, this.instanceConfig];
      configsToMerge = [this.config];
      docpad.mergeConfigurations(configPackages, configsToMerge);
      if (!this.isEnabled()) {
        // Remove listeners if we are disabled
        this.removeListeners();
      }
      return this;
    }

    getConfig() {
      return this.config;
    }

    /**
     * Alias for b/c
     * @private
     * @method bindEvents
     */
    bindEvents() {
      return this.addListeners();
    }

    /**
     * Bind Listeners
     * @private
     * @method bindListeners
     */
    bindListeners() {
      var docpad, events, pluginInstance;
      // Prepare
      pluginInstance = this;
      docpad = this.docpad;
      events = docpad.getEvents();
      // Bind events
      eachr(events, function(eventName) {
        var eventHandler;
        // Fetch the event handler
        eventHandler = pluginInstance[eventName];
        // Check it exists and is a function
        if (typeChecker.isFunction(eventHandler)) {
          // Bind the listener to the plugin
          return pluginInstance[eventName] = eventHandler.bind(pluginInstance);
        }
      });
      return this;
    }

    /**
     * Add Listeners
     * @private
     * @method addListeners
     */
    addListeners() {
      var docpad, events, pluginInstance;
      // Prepare
      pluginInstance = this;
      docpad = this.docpad;
      events = docpad.getEvents();
      // Bind events
      eachr(events, function(eventName) {
        var eventHandler, eventHandlerPriority, ignoredError;
        // Fetch the event handler
        eventHandler = pluginInstance[eventName];
        // Check it exists and is a function
        if (typeChecker.isFunction(eventHandler)) {
          // Apply the priority
          eventHandlerPriority = pluginInstance[eventName + 'Priority'] || pluginInstance.priority || null;
          if (eventHandler.priority == null) {
            eventHandler.priority = eventHandlerPriority;
          }
          try {
            eventHandler.name = `${pluginInstance.name}: {eventName}`;
            if (eventHandler.priority != null) {
              eventHandler.name += "(priority eventHandler.priority})";
            }
          } catch (error) {
            ignoredError = error;
          }
          // newer versions of node do not allow writing the name property on functions, as it is readonly

          // Wrap the event handler, and bind it to docpad
          return docpad.off(eventName, eventHandler).on(eventName, eventHandler);
        }
      });
      return this;
    }

    /**
     * Remove Listeners
     * @private
     * @method removeListeners
     */
    removeListeners() {
      var docpad, events, pluginInstance;
      // Prepare
      pluginInstance = this;
      docpad = this.docpad;
      events = docpad.getEvents();
      // Bind events
      eachr(events, function(eventName) {
        var eventHandler;
        // Fetch the event handler
        eventHandler = pluginInstance[eventName];
        // Check it exists and is a function
        if (typeChecker.isFunction(eventHandler)) {
          // Wrap the event handler, and unbind it from docpad
          return docpad.off(eventName, eventHandler);
        }
      });
      return this;
    }

    /**
     * Destructor. Calls removeListeners
     * @private
     * @method destroy
     */
    destroy() {
      this.removeListeners();
      return this;
    }

    /**
     * Is Enabled?
     * @method isEnabled
     * @return {Boolean}
     */
    isEnabled() {
      return this.config.enabled !== false;
    }

  };

  /**
   * Add support for BasePlugin.extend(proto)
   * @private
   * @property {Object} @extend
   */
  BasePlugin.extend = require('csextends');

  // ---------------------------------
  // Inherited
  /**
   * The DocPad Instance
   * @private
   * @property {Object} docpad
   */
  BasePlugin.prototype.docpad = null;

  // ---------------------------------
  // Variables
  /**
   * The plugin name
   * @property {String}
   */
  BasePlugin.prototype.name = null;

  /**
   * The plugin config
   * @property {Object}
   */
  BasePlugin.prototype.config = {};

  /**
   * The instance config.
   * @property {Object}
   */
  BasePlugin.prototype.instanceConfig = {};

  /**
   * Plugin priority
   * @private
   * @property {Number}
   */
  BasePlugin.prototype.priority = 500;

  return BasePlugin;

}).call(this);

// ---------------------------------
// Export Plugin
module.exports = BasePlugin;
