// Generated by CoffeeScript 2.3.1
  // =====================================
  // Requires

  // Standard Library
var CSON, DocPad, PluginTester, RendererTester, ServerTester, assert, balUtil, deepEqual, difference, docpadUtil, equal, errorEqual, extendr, joe, pathUtil, pluginPort, safefs, test, testers,
  indexOf = [].indexOf;

pathUtil = require('path');

// External
safefs = require('safefs');

balUtil = require('bal-util');

extendr = require('extendr');

joe = require('joe');

assert = require('assert');

({equal, deepEqual, errorEqual} = require('assert-helpers'));

CSON = require('cson');

({difference} = require('underscore'));

// Local
DocPad = require('./docpad');

docpadUtil = require('./util');

// =====================================
// Helpers

// Prepare
// We want the plugn port to be a semi-random number above 2000
pluginPort = 2000 + parseInt(String(Date.now()).substr(-6, 4));

testers = {CSON, DocPad};

// ---------------------------------
// Classes
/**
 * The Plugin Tester class
 * @class PluginTester
 * @constructor
 */
testers.PluginTester = PluginTester = (function() {
  class PluginTester {
    /**
     * Constructor method
     * @method constructor
     * @param {Object} [config={}]
     * @param {Object} [docpadConfig={}]
     * @param {Function} next
     */
    constructor(config = {}, docpadConfig = {}, next) {
      var base, base1, base2, base3, base4, base5, base6, base7, base8, defaultEnabledPlugins, tester;
      /**
       * Create the DocPad instance
       * @method testCreate
       */
      this.testCreate = this.testCreate.bind(this);
      /**
       * Test Loaded
       * @method
       */
      this.testLoad = this.testLoad.bind(this);
      // Perform Server
      this.testServer = this.testServer.bind(this);
      /**
       * Test generate
       * @method
       */
      this.testGenerate = this.testGenerate.bind(this);
      /**
       * Test everything
       * @method {Object}
       */
      this.testEverything = this.testEverything.bind(this);
      // Apply Configuration
      tester = this;
      this.config = extendr.deep({}, PluginTester.prototype.config, this.config, config);
      this.docpadConfig = extendr.deep({}, PluginTester.prototype.docpadConfig, this.docpadConfig, docpadConfig);
      if ((base = this.docpadConfig).port == null) {
        base.port = ++pluginPort;
      }
      if ((base1 = this.config).testerName == null) {
        base1.testerName = `${this.config.pluginName} plugin`;
      }
      // Extend Configuration
      (base2 = this.config).testPath || (base2.testPath = pathUtil.join(this.config.pluginPath, 'test'));
      (base3 = this.config).outExpectedPath || (base3.outExpectedPath = pathUtil.join(this.config.testPath, 'out-expected'));
      // Extend DocPad Configuration
      (base4 = this.docpadConfig).rootPath || (base4.rootPath = this.config.testPath);
      (base5 = this.docpadConfig).outPath || (base5.outPath = pathUtil.join(this.docpadConfig.rootPath, 'out'));
      (base6 = this.docpadConfig).srcPath || (base6.srcPath = pathUtil.join(this.docpadConfig.rootPath, 'src'));
      if ((base7 = this.docpadConfig).pluginPaths == null) {
        base7.pluginPaths = [this.config.pluginPath];
      }
      defaultEnabledPlugins = {};
      defaultEnabledPlugins[this.config.pluginName] = true;
      (base8 = this.docpadConfig).enabledPlugins || (base8.enabledPlugins = defaultEnabledPlugins);
      // Test API
      joe.describe(this.config.testerName, function(suite, task) {
        tester.describe = tester.suite = suite;
        tester.it = tester.test = task;
        tester.done = tester.exit = function(next) {
          var ref;
          return (ref = tester.docpad) != null ? ref.action('destroy', next) : void 0;
        };
        return typeof next === "function" ? next(null, tester) : void 0;
      });
      // Chain
      this;
    }

    /**
     * Get tester Configuration
     * @method getConfig
     * @return {Object}
     */
    getConfig() {
      return this.config;
    }

    /**
     * Get the plugin instance
     * @method getPlugin
     * @return {Object} the plugin
     */
    getPlugin() {
      return this.docpad.getPlugin(this.getConfig().pluginName);
    }

    testCreate() {
      var docpadConfig, tester;
      // Prepare
      tester = this;
      docpadConfig = this.docpadConfig;
      // Create Instance
      this.test("create", function(done) {
        return new DocPad(docpadConfig, function(err, docpad) {
          if (err) {
            return done(err);
          }
          tester.docpad = docpad;
          // init docpad in case the plugin is starting from scratch
          return tester.docpad.action('init', function(err) {
            if (err && err.message !== tester.docpad.getLocale().skeletonExists) {
              return done(err); // care about the error providing it isn't the skeleton exists error
            }
            
            // clean up the docpad out directory
            return tester.docpad.action('clean', function(err) {
              if (err) {
                return done(err);
              }
              // install anything on the website that needs to be installed
              return tester.docpad.action('install', done);
            });
          });
        });
      });
      return this;
    }

    testLoad() {
      var tester;
      // Prepare
      tester = this;
      // Test
      this.test(`load plugin ${tester.config.pluginName}`, function(done) {
        return tester.docpad.loadedPlugin(tester.config.pluginName, function(err, loaded) {
          if (err) {
            return done(err);
          }
          assert.ok(loaded);
          return done();
        });
      });
      return this;
    }

    testServer(next) {
      var tester;
      // Prepare
      tester = this;
      // Handle
      this.test("server", function(done) {
        return tester.docpad.action('server', function(err) {
          return done(err);
        });
      });
      return this;
    }

    testGenerate() {
      var tester;
      // Prepare
      tester = this;
      // Test
      this.test("generate", function(done) {
        return tester.docpad.action('generate', function(err) {
          return done(err);
        });
      });
      return this;
    }

    testEverything() {
      var tester;
      // Prepare
      tester = this;
      // Tests
      this.testCreate();
      this.testLoad();
      this.testGenerate();
      this.testServer();
      if (typeof this.testCustom === "function") {
        this.testCustom();
      }
      // Finish
      this.finish();
      return this;
    }

    /**
     * Finish
     * @method finish
     */
    finish() {
      var tester;
      // Prepare
      tester = this;
      // Finish
      if (tester.config.autoExit) {
        this.test('finish up', function(next) {
          return tester.exit(next);
        });
      }
      return this;
    }

  };

  // Add support for BasePlugin.extend(proto)
  PluginTester.extend = require('csextends');

  /**
   * Default plugin config
   * @property {Object}
   */
  PluginTester.prototype.config = {
    testerName: null,
    pluginName: null,
    pluginPath: null,
    testPath: null,
    outExpectedPath: null,
    removeWhitespace: false,
    contentRemoveRegex: null,
    autoExit: 'safe'
  };

  /**
   * Default DocPad config
   * @property {Object}
   */
  PluginTester.prototype.docpadConfig = {
    global: true,
    port: null,
    logLevel: ((indexOf.call(process.argv, '-d') >= 0) ? 7 : 5),
    rootPath: null,
    outPath: null,
    srcPath: null,
    pluginPaths: null,
    enableUnlistedPlugins: true,
    enabledPlugins: null,
    skipUnsupportedPlugins: false,
    catchExceptions: false,
    environment: null
  };

  /**
   * The DocPad instance
   * @private
   * @property {Object}
   */
  PluginTester.prototype.docpad = null;

  return PluginTester;

}).call(this);

/**
 * Server tester
 * @class ServerTester
 * @extends PluginTester
 * @constructor
 */
testers.ServerTester = ServerTester = class ServerTester extends PluginTester {};

/**
 * Rednderer tester
 * @class ServerTester
 * @extends PluginTester
 * @constructor
 */
testers.RendererTester = RendererTester = class RendererTester extends PluginTester {
  // Test Generation
  testGenerate() {
    var tester;
    // Prepare
    tester = this;
    // Test
    this.suite("generate", function(suite, test) {
      test('action', function(done) {
        return tester.docpad.action('generate', function(err) {
          return done(err);
        });
      });
      return suite('results', function(suite, test, done) {
        // Get actual results
        return balUtil.scanlist(tester.docpadConfig.outPath, function(err, outResults) {
          if (err) {
            return done(err);
          }
          // Get expected results
          return balUtil.scanlist(tester.config.outExpectedPath, function(err, outExpectedResults) {
            var outExpectedResultsKeys, outResultsKeys;
            if (err) {
              return done(err);
            }
            // Prepare
            outResultsKeys = Object.keys(outResults);
            outExpectedResultsKeys = Object.keys(outExpectedResults);
            // Check we have the same files
            test('same files', function() {
              var outDifferenceKeys;
              outDifferenceKeys = difference(outExpectedResultsKeys, outResultsKeys);
              deepEqual(outDifferenceKeys, [], 'The following file(s) should have been generated');
              outDifferenceKeys = difference(outResultsKeys, outExpectedResultsKeys);
              return deepEqual(outDifferenceKeys, [], 'The following file(s) should not have been generated');
            });
            // Check the contents of those files match
            outResultsKeys.forEach(function(key) {
              return test(`same file content for: ${key}`, function() {
                var actual, expected, replaceLinesRegex;
                // Fetch file value
                actual = outResults[key];
                expected = outExpectedResults[key];
                // Remove empty lines
                if (tester.config.removeWhitespace === true) {
                  replaceLinesRegex = /\s+/g;
                  actual = actual.replace(replaceLinesRegex, '');
                  expected = expected.replace(replaceLinesRegex, '');
                }
                // Content regex
                if (tester.config.contentRemoveRegex) {
                  actual = actual.replace(tester.config.contentRemoveRegex, '');
                  expected = expected.replace(tester.config.contentRemoveRegex, '');
                }
                // Compare
                return equal(actual, expected);
              });
            });
            // Forward
            return done();
          });
        });
      });
    });
    return this;
  }

};

/**
 * Test a plugin
 * test({pluginPath: String})
 * @property test
 */
testers.test = test = function(testerConfig, docpadConfig) {
  var complete;
  // Configure
  if (testerConfig.testerClass == null) {
    testerConfig.testerClass = PluginTester;
  }
  testerConfig.pluginPath = pathUtil.resolve(testerConfig.pluginPath);
  if (testerConfig.pluginName == null) {
    testerConfig.pluginName = pathUtil.basename(testerConfig.pluginPath).replace('docpad-plugin-', '');
  }
  if (testerConfig.testerPath == null) {
    testerConfig.testerPath = pathUtil.join('out', `${testerConfig.pluginName}.tester.js`);
  }
  if (testerConfig.testerPath) {
    testerConfig.testerPath = pathUtil.resolve(testerConfig.pluginPath, testerConfig.testerPath);
  }
  // Create tester
  complete = function() {
    if (typeof testerConfig.testerClass === 'string') {
      // Accept string inputs for testerClass
      testerConfig.testerClass = testers[testerConfig.testerClass];
    }
    // Create our tester
    return new testerConfig.testerClass(testerConfig, docpadConfig, function(err, testerInstance) {
      if (err) {
        throw err;
      }
      // Run the tests
      return testerInstance.testEverything();
    });
  };
  // Load the tester file
  if (testerConfig.testerPath) {
    safefs.exists(testerConfig.testerPath, function(exists) {
      if (exists) {
        testerConfig.testerClass = require(testerConfig.testerPath)(testers);
      }
      return complete();
    });
  } else {
    // User the default tester
    complete();
  }
  // Chain
  return testers;
};

// ---------------------------------
// Export Testers
module.exports = testers;
